;+---------------------------------------------------------------------
;
; CONT_FIT       11/2005
;
; Interactive tool for continuum-fitting single- and multiple-order
; spectra.
;
; The program can either compute an initial fit from scratch, or use a
; fit input using the parfile keyword.  The user then adjusts the fit
; for each order as deisred, and can write out the continuum fits and
; the normalized spectrum.
;
; CALLING SEQUENCE
;     cont_fit,fluxfile,errorfile,nspline=nspline,parfile=parfile,
;              smooth=smooth,wavefile=wavefile,/polyfit,
;              calibrated=calibrated
;
; INPUTS
;        fluxfile    - String, flux filename, may be single order
;                      or multiple orders
;        errorfile   - String, error filename, same size as fluxfile.
;                      If not provided, a fake error spectrum with
;                      uniform value will be used.
;        
; KEYWORDS
;        nspline     - Scalar, number of spline points per order to
;                      be used in the initial fit.  Default is 10.
;        parfile     - String, name of ASCII file containing 
;                      tabulated spline points to be used as the initial
;                      spline fit.   Format is three columns with
;                      [order, spline_x, spline_y] , where orders are
;                      numbered consecutively starting from 0.  See also
;                      the 'smooth' keyword.
;        smooth      - Scalar, number of pixels around each spline point
;                      over which to compute the weighted mean for
;                      the initial fit.  Default is 21 (+/- 10 pixels
;                      around a spline point).  If the parfile 
;                      keyword is set and smooth >= 1, then the initial
;                      fit will be computed around the x-values in
;                      the par file nd the y-values will not be used.
;                      If smooth=-1, then the initial fit will use
;                      the literal y-values in the parfile. 
;        wavefile    - String, wavelength filename, same size as
;                      fluxfile.  Default is to compute wavelengths
;                      from information in the flux file header, 
;                      although this is not currently set up for
;                      multiple orders with MAKEE-style wavvelength 
;                      information.
;        polyfit     - If set, do a robust polynomial fit around each
;                      spline point to set initial value, with crude
;                      rejection of absorption lines.  No effect if
;                      the parfile keyword is set.  Note: This does
;                      not typically give the correct continuum in the
;                      presence of strong absorption lines, but it may
;                      provide a more convenient starting point for
;                      fits done by hand.
;        ginterp     - Four element array.  If set, use a gaussian
;                      interpolation of the spectrum as the basis for
;                      the spline points.  Values are
;                      [wmin,wmax,degree,dv], where wmin and wmax are
;                      the min and max wavelengths over which the
;                      continuum is well sampled, degree is the degree
;                      of the polynomial fit used for doing local
;                      polynomial fits around each pixel (default =
;                      5), and dv is the width of the Gaussian
;                      weighting function, in km/s (default = 1000).
;                      If the keyword is set with no parameters, then
;                      the defaults are used, and wmin and wmax are
;                      set to the min and max input wavelengths.
;                      Program will attempt to filter out noise spikes
;                      and absorption lines.  Usefule for QSOs redward
;                      of the Lya forest.  If set, a large value for
;                      nspline is recommended.
;        calibrated  - String, FITS filename of 1D spectrum of
;                      calibrated continuum values for the object
;                      being fit.  If set, clicking 'Write Files' will
;                      also generate a response function, with the the
;                      same dimensions as the input flux.  The
;                      response function is calculated by dividing the
;                      the calibrated values by the fitted continuum
;                      (and currently multiplying by 1e30), such that
;                      calibrated flux = raw flux * response.  FITS
;                      files for many spectrophotometric standards can
;                      be obtained from the STScI CALSPEC website.
;        wcalibrated - String, FITS filename of 1D wavelengths for the
;                      calibrated spectrum.  Default is to get the
;                      wavelength information from the header of the
;                      flux file.
;        wcal2vac    - If set, convert 1D wavelengths for the 
;                      calibrated spectrum from air to vacuum.
;
; OUTPUTS
;        Clicking the 'Write Files' button will cause the program to
;        write the following files.  The user may interactively change
;        the file names.  Defaults are derived from the name of the
;        input fluxfile = <flux>.fits and errorfile = <error>.fits.
;             Cont-<flux>.fits  - Continuum FITS file
;             Cont-<flux).par   - ASCII file containing spline points
;             Norm-<flux>.fits  - Normalized flux FITS file
;             Norm-<error>.fits - Normalized error FITS file 
;        And if the calibrated keyword is set,
;             Resp-<flux>.fits  - Response function FITS file
; 
; HISTORY
;        Written 11/5/2005 GDB
;        Added GINTERP keyword 4/27/2009 GDB
;        Added wcal2vac keyword 1/28/2014 GDB
;----------------------------------------------------------------------

;----------------------------------------------------------------------
; Initialize common blocks and other settings
;----------------------------------------------------------------------

Pro cont_fit_initialize

common cf_state, state
common cf_data, wave, flux, error, weight, splfit, spoints, svals, $
                spline_x, spline_y, norm_flux, norm_error, $
                file, fluxhdr, errorhdr, binflux, binerror, $
                binwave, norm_binflux, norm_binerror, input_type, $
                fcalib, wcalib

state = {              $
          plotwindow: 0L, $        ; Plot window number
          plot_xsz: 1000L, $       ; Screen x-size of plot window
          plot_ysz: 750L, $        ; Screen y-size of plot window
          xdispmin: 0d0, $         ; Min display x-value 
          xdispmax: 0d0, $         ; Max display x-value
          ydispmin: 0d0, $         ; Min display x-value 
          ydispmax: 0d0, $         ; Max display x-value
          default_xdispmin: 0d0, $ ; Default min display x-value 
          default_xdispmax: 0d0, $ ; Default max display x-value
          default_ydispmin: 0d0, $ ; Default min display x-value 
          default_ydispmax: 0d0, $ ; Default max display x-value
          binsz: 1L, $             ; Bin size for displayed spectrum
          histogram: 0L, $         ; Flag to plot in histogram mode
          plotnorm: 0L, $          ; Flag to plot the normalized arrays
          minwave: 0L, $           ; Minimum wavelength for the current order
          maxwave: 0L, $           ; Maximum wavelength for the current order
          xcoord: 0., $            ; Data x-coordinate of mouse click
          ycoord: 0., $            ; Data y-coordinate of mouse click
          order: 0L, $             ; Index of current order
          norders: 0L, $           ; Number of orders
          npix: 0L, $              ; Number of pixels in each order
          writeresponse: 0L, $     ; Write response function
          max_spoints: 1000, $      ; Max number of spline points per order
          buttonbase_xsz: 0L, $    ; Screen x-size of button base
          buttonbase_ysz: 0L, $    ; Screen y-size of button base
          base_id: 0L, $           ; WIDGET IDS
          subbase_id: 0L, $        ;
          buttonbase_id: 0L, $     ;
          plotwindow_id: 0L, $     ; 
          write_button_id: 0L, $   ;
          quit_button_id: 0L $     ;
        }

file =  {          $
          flux: '', $
          error: '', $
          norm_flux: '', $
          norm_error: '', $
          cont: '', $
          param: '', $
          response: '' $
        }

end

;------------------------------------------------------------------------
; Widget events (main)
;------------------------------------------------------------------------

Pro cont_fit_window_event,event

common cf_state
common cf_data

widget_control,event.id,get_uvalue=eventval

case eventval of 
    'plotwindow' : begin
                      if event.clicks then begin
                         ; Mouse event: act when mouse button is pressed
                         if (event.press gt 0) then mouse_event,event
                      endif else begin
                         ; Keyboard event: act when key is pressed
                         if (event.press gt 0) then keyboard_event,event
                      endelse
                   end
    'write'      : write_cf_data
    'quit'       : begin
                     reply = ''
                     read,'Really quit? [y]: ',reply
                     if (reply eq '' or reply eq 'y') then begin
                        widget_control,event.top,/destroy
                     endif
                   end
    else         : print,'Unknown event'
endcase

end

;------------------------------------------------------------------------
; Mouse events
;------------------------------------------------------------------------

Pro mouse_event,event

common cf_state

; Convert coordinates from device to data
coords = convert_coord(event.x,event.y,$
                       /device,/to_data)
state.xcoord = coords(0)
state.ycoord = coords(1)

case event.press of
   1: begin  ; Left
         ; Get new x-range
         print,'Select second x-bound'
         cursor,xselect,yselect,/down,/data
         state.xdispmin = min([xselect,state.xcoord])
         state.xdispmax = max([xselect,state.xcoord])
      end
   2: begin  ; Center
         ; Get new x/y-range
         print,'Select second bound'
         cursor,xselect,yselect,/down,/data
         state.xdispmin = min([xselect,state.xcoord])
         state.xdispmax = max([xselect,state.xcoord])
         state.ydispmin = min([yselect,state.ycoord])
         state.ydispmax = max([yselect,state.ycoord])
      end
   4: begin  ; Right
         ; Get new y-range
         print,'Select second y-bound'
         cursor,xselect,yselect,/down,/data
         state.ydispmin = min([yselect,state.ycoord])
         state.ydispmax = max([yselect,state.ycoord])
      end
   else: print,'Unknown mouse button'
endcase

cf_draw_plot

end

;------------------------------------------------------------------------
; Keyboard events
;------------------------------------------------------------------------

Pro keyboard_event,event

common cf_state
common cf_data

; Convert coordinates from device to data
coords = convert_coord(event.x,event.y,$
                       /device,/to_data)
state.xcoord = coords(0)
state.ycoord = coords(1)

; Determine which key was pressed
if (event.ch ne 0) then begin
   case event.ch of
      8:    key = 'backspace'
      9:    key = 'tab'
      13:   key = 'return'
      27:   key = 'escape'
      32:   key = 'space'
      127:  key = 'delete'
      else: key = string(event.ch)
   endcase
endif else begin
   case event.key of
      5:    key = 'leftarrow'
      6:    key = 'rightarrow'
      7:    key = 'uparrow'
      8:    key = 'downarrow'
      9:    key = 'pageup'
      10:   key = 'pagedown'
      11:   key = 'home'
      12:   key = 'end'
      else: key = 'unknown'
   endcase
endelse

case key of
   ;;;
   ;;; Display keys
   ;;;
   'h': begin   ; Toggle histogram mode
           if (state.histogram eq 0) then state.histogram = 1 else $
              state.histogram = 0
           cf_draw_plot
        end
   ']': begin   ; Step up in wavelength
           if (state.xdispmax lt state.maxwave) then begin
              dwave = state.xdispmax - state.xdispmin
              state.xdispmax = state.xdispmax + 0.95*dwave
              state.xdispmin = state.xdispmax - dwave
              cf_draw_plot
           endif
        end
   '[': begin   ; Step down in wavelength
           if (state.xdispmin gt state.minwave) then begin
              dwave = state.xdispmax - state.xdispmin
              state.xdispmin = state.xdispmin - 0.95*dwave
              state.xdispmax = state.xdispmin + dwave
              cf_draw_plot
           endif
        end
   'p': begin   ; Pan to the cursor wavelength position
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 0.5*dwave
           state.xdispmax = state.xcoord + 0.5*dwave
           cf_draw_plot
        end
   'i': begin   ; Zoom in at the cursor position (and rescale in y)
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 0.25*dwave
           state.xdispmax = state.xcoord + 0.25*dwave
           display_ls = where(wave(*,state.order) ge state.xdispmin and $
                              wave(*,state.order) le state.xdispmax,n_display)
           if (n_display gt 0) then begin
              disp_flux_range = max(flux(display_ls,state.order)) - $
                                min(flux(display_ls,state.order))
              state.ydispmin = 0 < (min(flux(display_ls,state.order)) - $
                                    0.05*disp_flux_range)
              state.ydispmax = max(flux(display_ls,state.order)) + $
                                    0.05*disp_flux_range
           endif
           cf_draw_plot
        end
   'o': begin   ; Zoom out at the cursor position (and rescale in y)
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 1.0*dwave
           state.xdispmax = state.xcoord + 1.0*dwave
           display_ls = where(wave(*,state.order) ge state.xdispmin and $
                              wave(*,state.order) le state.xdispmax,n_display)
           if (n_display gt 0) then begin
              disp_flux_range = max(flux(display_ls,state.order)) - $
                                min(flux(display_ls,state.order))
              state.ydispmin = 0 < (min(flux(display_ls,state.order)) - $
                                    0.05*disp_flux_range)
              state.ydispmax = max(flux(display_ls,state.order)) + $
                                    0.05*disp_flux_range
           endif
           cf_draw_plot
        end
   'u': begin   ; Zoom out in y
           if (state.ydispmax gt 0) then state.ydispmax = 2*state.ydispmax
           if (state.ydispmin lt 0) then state.ydispmin = 2*state.ydispmin
           cf_draw_plot
        end
   'w': begin   ; Reset plot range
           state.xdispmin = state.default_xdispmin
           state.xdispmax = state.default_xdispmax
           if state.plotnorm then begin
              state.ydispmin = -0.2
              state.ydispmax =  1.4
           endif else begin
              state.ydispmin = state.default_ydispmin
              state.ydispmax = state.default_ydispmax
           endelse
           cf_draw_plot
        end
   '-': begin   ; Change to lower order
           if (state.order gt 0) then $
              change_order,state.order-1
        end
   '=': begin   ; Change to higher order
           if (state.order lt state.norders-1) then $
              change_order,state.order+1
        end
   'n': begin   ; Toggle plotting normalized arrays
           if (state.plotnorm eq 0) then begin
              state.plotnorm = 1 
              state.ydispmin = -0.2
              state.ydispmax =  1.4
           endif else begin
              state.plotnorm = 0
              display_ls = where(wave(*,state.order) ge state.xdispmin and $
                              wave(*,state.order) le state.xdispmax,n_display)
              if (n_display gt 0) then begin
                 disp_flux_range = max(flux(display_ls,state.order)) - $
                                   min(flux(display_ls,state.order))
                 state.ydispmin = 0 < (min(flux(display_ls,state.order)) - $
                                       0.05*disp_flux_range)
                 state.ydispmax = max(flux(display_ls,state.order)) + $
                                       0.05*disp_flux_range
              endif
           endelse
           cf_draw_plot
        end
   '1': cf_bin_arrays,1
   '2': cf_bin_arrays,2
   '3': cf_bin_arrays,3
   '4': cf_bin_arrays,4
   '5': cf_bin_arrays,5
   '6': cf_bin_arrays,6
   '7': cf_bin_arrays,7
   '8': cf_bin_arrays,8
   '9': cf_bin_arrays,9
   '0': cf_bin_arrays,10
   'tab': print,'Tab key pressed.  Be careful of highlighted buttons!'
   ;;; 
   ;;; Spline fitting keys
   ;;;
   'a': begin   ; Add a spline point at the cursor position
           if not(state.plotnorm) then begin
              spline_x = [spline_x,state.xcoord]
              spline_y = [spline_y,state.ycoord]
              arrange  = sort(spline_x)
              spline_x = spline_x(arrange)
              spline_y = spline_y(arrange)
              compute_spline
              cf_draw_plot
           endif
        end
   'r': begin   ; Move nearest spline point to cursor position
           if not(state.plotnorm) then begin
              nearest = where(abs(spline_x-state.xcoord) eq $
                              min(abs(spline_x-state.xcoord)))
              spline_x(nearest) = state.xcoord
              spline_y(nearest) = state.ycoord
              compute_spline
              cf_draw_plot
           endif
        end
   'd': begin   ; Delete the nearest spline point
           if not(state.plotnorm) then begin
              nearest = where(abs(spline_x-state.xcoord) eq $
                              min(abs(spline_x-state.xcoord)),$
                              complement=other_ls,ncomplement=n_other)
              if (n_other lt 3) then begin
                 print,'Must have at least 3 spline points'
              endif else begin
                 spline_x = spline_x(other_ls)
                 spline_y = spline_y(other_ls)
                 compute_spline
                 cf_draw_plot
              endelse
           endif
        end
   't': begin   ; Set the nearest spline point to one
           if not(state.plotnorm) then begin
              nearest = where(abs(spline_x-state.xcoord) eq $
                              min(abs(spline_x-state.xcoord)))
              spline_y(nearest) = 1.0
              compute_spline
              cf_draw_plot
           endif
        end
   'b': begin  ; Set all points blueward equal to one
           if not(state.plotnorm) then begin
              reply = 'n'
              read,'Set all points blueward of '+strtrim(state.xcoord,2)+$
                   ' to 1.0? (y/n): ',reply
              if (reply eq 'y') then begin
                 blue_ls = where(spline_x le state.xcoord,n_blue,compl=red_ls)
                 if (n_blue ne 0) then spline_y(blue_ls) = 1.0
              endif
              compute_spline
              cf_draw_plot
           endif
        end
   'v': begin  ; Set all point blueward equal to a user-specified value
           if not(state.plotnorm) then begin
              reply = -1.0
              read,'Set all points blueward of '+strtrim(state.xcoord,2)+$
                   ' to this value: ',reply
              reply = float(reply)
              if (reply ne -1.0) then begin
                 blue_ls = where(spline_x le state.xcoord,n_blue,compl=red_ls)
                 if (n_blue ne 0) then spline_y(blue_ls) = reply
              endif
              compute_spline
              cf_draw_plot
           endif
        end

   ;;;
   ;;; Other keys
   ;;;
   '?': begin
           print,'Display Commands'
           print,'     left click : Set x range'
           print,'    right click : Set y range'
           print,'   center click : Set x and y ranges'
           print,'              [ : Pan left'
           print,'              ] : Pan right'
           print,'              p : Pan to the cursor position'
           print,'              i : Zoom in'
           print,'              o : Zoom out'
           print,'              u : Zoom out in y'
           print,'              w : Display whole order'
           print,'              - : Change order up'
           print,'              = : Change order down'
           print,'              n : Plot normalized spectrum (toggle)'
           print,'            0-9 : Bin spectrum'
           print,'              h : Plot in histogram mode (toggle)'
           print,'Fitting Commands'
           print,'              a : Add spline point'
           print,'              r : Relocate nearest spline point'
           print,'              d : Delete nearest spline point'
           print,'              t : Set the nearest spline point to 1'
           print,'              b : Set all points blueward to 1'
           print,'              v : Set all points blueward to new value'
        end
   else: print,'No command for '+key+' key.  Press ? for menu.'
endcase

end

;------------------------------------------------------------------------
; Resize window events
;------------------------------------------------------------------------

Pro resize_cont_fit_window,event

common cf_state

; Resize plotting window
widget_control,state.plotwindow_id,$
               draw_xsize=((event.x - 6) > state.buttonbase_xsz),$
               draw_ysize=((event.y - state.buttonbase_ysz - 9) > 1)

; Redraw plot
cf_draw_plot

end

;------------------------------------------------------------------------
; Initialize widgets
;------------------------------------------------------------------------

Pro create_cont_fit_window,group=group

common cf_state

; Top level base (for resizing)

state.base_id = widget_base(title='Continuum Fit',tlb_size_events=1)

; Create base to actually hold all the widgets and direct events
; to the correct proceedure.

state.subbase_id = widget_base(state.base_id,/column,/base_align_right,$
                               event_pro='cont_fit_window_event')

; Adjust plot windows to fit screen, if necessary
device,get_screen_size=screen
state.plot_xsz = state.plot_xsz  < (screen(0)-100)
state.plot_ysz = state.plot_ysz  < (screen(1)-100)

state.plotwindow_id = widget_draw(state.subbase_id,$
                                  scr_xsize=state.plot_xsz,$
                                  scr_ysize=state.plot_ysz,$
                                  uvalue='plotwindow',$
                                  keyboard_events=1,$
                                  /button_events)

state.buttonbase_id = widget_base(state.subbase_id,/row)

state.write_button_id = widget_button(state.buttonbase_id,value='Write Files',$
                                      uvalue='write')

state.quit_button_id = widget_button(state.buttonbase_id,value='Quit',$
                                     uvalue='quit')

; Get size of button row
buttonbase_geom = widget_info(state.buttonbase_id,/geometry)
state.buttonbase_xsz = buttonbase_geom.xsize
state.buttonbase_ysz = buttonbase_geom.ysize

; Realize widgets
widget_control,state.base_id,/realize

; Determine window numbers
widget_control,state.plotwindow_id, get_value = tmp_value
state.plotwindow = tmp_value

end

;------------------------------------------------------------------------
; Compute spline
;------------------------------------------------------------------------

Pro compute_spline

common cf_state
common cf_data

tension = 2.0

fit_ls = where(wave(*,state.order) gt 0,n_fit,comp=nonfit_ls,ncomp=n_nonfit)
splfit(fit_ls,state.order) = spline(spline_x,spline_y,$
                                    wave(fit_ls,state.order),tension)
if (n_nonfit ne 0) then splfit(nonfit_ls) = 0.

binfit_ls = where(binwave(*,state.order) gt 0,n_binfit,$
                  comp=binnonfit_ls,ncomp=n_binnonfit)
temp_binsplfit = 0.*binwave(*,state.order)
temp_binsplfit(binfit_ls) = spline(spline_x,spline_y,$
                                   binwave(binfit_ls,state.order),tension)

; Normalized arrays

norm_flux(fit_ls,state.order)  = flux(fit_ls,state.order) / $
                                    splfit(fit_ls,state.order) 
norm_error(fit_ls,state.order) = error(fit_ls,state.order) / $
                                    splfit(fit_ls,state.order)
if (n_nonfit ne 0) then begin
   norm_flux(nonfit_ls,state.order)  = 1.
   norm_error(nonfit_ls,state.order) = 1.
endif

norm_binflux(binfit_ls,state.order)  = binflux(binfit_ls,state.order) / $
                                          temp_binsplfit(binfit_ls)
norm_binerror(binfit_ls,state.order) = binerror(binfit_ls,state.order) / $
                                          temp_binsplfit(binfit_ls)
if (n_binnonfit ne 0) then begin
   norm_binflux(binnonfit_ls,state.order)  = 1.
   norm_binerror(binnonfit_ls,state.order) = 1.
endif 

end

;------------------------------------------------------------------------
; Draw plot
;------------------------------------------------------------------------

Pro cf_draw_plot

common cf_state
common cf_data

wset,state.plotwindow

fit_thick = 2.0

; Define circle plot symbol
plotsym,0,thick=fit_thick,/fill

; Histogram mode?
if (state.histogram eq 1) then  data_psym = 10 else data_psym = 0

; Plot only pixels with positive wavelengths
plot_ls    = where(wave(*,state.order) gt 0)
binplot_ls = where(binwave(*,state.order) gt 0)

if state.plotnorm then begin
   ; Plot normalized arrays
   ; Zero line
   plot,[-1e9,1e9],[0,0],linestyle=1,$
                  xrange=[state.xdispmin,state.xdispmax],/xs,$
                  yrange=[state.ydispmin,state.ydispmax],/ys,$
                  title='Row '+strtrim(state.order,2)+' (Normalized)'
   ; Unity line
   oplot,[-1e9,1e9],[1,1],linestyle=1
   ; Flux
   oplot,binwave(binplot_ls,state.order),norm_binflux(binplot_ls,state.order),$
         psym=data_psym
endif else begin
   ; Plot original arrays with spline fit
   ; Zero line
   plot,[-1e9,1e9],[0,0],linestyle=1,$
                  xrange=[state.xdispmin,state.xdispmax],/xs,$
                  yrange=[state.ydispmin,state.ydispmax],/ys,$
                  title='Row '+strtrim(state.order,2)
   ; Flux
   oplot,binwave(binplot_ls,state.order),binflux(binplot_ls,state.order),$
         psym=data_psym
   ; Spline fit
   oplot,wave(plot_ls,state.order),splfit(plot_ls,state.order),color=250,$
         thick=fit_thick
   ; Spline points
   oplot,spline_x,spline_y,psym=8,color=250,symsize=1.5
endelse

end

;------------------------------------------------------------------------
; Change order
;------------------------------------------------------------------------

Pro change_order,new_order,no_save=no_save

common cf_state
common cf_data

old_order   = state.order
state.order = new_order

; Save spline points for old order
if not(keyword_set(no_save)) then begin
   spoints(*,old_order) = -9999
   svals(*,old_order)   = -9999
   n_old_spoints = n_elements(spline_x)
   spoints(0:n_old_spoints-1,old_order) = spline_x
   svals(0:n_old_spoints-1,old_order)   = spline_y
endif

; Activate spline points for the new order
active_ls   = where(spoints(*,state.order) ne -9999)
spline_x    = spoints(active_ls,state.order)
spline_y    = svals(active_ls,state.order)

; Set default plot limits for the new order
pos_ls        = where(wave(*,state.order) gt 0)
state.minwave = min(wave(pos_ls,state.order))
state.maxwave = max(wave(pos_ls,state.order))
wave_range = state.maxwave - state.minwave
state.default_xdispmin = min(wave(pos_ls,state.order)) - 0.02*wave_range
state.default_xdispmax = max(wave(pos_ls,state.order)) + 0.02*wave_range
; Ignore outlying 1% of flux values when setting display limits
fluxorder  = sort(flux(pos_ls,state.order))
fluxlist   = flux(fluxorder,state.order)
npix       = n_elements(fluxlist)
flux_range = fluxlist(0.99*npix) - fluxlist(0.01*npix)
state.default_ydispmin = 0 < (fluxlist(0.01*npix) - 0.05*flux_range)
state.default_ydispmax = fluxlist(0.99*npix) + 0.05*flux_range

; Set plot limits to the defaults, unless plotting normalized arrays
state.xdispmin = state.default_xdispmin
state.xdispmax = state.default_xdispmax
if state.plotnorm then begin
   state.ydispmin = -0.2
   state.ydispmax =  1.4
endif else begin
   state.ydispmin = state.default_ydispmin
   state.ydispmax = state.default_ydispmax
endelse

cf_draw_plot

end

;------------------------------------------------------------------------
; Bin arrays
;------------------------------------------------------------------------

Pro cf_bin_arrays,binsz

common cf_state
common cf_data

binsz = long(binsz)

state.binsz = binsz

n_bins = floor(1.*state.npix / binsz)

if (binsz eq 1) then begin
   binflux  = flux
   binerror = error
   binwave  = wave
endif else begin
   binflux  = fltarr(n_bins,state.norders)
   binerror = fltarr(n_bins,state.norders)
   binwave  = dblarr(n_bins,state.norders)
   for i=0L,long(n_bins)-1 do binflux(i,*) = $
                           total(flux(i*binsz:(i+1)*binsz-1,*),1)
   for i=0L,long(n_bins)-1 do binerror(i,*) = $
                           total(error(i*binsz:(i+1)*binsz-1,*),1)
   for i=0L,long(n_bins)-1 do binwave(i,*) = $
                           total(wave(i*binsz:(i+1)*binsz-1,*),1)
   binflux  = binflux / binsz
   binerror = binerror / binsz
   binwave  = binwave / binsz
endelse

; Recompute binned normalized arrays
norm_binflux = 0.*binflux
norm_binerror = 0.*binerror
compute_spline

cf_draw_plot

end

;------------------------------------------------------------------------
; Write files
;------------------------------------------------------------------------

Pro write_cf_data

common cf_state
common cf_data

; Save spline points for the current order
spoints(*,state.order) = -9999
svals(*,state.order)   = -9999
n_old_spoints = n_elements(spline_x)
spoints(0:n_old_spoints-1,state.order) = spline_x
svals(0:n_old_spoints-1,state.order)   = spline_y

; Allow user to change filenames

new_cont  = ''
new_param = ''
new_norm_flux = ''
new_norm_error = ''
new_response = ''

read,new_cont,prompt='Continuum FITS file ['+file.cont+']:'
if (new_cont ne '') then file.cont = new_cont

read,new_param,prompt='Continuum param file ['+file.param+']:'
if (new_param ne '') then file.param = new_param

read,new_norm_flux,prompt=$
     'Normalized flux FITS file ['+file.norm_flux+']:'
if (new_norm_flux ne '') then file.norm_flux = new_norm_flux

read,new_norm_error,prompt=$
     'Normalized error  FITS file ['+file.norm_error+']:'
if (new_norm_error ne '') then file.norm_error = new_norm_error

if state.writeresponse then begin
   read,new_response,prompt=$
        'Response FITS file ['+file.response+']:'
   if (new_response ne '') then file.response = new_response
   ; Calculate response function for each order
   resp_fn = 0d0*norm_flux
   for i=0,state.norders-1 do begin
         pos_ls = where(splfit(*,i) gt 0,n_pos,comp=neg_ls,ncomp=n_neg)
         resp_fn(pos_ls,i) = interpol(fcalib,wcalib,wave(pos_ls,i)) / $
                                splfit(pos_ls,i)
         ; Suppress regions with very little transmission
         max_splfit = max(splfit(*,i))
         max_pix    = where(splfit(*,i) eq max_splfit)
         max_pix    = max_pix(0)
         min_resp   = resp_fn(max_pix,i)
         low_ls = where(splfit(*,i) le max_splfit/(1e3),n_low)
         if (n_low ne 0) then resp_fn(low_ls,i) = 1e3 * min_resp
         if (n_neg ne 0) then resp_fn(neg_ls,i) = 1e3 * min_resp
   endfor
   zero_ls = where(wave le 0,n_zero)
   if (n_zero ne 0) then resp_fn(zero_ls) = 0.
   ; Avoid very small numbers
   resp_fn = resp_fn * 1d30
endif

; If input was in double precision, then output in double precision.
; Otherwise, output in floating type.
if (input_type eq 5) then begin
   writefits,file.cont,double(splfit),fluxhdr
   writefits,file.norm_flux,double(norm_flux),fluxhdr
   writefits,file.norm_error,double(norm_error),errorhdr
   openw,lun,file.param,/get_lun
   for i=0,state.norders-1 do begin
      active_ls = where(spoints(*,i) ne -9999)
      output_x  = spoints(active_ls,i)
      output_y  = svals(active_ls,i)
      n_output  = n_elements(output_x)
      for j=0,n_output-1 do printf,lun,i,output_x(j),double(output_y(j))
   endfor 
   free_lun,lun
   if state.writeresponse then $
      writefits,file.response,double(resp_fn),fluxhdr
endif else begin
   writefits,file.cont,float(splfit),fluxhdr
   writefits,file.norm_flux,float(norm_flux),fluxhdr
   writefits,file.norm_error,float(norm_error),errorhdr
   openw,lun,file.param,/get_lun
   for i=0,state.norders-1 do begin
      active_ls = where(spoints(*,i) ne -9999)
      output_x  = spoints(active_ls,i)
      output_y  = svals(active_ls,i)
      n_output  = n_elements(output_x)
      for j=0,n_output-1 do printf,lun,i,output_x(j),float(output_y(j))
   endfor 
   free_lun,lun
   if state.writeresponse then $
      writefits,file.response,float(resp_fn),fluxhdr
endelse
print,'Output files written.'

end

;------------------------------------------------------------------------
; Main program - see beginning of file for header
;------------------------------------------------------------------------

Pro cont_fit,fluxfile,errorfile,nspline=nspline,parfile=parfile,$
             smooth=smooth,wavefile=wavefile,polyfit=polyfit,$
             ginterp=ginterp,calibrated=calibrated,wcalibrated=wcalibrated,$
             wcal2vac=wcal2vac

common cf_state
common cf_data

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:'
   print,'   cont_fit,fluxfile,errorfile,nspline=nspline,parfile=parfile,'
   print,'            smooth=smooth,wavefile=wavefile,/polyfit,/ginterp,'
   print,'            calibrated=calibrated,wcalibrated=wcalibrated,'
   print,'            /wcal2vac'
   return
endif

; Initialize common blocks and settings 
cont_fit_initialize

; Read in data arrays
fluxexists = findfile(fluxfile,count=nfluxfiles)
if (nfluxfiles eq 0) then begin
   print,'Flux file not found.'
   return
endif else begin
   if keyword_set(wavefile) then begin
      ; Get wavelength information from separate file
      flux = readfits(fluxfile,fluxhdr)
      waveexists = findfile(wavefile,count=nwavefiles)
      if (nwavefiles eq 0) then begin
         print,'Wavelength file not found.'
         return
      endif else begin
         wave = readfits(wavefile,wavehdr)
      endelse
   endif else begin
      ; Get wavelength information from flux header
      flux = rflam(fluxfile,fluxhdr,wave)
   endelse
endelse
; If an error file has been provided, read in the error array.
; Otherwise, use a fake error array.
if (n_params() ge 2) then begin
   errorexists = findfile(errorfile,count=nerrorfiles)
   if (nerrorfiles eq 0) then begin
      print,'Error file not found.'
      return
   endif else begin
      error = readfits(errorfile,errorhdr)
   endelse
endif else begin
   errorfile = 'junk_error.fits'
   errorhdr  = fluxhdr
   error     = 1 + 0.*flux
endelse

; Calibrated continuum values for this object (optional)
if keyword_set(calibrated) then begin
   if keyword_set(wcalibrated) then begin
      ; Separated wavelength file
      fcalib = readfits(calibrated,hcalib)
      wcalib = readfits(wcalibrated)
   endif else begin
      ; Wavelength info in header
      fcalib = rflam(calibrated,hcalib,wcalib)
   endelse
   state.writeresponse = 1
   ; Convert wavelengths to vacuum?
   if keyword_set(wcal2vac) then airtovac,wcalib
endif else begin
   fcalib = 0 & hcalib = 0 & wcalib = 0
   state.writeresponse = 0
endelse

; Use double-precision for fitting
input_type = size(flux,/type)
flux       = double(flux)
error      = double(error)

; Starting number of spline points
if keyword_set(nspline) then n_spoints = nspline else n_spoints = 10.

; Number of pixels around each spline point over which to compute
; starting spline values
if keyword_set(smooth) then smth = smooth else smth = 21.

; Determine number of orders (rows)
sz = size(flux)
if (sz(0) eq 1) then begin
   state.norders = 1
   state.npix    = sz(1)
endif else begin
   state.norders = sz(2)
   state.npix    = sz(1)
endelse

; Set the initial binned arrays to be the input arrays
binflux  = flux
binerror = error
binwave  = wave

; Initialze file names
fluxroot  = filename(fluxfile,dirname=fluxdir,/noextension)
errorroot = filename(errorfile,dirname=errordir,/noextension)
file.flux       = fluxfile
file.error      = errorfile
file.norm_flux  = fluxdir+'Norm-'+fluxroot+'.fits'
file.norm_error = errordir+'Norm-'+errorroot+'.fits'
file.cont       = fluxdir+'Cont-'+fluxroot+'.fits'
file.param      = fluxdir+'Cont-'+fluxroot+'.par'
file.response   = fluxdir+'Resp-'+fluxroot+'.fits'

; Initialize binned and normalized arrays
binflux       = flux
binerror      = error
norm_flux     = 0.*flux
norm_error    = 0.*error
norm_binflux  = 0.*flux
norm_binerror = 0.*error

; Initialize spline fits
spoints = dblarr(state.max_spoints,state.norders) - 9999
svals   = dblarr(state.max_spoints,state.norders) - 9999
splfit  = 0.*flux
nside   = (smth-1)/2.
weight  = 1./(error^2)

if keyword_set(parfile) then begin
   ; Read in spline points from file
   stempl = {   $
              version: 1.0, $    
              datastart: 0L, $   
              delimiter: ' ', $
              missingvalue: -99.0, $
              commentsymbol: '#', $
              fieldcount: 3L, $
              fieldtypes: [3L,4L,5L], $
              fieldnames: ['order','wavelength','value'], $
              fieldlocations: [7L,15L,31L], $
              fieldgroups: [0L,1L,2L] $
            }
   sdata = read_ascii(parfile,template=stempl)   
   for i=0,state.norders - 1 do begin
      state.order  = i
      spoints_ls   = where(sdata.order eq i,n_spoints)
      pos_ls       = where(wave(*,i) gt 0 and finite(flux(*,i)) eq 1 and $
                           finite(weight(*,i)) eq 1,order_npix)
      order_wave   = wave(pos_ls,i)
      order_flux   = flux(pos_ls,i)
      order_weight = weight(pos_ls,i)
      if (n_spoints ne 0) then begin
         spline_x = sdata.wavelength(spoints_ls)
         if (smth eq -1) then begin
            print,'Using literal spline points from input file'
            spline_y = sdata.value(spoints_ls)
         endif else begin
            spline_y = fltarr(n_spoints)
            for j=0,n_spoints-1 do begin
               index = where(abs(order_wave-spline_x(j)) eq $
                             min(abs(order_wave-spline_x(j))))
               lo = (index-nside) > 0
               hi = (index+nside) < (order_npix-1)
               temp_flux   = order_flux(lo:hi)
               temp_weight = order_weight(lo:hi)
               spline_y(j) = total(temp_flux*temp_weight) / $
                             total(temp_weight)
            endfor   
         endelse
         ; Evaluate spline
         compute_spline
         ; Store spline points
         spoints(0:n_spoints-1,i) = spline_x
         svals(0:n_spoints-1,i)   = spline_y
      endif else begin
         print,'Missing spline points for order '+strtrim(i,2)
         return
      endelse
   endfor
endif else begin
   ; Compute new spline fit for each order.  Ignore zero wavelength.
   for i=0,state.norders - 1 do begin
      state.order = i
      notzero = where(abs(wave(*,i)) ge 1e-4)
      minwave = min(wave(notzero,i))
      maxwave = max(wave(notzero,i))
      ; Space break points evenly in wavelength
      bkpoint_space = (maxwave - minwave) / n_spoints
      spline_x = minwave + bkpoint_space*(0.5+dindgen(n_spoints))
      ; Compute the flux value at each break point.
      spline_y = fltarr(n_spoints)
      method = 'WEIGHTEDMEAN'
      if keyword_set(polyfit) then method = 'POLYFIT'
      if keyword_set(ginterp) then method = 'GINTERP'
      case method of 
         'POLYFIT': begin
            ; Local polynomial fit
            for j=0,n_spoints-1 do begin
               ls = where(abs(wave(*,i)-spline_x(j)) le bkpoint_space and $
                          finite(flux) eq 1)
               tmp_wave = wave(ls,i)
               tmp_flux = flux(ls,i)
               pfit     = robust_poly_fit(tmp_wave,tmp_flux,1,fluxfit)
               sdev     = robust_sigma(tmp_flux-fluxfit)
               ; Rough absorption line rejection
               keep_ls  = where(tmp_flux-fluxfit ge -2*sdev)
               pfit2    = robust_poly_fit(tmp_wave(keep_ls),tmp_flux(keep_ls),$
                                          1,fluxfit2)
               spline_y(j) = pfit2(0) + spline_x(j)*pfit2(1)
            endfor
            end
         'WEIGHTEDMEAN': begin
            ; Simple weighed mean (default)
            for j=0,n_spoints-1 do begin
               index = where(abs(wave(*,i)-spline_x(j)) eq $
                             min(abs(wave(*,i)-spline_x(j))))
               index = mean(index)
               lo = (index-nside) > 0
               hi = (index+nside) < (state.npix-1)
               spline_y(j) = total(flux(lo:hi,i)*weight(lo:hi,i)) / $
                             total(weight(lo:hi,i))
            endfor
            end
         'GINTERP': begin
            ; Gaussian interpolation scheme
            print,'Performing Gaussian interpolation...'
            f = flux(*,i)
            w = wave(*,i)
            e = error(*,i)
            pos_ls = where(wave(*,i) gt 0)
            f = f(pos_ls)
            w = w(pos_ls)
            e = e(pos_ls)
            if (n_elements(ginterp) eq 1) then begin
               ; Use default fitting values
               wmin   = min(w)
               wmax   = max(w)
               degree = 5
               dv     = 1000
            endif else begin
               wmin   = ginterp(0) > min(w)
               wmax   = ginterp(1) < max(w)
               degree = ginterp(2)
               dv     = ginterp(3)
            endelse
            ; Mask points with error <= 0
            neg_ls = where(e le 0,n_neg)
            if (n_neg ne 0) then e(neg_ls) = 1e6
            ; First pass - median
            med = median(f,100)          ; 300 chosen for MIKE data
            ; Scale errors               ; 100 for HIRES
            scaled_diff = (f-med)/e
            ls     = where(w ge wmin and w le wmax and abs(scaled_diff) le 20 $
                           and scaled_diff ne 0)
            plothist,(scaled_diff)(ls),bin=0.1,/half,x,y,/noplot
            gfit   = gaussfit(x,y,gterms,nterms=3)
            escale = e*gterms(2)
            ; Filter out noise spikes and absorption lines
            bad_ls = where(abs(f-med) gt 3*escale,n_bad)
            temp_e = escale
            if (n_bad ne 0) then temp_e(bad_ls) = 1e6
            ; Dispersion in km/s
            c    = 2.998e5
            vmed = c * median(convol(w(ls),[-1,1])/w(ls))
            v    = vmed*findgen(n_elements(f))
            spline_v = interpol(v,w,spline_x)
            ; First-pass Gaussian interpolation
            for j=0,n_spoints-1 do $
               spline_y(j) = gauss_interp(v,f,temp_e,degree,dv,x0=spline_v(j))
            compute_spline
            ; Second pass filter
            scaled_diff = (f-splfit(pos_ls,i))/e
            ls     = where(w ge wmin and w le wmax and abs(scaled_diff) le 20 $
                           and scaled_diff ne 0)
            plothist,(scaled_diff)(ls),bin=0.1,/half,x,y,/noplot
            gfit   = gaussfit(x,y,gterms,nterms=3)
            escale = e*gterms(2)
            bad_ls = where(abs(f-splfit(pos_ls,i)) gt 2*escale,n_bad)
            temp_e = escale
            if (n_bad ne 0) then temp_e(bad_ls) = 1e6
            ; Second pass Gaussian interpolation
            for j=0,n_spoints-1 do $
               spline_y(j) = gauss_interp(v,f,temp_e,degree,dv,x0=spline_v(j))
            end
      endcase

      ; Reject bad spline points
      good_ls = where(finite(spline_y) eq 1,n_good)
      if (n_good lt 3) then begin
         print,'Fewer than 3 good spline points in order '+strtrim(i,2)
         spline_y(*) = interpol(flux(*,i),wave(*,i),spline_x)
      endif else begin
         spline_x  = spline_x(good_ls)
         spline_y  = spline_y(good_ls)
         n_spoints = n_good
      endelse
      ; Evaluate spline
      compute_spline
      ; Store spline points
      spoints(0:n_spoints-1,i) = spline_x
      svals(0:n_spoints-1,i)   = spline_y
   endfor
endelse

; Use rainbow color indices
device,decomp=0
loadct,39

; One plot per window
!p.multi = [0,1,1]

; Create a new window if one does not already exist
if not(xregistered('cont_fit_window')) then create_cont_fit_window

; Start with first order
change_order,0,/no_save

; Print help reminder
print,"Type '?' for menu."

; Register widgets with blocking
xmanager,'cont_fit_window',state.base_id,group_leader=group,$
         event_handler='resize_cont_fit_window'

return
end
