;+---------------------------------------------------------------------
;
; SPECPLOT       3/2007
;
; Interactive tool for plotting an analyzing single- and multiple-order
; spectra.  Modeled in large part after XPLOT.
;
; CALLING SEQUENCE
;     specplot,fluxfile,errorfile,wavefile=wavefile,/norm,z=z,$
;              f2=f2,w2=w2,yrange=yrange,/block
;
; INPUTS
;        fluxfile    - String, flux filename, may be single order
;                      or multiple orders, OR a 1D array of flux values.
;        errorfile   - String, error filename, same size as fluxfile,
;                      OR a 1D array of error values.  If not
;                      provided, a fake error spectrum with uniform
;                      value will be used.
;        
; KEYWORDS
;        wavefile    - String, wavelength filename, same size as
;                      fluxfile.  Default is to compute wavelengths
;                      from information in the flux file header, 
;                      although this is not currently set up for
;                      multiple orders with MAKEE-style wavvelength 
;                      information.
;        norm        - If set, assumes input spectrum is normalized.
;        z           - Scalar, redshift of the input spectrum
;        f2          - String, flux filename, may be single order
;                      or multiple orders, OR a 1D array of flux values
;                      of second spectrum to overplot.
;        w2          - String, wavelength filename, OR a 1D array of
;                      wavelengths for f2.  If not supplied, and there
;                      is no wavelengths information in the f2 header,
;                      then the wavelengths for the primary flux array
;                      are used.
;        yrange      - 2-element array specifying the [ymin,ymax] values
;                      to use when plotting.  Default is to
;                      automatically determine the range.
;        block       - If set, widget will open with blocking.  Default
;                      is to allow other IDL processes while specplot
;                      is running.
;
; OUTPUTS
;        <none>
;
; HISTORY
;        Adapted from CONT_FIT 3/12/2007 GDB
;----------------------------------------------------------------------

;----------------------------------------------------------------------
; Initialize common blocks and other settings
;----------------------------------------------------------------------

Pro specplot_initialize

common sp_state, state
common sp_data, wave, flux, error, $
                file, fluxhdr, errorhdr, binflux, binerror, $
                binwave, input_type, marked_lines, $
                flux2, wave2, flux2hdr, fluxfit
common sp_atom, rest_wave, fvalue, atom, lya, lyb, c

state = {              $
          plotwindow: 0L, $        ; Plot window number
          plot_xsz: 1700L, $       ; Screen x-size of plot window
          plot_ysz: 730L, $        ; Screen y-size of plot window
          xdispmin: 0d0, $         ; Min display x value 
          xdispmax: 0d0, $         ; Max display x value
          ydispmin: 0d0, $         ; Min display y value 
          ydispmax: 0d0, $         ; Max display y value
          default_xdispmin: 0d0, $ ; Default min display x value 
          default_xdispmax: 0d0, $ ; Default max display x value
          default_ydispmin: 0d0, $ ; Default min display y value 
          default_ydispmax: 0d0, $ ; Default max display y value
          fixedyrange: 0L, $       ; Flag to use fixed y range
          fixed_ydispmin: 0d0, $   ; Fixed min display y value
          fixed_ydispmax: 0d0, $   ; Fixed max display y value
          norm: 0L, $              ; Flag to indicate normalized array
          scale_factor: 1.0, $     ; Flux scaling factor for plotting
          plot_error: 0L, $        ; Flag to plot error array
          plot_flux2: 0L, $        ; Flag to plot second flux array
          plot_fluxfit: 0L, $      ; Flag to plot fit to flux array
          modified_arrays: 0L, $   ; Flag that arrays have been modified
          binsz: 1L, $             ; Bin size for displayed spectrum
          histogram: 0L, $         ; Flag to plot in histogram mode
          minwave: 0L, $           ; Minimum wavelength for the current order
          maxwave: 0L, $           ; Maximum wavelength for the current order
          xcoord: 0d0, $           ; Data x-coordinate of mouse click
          ycoord: 0d0, $           ; Data y-coordinate of mouse click
          order: 0L, $             ; Index of current order
          norders: 0L, $           ; Number of orders
          npix: 0L, $              ; Number of pixels in each order
          buttonbase_xsz: 0L, $    ; Screen x-size of button base
          buttonbase_ysz: 0L, $    ; Screen y-size of button base
          base_id: 0L, $           ; WIDGET IDS
          subbase_id: 0L, $        ;
          buttonbase_id: 0L, $     ;
          plotwindow_id: 0L, $     ; 
          quit_button_id: 0L, $    ;
          zqso: -1d0, $            ; QSO redshift
          zabs: 0d0, $             ; Absorption redshift
          first_mark_set: 0L, $    ; First mark set for masking 
          first_mark_wave: 0d0, $  ; Wavlength of first mark
          plot_line_list: 0L, $    ; Mark lines from list
          line_file: '' $          ; Name of line list file
        }

file =  {          $
          flux:  'sp_flux.fits', $     ; Default file names
          error: 'sp_error.fits', $   
          flux2: 'sp_flux2', $
          edt_flux: '', $
          edt_error: '', $
          edt_flux2: '' $
        }

atom =  {             $
          doublet_lam1: 0d0, $     ; Doublet wavelength 1
          doublet_lam2: 0d0, $     ; Doublet wavelength 2
          doublet_name: '', $      ; Doublet name
          metals_wmin: 0d0, $      ; Min wavelength for plotting metals
          metals_wmax: 0d0, $      ; Max wavelength for plotting metals
          metals_fmin: 0e0, $      ; Min f-value for plotting metals
          metals_dv: 200e0, $      ; +/- velocity range for plotting metals
          metals_ps: 0L, $         ; Create PS plot for metals lines
          ztol: 1e-4, $            ; Redshift tolerance for identifying lines
          zmax: 6.5 $              ; Maximum redshift for identifying lines
        }

; Physical constanzts

lya = 1215.6701d0   ; H I Lya
lyb = 1025.7223d0   ; H I Lyb
c   = 2.99792458d5  ; Speed of light (km/s)

; Rest wavelength and oscilator strength of measured line

rest_wave = 1548.202d0 ; C IV 1548
fvalue    = 0.189900d0 ;

end

;------------------------------------------------------------------------
; Widget events (main)
;------------------------------------------------------------------------

Pro specplot_window_event,event

common sp_state
common sp_data

widget_control,event.id,get_uvalue=eventval

case eventval of 
    'plotwindow' : begin
                      if event.clicks then begin
                         ; Mouse event: act when mouse button is pressed
                         if (event.press gt 0) then mouse_event,event
                      endif else begin
                         ; Keyboard event: act when key is pressed
                         if (event.press gt 0) then keyboard_event,event
                      endelse
                   end
    'quit'       : begin
                      ; Check whether a fit has been created
                      if (state.plot_fluxfit eq 1) then begin
                         save_fit = ''
                         read,save_fit,$
                            prompt = 'Save fit? [y]: '
                         if (save_fit eq ''  or $
                             save_fit eq 'y' or $
                             save_fit eq 'yes') then begin
                             writefits,'sp_fit.fits',fluxfit
                             print,'Fit written to sp_fit.fits'
                         endif
                      endif
                      ; Check whether arrays have been modified
                      if (state.modified_arrays eq 1) then begin
                         save_arrays = ''
                         read,save_arrays,$
                            prompt = 'Modified arrays exist. Save? [y]: '
                         if (save_arrays eq 'n' or $
                             save_arrays eq 'no') then begin
                            quit_check = ''
                            read,quit_check,$
                               prompt = 'Really quit without saving? [y]: '
                            if (quit_check eq '' or $
                                quit_check eq 'y' or $
                                quit_check eq 'yes') then begin
                               widget_control,event.top,/destroy
                            endif else begin
                               print,'Not quitting'
                            endelse
                         endif else begin
                            sp_save_arrays
                            widget_control,event.top,/destroy
                         endelse
                      endif else begin
                         widget_control,event.top,/destroy
                      endelse
                   end
    else         : print,'Unknown event'
endcase

end

;------------------------------------------------------------------------
; Mouse events
;------------------------------------------------------------------------

Pro mouse_event,event

common sp_state

; Convert coordinates from device to data
coords = convert_coord(event.x,event.y,$
                       /device,/to_data)
state.xcoord = coords(0)
state.ycoord = coords(1)

case event.press of
   1: begin  ; Left
         ; Get new x-range
         print,'Select second x-bound'
         cursor,xselect,yselect,/down,/data
         state.xdispmin = min([xselect,state.xcoord])
         state.xdispmax = max([xselect,state.xcoord])
      end
   2: begin  ; Center
         ; Get new x/y-range
         print,'Select second bound'
         cursor,xselect,yselect,/down,/data
         state.xdispmin = min([xselect,state.xcoord])
         state.xdispmax = max([xselect,state.xcoord])
         state.ydispmin = min([yselect,state.ycoord])
         state.ydispmax = max([yselect,state.ycoord])
      end
   4: begin  ; Right
         ; Get new y-range
         print,'Select second y-bound'
         cursor,xselect,yselect,/down,/data
         state.ydispmin = min([yselect,state.ycoord])
         state.ydispmax = max([yselect,state.ycoord])
      end
   else: print,'Unknown mouse button'
endcase

sp_draw_plot

end

;------------------------------------------------------------------------
; Keyboard events
;------------------------------------------------------------------------

Pro keyboard_event,event

common sp_state
common sp_data
common sp_atom

; Convert coordinates from device to data
coords = convert_coord(event.x,event.y,$
                       /device,/to_data)
state.xcoord = coords(0)
state.ycoord = coords(1)

; Determine which key was pressed
if (event.ch ne 0) then begin
   case event.ch of
      8:    key = 'backspace'
      9:    key = 'tab'
      13:   key = 'return'
      27:   key = 'escape'
      32:   key = 'space'
      127:  key = 'delete'
      else: key = string(event.ch)
   endcase
endif else begin
   case event.key of
      5:    key = 'leftarrow'
      6:    key = 'rightarrow'
      7:    key = 'uparrow'
      8:    key = 'downarrow'
      9:    key = 'pageup'
      10:   key = 'pagedown'
      11:   key = 'home'
      12:   key = 'end'
      else: key = 'unknown'
   endcase
endelse

case key of
   ;;;
   ;;; Display keys
   ;;;
   'h': begin   ; Toggle histogram mode
           if (state.histogram eq 0) then state.histogram = 1 else $
              state.histogram = 0
           sp_draw_plot
        end
   ']': begin   ; Step up in wavelength
           if (state.xdispmax lt state.maxwave) then begin
              dwave = state.xdispmax - state.xdispmin
              state.xdispmax = state.xdispmax + 0.95*dwave
              state.xdispmin = state.xdispmax - dwave
              sp_draw_plot
           endif
        end
   '[': begin   ; Step down in wavelength
           if (state.xdispmin gt state.minwave) then begin
              dwave = state.xdispmax - state.xdispmin
              state.xdispmin = state.xdispmin - 0.95*dwave
              state.xdispmax = state.xdispmin + dwave
              sp_draw_plot
           endif
        end
   'p': begin   ; Pan to the cursor wavelength position
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 0.5*dwave
           state.xdispmax = state.xcoord + 0.5*dwave
           sp_draw_plot
        end
   'i': begin   ; Zoom in at the cursor position (and rescale in y)
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 0.25*dwave
           state.xdispmax = state.xcoord + 0.25*dwave
           display_ls = where(wave(*,state.order) ge state.xdispmin and $
                              wave(*,state.order) le state.xdispmax,n_display)
           if (n_display gt 0) then begin
              disp_flux_range = max(flux(display_ls,state.order)) - $
                                min(flux(display_ls,state.order))
              state.ydispmin = 0 < (min(flux(display_ls,state.order)) - $
                                    0.05*disp_flux_range)
              state.ydispmax = max(flux(display_ls,state.order)) + $
                                    0.05*disp_flux_range
           endif
           sp_draw_plot
        end
   'o': begin   ; Zoom out at the cursor position (and rescale in y)
           dwave = state.xdispmax - state.xdispmin
           state.xdispmin = state.xcoord - 1.0*dwave
           state.xdispmax = state.xcoord + 1.0*dwave
           display_ls = where(wave(*,state.order) ge state.xdispmin and $
                              wave(*,state.order) le state.xdispmax,n_display)
           if (n_display gt 0) then begin
              disp_flux_range = max(flux(display_ls,state.order)) - $
                                min(flux(display_ls,state.order))
              state.ydispmin = 0 < (min(flux(display_ls,state.order)) - $
                                    0.05*disp_flux_range)
              state.ydispmax = max(flux(display_ls,state.order)) + $
                                    0.05*disp_flux_range
           endif
           sp_draw_plot
        end
   'u': begin   ; Zoom out in y
           if (state.ydispmax gt 0) then state.ydispmax = 2*state.ydispmax
           if (state.ydispmin lt 0) then state.ydispmin = 2*state.ydispmin
           sp_draw_plot
        end
   'r': begin   ; Refresh plot
           sp_draw_plot
        end
   'w': begin   ; Reset plot range
           state.xdispmin = state.default_xdispmin
           state.xdispmax = state.default_xdispmax
           state.ydispmin = state.default_ydispmin
           state.ydispmax = state.default_ydispmax
           sp_draw_plot
        end
   '-': begin   ; Change to lower order
           if (state.order gt 0) then $
              change_order,state.order-1
        end
   '=': begin   ; Change to higher order
           if (state.order lt state.norders-1) then $
              change_order,state.order+1
        end
   'j': begin   ; Jump to a different order
           new_order = ''
           read,new_order,$
              prompt = '   Jump to order number: '
           if (new_order ne '') then begin
              if (new_order gt state.norders-1 or $
                  new_order lt 0) then begin
                 print,'No such order'
              endif else begin
                 change_order,new_order
              endelse
           endif
        end
   '1': sp_bin_arrays,1
   '2': sp_bin_arrays,2
   '3': sp_bin_arrays,3
   '4': sp_bin_arrays,4
   '5': sp_bin_arrays,5
   '6': sp_bin_arrays,6
   '7': sp_bin_arrays,7
   '8': sp_bin_arrays,8
   '9': sp_bin_arrays,9
   '0': sp_bin_arrays,10
   'tab': print,'Tab key pressed.  Be careful of highlighted buttons!'

   ;;;
   ;;; Edit and save arrays
   ;;;
   'e': sp_edit_arrays
   's': sp_save_arrays
           
   ;;;
   ;;; Line identification
   ;;;
   'Q': begin
           atom.doublet_name = 'C IV'
           atom.doublet_lam1 = 1548.202d0
           atom.doublet_lam2 = 1550.774d0
           sp_mark_doublet
        end
   'W': begin
           atom.doublet_name = 'Mg II'
           atom.doublet_lam1 = 2796.352d0
           atom.doublet_lam2 = 2803.531d0
           sp_mark_doublet
        end
   'R': begin
           atom.doublet_name = 'Si IV'
           atom.doublet_lam1 = 1393.755d0 
           atom.doublet_lam2 = 1402.770d0
           sp_mark_doublet
        end
   'Z': begin  ; Other doublet
           oplot,[state.xcoord,state.xcoord],[-1e29,1e29],color=200,linestyle=1
           print,'Enter key'
           print,'   a = O VI   1031.927  1037.616'
           print,'   b = N V    1238.821  1242.804'
           print,'   c = Si IV  1393.755  1402.770'
           print,'   d = C IV   1548.202  1550.774'
           print,'   e = Al III 1854.716  1862.790'
           print,'   f = Mg II  2796.352  2803.531'
           print,'   g = Ca II  3934.777  3969.591'
           print,'   h = Na I   5891.583  5897.558'
           doublet_key=get_kbrd()
           case doublet_key of
              'a': begin
                      atom.doublet_name = 'O VI'
                      atom.doublet_lam1 = 1031.9265d0
                      atom.doublet_lam2 = 1037.6155d0
                      sp_mark_doublet
                   end
              'b': begin
                      atom.doublet_name = 'N V'
                      atom.doublet_lam1 = 1238.821d0
                      atom.doublet_lam2 = 1242.804d0
                      sp_mark_doublet
                   end
              'c': begin
                      atom.doublet_name = 'Si IV'
                      atom.doublet_lam1 = 1393.755d0
                      atom.doublet_lam2 = 1402.770d0
                      sp_mark_doublet
                   end
              'd': begin
                      atom.doublet_name = 'C IV'
                      atom.doublet_lam1 = 1548.202d0
                      atom.doublet_lam2 = 1550.774d0
                      sp_mark_doublet
                   end
              'e': begin
                      atom.doublet_name = 'Al III'
                      atom.doublet_lam1 = 1854.7164d0
                      atom.doublet_lam2 = 1862.7895d0
                      sp_mark_doublet
                   end
              'f': begin
                      atom.doublet_name = 'Mg II'
                      atom.doublet_lam1 = 2796.352d0
                      atom.doublet_lam2 = 2803.531d0
                      sp_mark_doublet
                   end
              'g': begin
                      atom.doublet_name = 'Ca II'
                      atom.doublet_lam1 = 3934.777d0
                      atom.doublet_lam2 = 3969.591d0
                      sp_mark_doublet
                   end
              'h': begin
                      atom.doublet_name = 'Na I'
                      atom.doublet_lam1 = 5891.5833d0
                      atom.doublet_lam2 = 5897.5581d0
                      sp_mark_doublet
                   end
              else: begin
                       print,'No doublet marked'
                    end
           endcase
        end
   'return': begin   ; Print cursor position
                print,'Cursor x,y = '+strtrim(state.xcoord,2)+', '+$
                      strtrim(state.ycoord,2)
             end
   'l': begin   ; Compute redshift of Lya at the cursor position
           oplot,[state.xcoord,state.xcoord],[state.ydispmin,state.ydispmax],$
                 linestyle=2,color=200
           z_lya = state.xcoord / lya - 1
           print,'Cursor x = '+strtrim(state.xcoord,2)
           print,'Lya redshift = '+strtrim(z_lya,2)           
           state.zabs = z_lya
        end
   'L': begin   ; Compute redshift of Lyb at the cursor position
           oplot,[state.xcoord,state.xcoord],[state.ydispmin,state.ydispmax],$
                 linestyle=2,color=200
           z_lya = state.xcoord / lyb - 1
           print,'Cursor x = '+strtrim(state.xcoord,2)
           print,'Lyb redshift = '+strtrim(z_lya,2)           
           state.zabs = z_lya
        end
   'z': begin   ; Plot absorption lines at a specified redshift
           new_zabs = ''
           read,new_zabs,$
              prompt = '   Absorption redshift ['+strtrim(state.zabs,2)+']: '
           if (new_zabs ne '') then state.zabs = new_zabs
           sp_plot_metals,state.zabs
        end
   'I': begin  ; Attempt to identify marked lines
           sp_id_lines
        end

   ;;;
   ;;; Other keys
   ;;;
   'm': begin ; Mark cursor position and add to line list
           print,'Marking line at x = '+strtrim(state.xcoord,2)
           oplot,[state.xcoord,state.xcoord],[state.ydispmin,state.ydispmax],$
                 linestyle=1,color=150
           if (total(marked_lines) eq -1) then begin
              marked_lines = [state.xcoord]
           endif else begin
              marked_lines = [marked_lines,state.xcoord]
           endelse
        end
   'M': begin ; Print wavelengths of marked lines
           ;if (total(marked_lines) ne -1) then begin
              n_marked_lines = n_elements(marked_lines)
              ;;Print in descedining order
              ;line_order = reverse(sort(marked_lines))
              ;for i=0,n_marked_lines-1 do $
              ;   print,marked_lines(line_order(i))
              ; Print in an easy-to-copy array format
              line_order = sort(marked_lines)
              line_print = '['
              for i=0,n_marked_lines-1 do begin
                 line_print = line_print+strtrim(marked_lines(line_order(i)),2)
                 if (i le n_marked_lines-2) then $
                    line_print = line_print+',' else $
                    line_print = line_print+']'
              endfor
              print,line_print
           ;endif else begin
           ;   print,'No lines marked'
           ;endelse
        end
   'C': begin ; Clear marked lines
           print,'Clearing marked lines'
           marked_lines = -1
           sp_draw_plot
        end
   'P': begin ; Toggle plotting line list
           if (state.plot_line_list eq 0) then begin
              state.plot_line_list = 1
              new_line_file = ''
              read,new_line_file,$
                 prompt='Line file ['+state.line_file+']:'
              if (new_line_file ne '') then begin
                 ; Does file exist
                 fileexists = findfile(new_line_file,count=nfile)
                 if (nfile eq 0) then begin
                    print,'Line list not found'
                    state.plot_line_list = 0
                 endif else begin
                    state.line_file = new_line_file
                 endelse
              endif
           endif else begin
              state.plot_line_list = 0
           endelse
           sp_draw_plot
        end
   'c': begin ; Compute centroid and equivalent width
           sp_measure_ew
        end
   'f': begin ; Fit Voigt profiles using VPFIT
           sp_vpfit
        end
   '*': begin ; Change flux scale factor
           display_scale_factor = strtrim(state.scale_factor,2)
           read,new_scale_factor,$
              prompt='Enter scale factor ['+display_scale_factor+']:'
           if (new_scale_factor ne '') then $
              state.scale_factor = new_scale_factor
           sp_draw_plot
        end
   ':': begin ; Get command-line command
           command = ''
           read,command,prompt='Enter command [? for help]: '
           case command of
              'vpin': begin  ; Generate VPFIT input file for auto Lya fitting
                       if (n_elements(marked_lines) gt 1) then begin
                          vpfilename = 'vpfit.in'
                          new_vpfilename = ''
                          read,new_vpfilename,prompt='Enter VPFIT input '+$
                               'filename ['+vpfilename+']: '
                          if (new_vpfilename ne '') then $
                             vpfilename = new_vpfilename
                          line_order = reverse(sort(marked_lines))
                          ;openw,unit,vpfilename,/get_lun
                          for i=0,n_elements(marked_lines)-2 do begin
                             entry = ' %% '+file.flux+'  1   '+$
                                  strtrim(marked_lines(line_order(i+1)),2)+$
                                  '  '+strtrim(marked_lines(line_order(i)),2)
                             print,entry
                             print,'  '
                          endfor
                          ;free_lun,unit
                       endif else begin
                          print,'Must mark two or more wavelengths'
                       endelse
                     end
              '?':  begin
                       print,'   vpin: Create VPFIT input file'
                    end
              else: print,'Command not found'
           endcase
        end
   'v': begin ; Set Flux2 array to zero over a region
           if (state.first_mark_set eq 0) then begin
              ; First mark
              state.first_mark_set  = 1
              state.first_mark_wave = state.xcoord
              print,"Enter 'v' again to mark region"
           endif else begin
              ; Second mark
              x_lo    = state.xcoord < state.first_mark_wave
              x_hi    = state.xcoord > state.first_mark_wave
              mark_ls = where(wave2 ge x_lo and wave2 le x_hi,n_mark)
              if (n_mark ne 0) then flux2(mark_ls) = 0.
              state.first_mark_set  = 0
              state.first_mark_wave = 0
              state.modified_arrays = 1
              sp_draw_plot
           endelse
        end
   '?': begin
           print,'Display Commands'
           print,'     left click : Set x range'
           print,'    right click : Set y range'
           print,'   center click : Set x and y ranges'
           print,'              [ : Pan left'
           print,'              ] : Pan right'
           print,'              p : Pan to the cursor position'
           print,'              i : Zoom in'
           print,'              o : Zoom out'
           print,'              u : Zoom out in y'
           print,'              r : Refresh plot'
           print,'              w : Display whole order'
           print,'              - : Change order up'
           print,'              = : Change order down'
           print,'              j : Jump to new order'
           print,'            0-9 : Bin spectrum'
           print,'              h : Plot in histogram mode (toggle)'
           print,'Line ID Commands'
           print,'              Q : C IV'
           print,'              W : Mg II'
           print,'              R : Si IV'
           print,'              Z : Other doublet'
           print,'              l : Display Ly-alpha redshift'
           print,'              L : Display Ly-beta redshift'
           print,'              z : Plot abs lines at a specified redshift'
           print,'              I : Attempt to identify marked lines'
           print,'Other Commands'
           print,'              m : Mark cursor position'
           print,'              M : Print wavelengths of marked positions'
           print,'              C : Clear marked positions'
           print,'              P : Plot lines from file'
           print,'              c : Measure centroid and EW'
           print,'              f : Fit Voigt profiles using VPFIT'
           print,'              : : Enter command-line command'
           print,'              * : Change scale factor'
           print,'              e : Edit arrays'
           print,'              s : Save arrays to FITS files'
        end
   else: print,'No command for '+key+' key.  Press ? for menu.'
endcase

end

;------------------------------------------------------------------------
; Resize window events
;------------------------------------------------------------------------

Pro resize_specplot_window,event

common sp_state

; Resize plotting window
widget_control,state.plotwindow_id,$
               draw_xsize=((event.x - 6) > state.buttonbase_xsz),$
               draw_ysize=((event.y - state.buttonbase_ysz - 9) > 1)

; Redraw plot
sp_draw_plot

end

;------------------------------------------------------------------------
; Initialize widgets
;------------------------------------------------------------------------

Pro create_specplot_window,group=group

common sp_state

; Top level base (for resizing)

state.base_id = widget_base(title='SpecPlot',tlb_size_events=1)

; Create base to actually hold all the widgets and direct events
; to the correct proceedure.

state.subbase_id = widget_base(state.base_id,/column,/base_align_right,$
                               event_pro='specplot_window_event')

; Adjust plot windows to fit screen, if necessary
device,get_screen_size=screen
state.plot_xsz = state.plot_xsz  < (screen(0)-100)
state.plot_ysz = state.plot_ysz  < (screen(1)-100)

state.plotwindow_id = widget_draw(state.subbase_id,$
                                  scr_xsize=state.plot_xsz,$
                                  scr_ysize=state.plot_ysz,$
                                  uvalue='plotwindow',$
                                  keyboard_events=1,$
                                  /button_events)

state.buttonbase_id = widget_base(state.subbase_id,/row)

state.quit_button_id = widget_button(state.buttonbase_id,value='Quit',$
                                     uvalue='quit')

; Get size of button row
buttonbase_geom = widget_info(state.buttonbase_id,/geometry)
state.buttonbase_xsz = buttonbase_geom.xsize
state.buttonbase_ysz = buttonbase_geom.ysize

; Realize widgets
widget_control,state.base_id,/realize

; Determine window numbers
widget_control,state.plotwindow_id, get_value = tmp_value
state.plotwindow = tmp_value

end

;------------------------------------------------------------------------
; Draw plot
;------------------------------------------------------------------------

Pro sp_draw_plot

common sp_state
common sp_data

wset,state.plotwindow

fit_thick = 2.0

; Load color table
loadct,39,/silent

; Define circle plot symbol
plotsym,0,thick=fit_thick,/fill

; Histogram mode?
if (state.histogram eq 1) then  data_psym = 10 else data_psym = 0

plot,[-1e9,1e9],[0,0],linestyle=1,$
               xrange=[state.xdispmin,state.xdispmax],/xs,$
               yrange=[state.ydispmin,state.ydispmax],/ys,$
               title='Row '+strtrim(state.order,2)

; Error (optional)
if state.plot_error then $
   oplot,binwave(*,state.order),binerror(*,state.order),psym=data_psym,color=250

; Flux
oplot,binwave(*,state.order),state.scale_factor*binflux(*,state.order),$
      psym=data_psym

; Second spectrum (optional)
if state.plot_flux2 then $
   oplot,wave2(*,state.order),flux2(*,state.order),psym=data_psym,color=150

; Fit to flux (optional)
if state.plot_fluxfit then $
   oplot,wave(*,state.order),fluxfit(*,state.order),color=80

; Marked lines
if (total(marked_lines) ne -1) then begin
   n_marked = n_elements(marked_lines)
   for i=0,n_marked-1 do $
      oplot,replicate(marked_lines(i),2),[-1e29,1e29],linestyle=1,color=150
endif

; Lines from file
if (state.plot_line_list eq 1) then begin
   ; Read in line information
   line_templ = {version: 1.0, $    
                 datastart: 0, $   
                 delimiter: 0B, $        ; Fields arranged in columns
                 missingvalue: -99.0, $
                 commentsymbol: '#', $
                 fieldcount: 5, $
                 fieldtypes: [5,5,7,5,4], $
                 fieldnames: ['wobs','z','ion','wrest','f'], $
                 fieldlocations: [6,16,27,34,45], $
                 fieldgroups: lindgen(5) }
   line_data = read_ascii(state.line_file,templ=line_templ)
   n_line    = n_elements(line_data.wobs)
   ; Mark lines
   line_wobs  = line_data.wobs
   line_z     = line_data.z
   line_ion   = line_data.ion
   line_wrest = line_data.wrest
   line_f     = line_data.f
   for i=0,n_line-1 do begin
      tmp_wobs = (1+line_z(i))*line_wrest(i)
      oplot,replicate(tmp_wobs,2),[state.ydispmin,state.ydispmax],$
            linestyle=1,color=90
      tmp_label = line_ion(i)+' '+strtrim(line_wrest(i),2)+$
                  ' ('+strtrim(line_z(i),2)+')'
      xyouts,tmp_wobs,state.ydispmin+0.8*(state.ydispmax-state.ydispmin),$
             tmp_label,orient=90,align=0.5,color=90
   endfor
endif

end

;------------------------------------------------------------------------
; Change order
;------------------------------------------------------------------------

Pro change_order,new_order,no_save=no_save

common sp_state
common sp_data

old_order   = state.order
state.order = new_order

; Set default plot limits for the new order
good_ls        = where(wave(*,state.order) gt 0 and $
                       finite(wave(*,state.order)) eq 1 and $
                       finite(flux(*,state.order)) eq 1,n_good)
good_wave     = wave(good_ls,state.order)
good_flux     = state.scale_factor*flux(good_ls,state.order)
state.minwave = min(good_wave)
state.maxwave = max(good_wave)
wave_range = state.maxwave - state.minwave
state.default_xdispmin = min(good_wave) - 0.02*wave_range
state.default_xdispmax = max(good_wave) + 0.02*wave_range
if (state.fixedyrange eq 1) then begin
   ; Use fixed user-specified values
   state.default_ydispmin = state.fixed_ydispmin
   state.default_ydispmax = state.fixed_ydispmax
endif else begin
   ; Automatically determine y range.  Ignore outlying 1% of flux
   ; values when setting display limits
   fluxorder  = sort(good_flux)
   fluxlist   = good_flux(fluxorder)
   flux_range = fluxlist(0.99*n_good) - fluxlist(0.01*n_good)
   state.default_ydispmin = 0 < (fluxlist(0.01*n_good) - 0.05*flux_range)
   state.default_ydispmax = fluxlist(0.99*n_good) + 0.05*flux_range
endelse

; Set plot limits to the defaults
state.xdispmin = state.default_xdispmin
state.xdispmax = state.default_xdispmax
state.ydispmin = state.default_ydispmin
state.ydispmax = state.default_ydispmax

sp_draw_plot

end

;------------------------------------------------------------------------
; Bin arrays
;------------------------------------------------------------------------

Pro sp_bin_arrays,binsz

common sp_state
common sp_data

binsz = long(binsz)

state.binsz = binsz

n_bins = floor(1.*state.npix / binsz)

if (binsz eq 1) then begin
   binflux  = flux
   binerror = error
   binwave  = wave
endif else begin
   binflux  = fltarr(n_bins,state.norders)
   binerror = fltarr(n_bins,state.norders)
   binwave  = dblarr(n_bins,state.norders)
   i = 0L
   for i=0L,long(n_bins-1) do binflux(i,*) = $
                           total(flux(i*binsz:(i+1)*binsz-1,*),1)
   for i=0L,long(n_bins-1) do binerror(i,*) = $
                           total(error(i*binsz:(i+1)*binsz-1,*),1)
   for i=0L,long(n_bins-1) do binwave(i,*) = $
                           total(wave(i*binsz:(i+1)*binsz-1,*),1)
   binflux  = binflux / binsz
   binerror = binerror / binsz
   binwave  = binwave / binsz
endelse

sp_draw_plot

end

;------------------------------------------------------------------------
; Mark doublet wavelengths
;------------------------------------------------------------------------

Pro sp_mark_doublet

common sp_state
common sp_data
common sp_atom

; Case 1: Lower-wavelength line is at the cursor position
z1     = state.xcoord/atom.doublet_lam1 - 1
lam_hi = (1+z1)*atom.doublet_lam2
; Case 2: Higher-wavelength line is at the cursor position 
z2     = state.xcoord/atom.doublet_lam2 - 1
lam_lo = (1+z2)*atom.doublet_lam1

print,atom.doublet_name+' '+strtrim(atom.doublet_lam1,2)+'  z1 = '+$
      strtrim(z1,2)
print,atom.doublet_name+' '+strtrim(atom.doublet_lam2,2)+'  z2 = '+$
      strtrim(z2,2)

; Mark potetial doublet positions
oplot,[state.xcoord,state.xcoord],[-1e29,1e29],color=200,linestyle=1
oplot,[lam_lo,lam_lo],[-1e29,1e29],color=200,linestyle=1
oplot,[lam_hi,lam_hi],[-1e29,1e29],color=200,linestyle=1

end

;------------------------------------------------------------------------
; Measure centroid and equivalent width
;------------------------------------------------------------------------

Pro sp_measure_ew

common sp_state
common sp_data
common sp_atom

print,'Measuring centroid and EW:'

; Get min and max wavelengths over which to measure centroid and EW
print,'   Left-click twice to mark min and max wavelengths'
cursor,wlo,y,/down
oplot,[wlo,wlo],[-1e6,1e6],color=250
cursor,whi,y,/down
oplot,[whi,whi],[-1e6,1e6],color=250

; Set continuum
if (state.norm eq 1) then begin
   print,'   Normalized spectrum, using continuum = 1'
   temp_cont = 1
endif else begin
   print,'   Left-click once to mark continuum'
   cursor,x,temp_cont,/down
endelse
oplot,[-1e6,1e6],[temp_cont,temp_cont],color=250

; Calculate centroid.  Weight by absorption
cent_ls = where(wave ge wlo and wave le whi,n_cent)
if (n_cent eq 0) then begin
   print,'   No pixels selected'
   return
endif
temp_flux = flux(cent_ls)
temp_abs  = temp_cont - temp_flux
temp_wave = wave(cent_ls)
centroid  = total(temp_wave * temp_abs) / total(temp_abs)
print,'   Cent  = '+strtrim(centroid,2)
oplot,[centroid,centroid],[-1e6,1e6],linestyle=1,color=220

; Calculate EW
temp_dispersion = wave(round(mean(cent_ls))) - wave(round(mean(cent_ls))-1)
equiv_width     = temp_dispersion * total(temp_abs) / temp_cont
print,'   EW    = '+strtrim(equiv_width,2)

; Calculate velocity interval between marks
delta_v = c*(whi-wlo)/centroid
print,'   dv    = '+strtrim(delta_v,2)+' km/s (between marks)'

; Calculate redshift, rest EW, and column density
new_rest_wave = ''
new_fvalue    = ''
read,new_rest_wave,prompt='   Rest wavelength ['+strtrim(rest_wave,2)+']: '
read,new_fvalue,prompt='   Oscilator strength ['+strtrim(fvalue,2)+']: '
if (new_rest_wave ne '') then rest_wave = new_rest_wave
if (new_fvalue ne '') then fvalue = new_fvalue
temp_z = centroid / rest_wave - 1
rest_equiv_width = equiv_width / (1+temp_z)
temp_tau  = -alog(temp_flux/temp_cont)
temp_dvperpix    = 2.99792458d5 * convol(temp_wave,[-1,1])/temp_wave
temp_dvperpix(0) = temp_dvperpix(1)
temp_logN        = float(alog10((3.768e14/(double(rest_wave)*double(fvalue)))*$
                          total(temp_tau*temp_dvperpix)))
print,'   z    = '+strtrim(temp_z,2)
print,'   EW_0 = '+strtrim(rest_equiv_width,2)
print,'   logN = '+strtrim(temp_logN,2)

end

;------------------------------------------------------------------------
; Fit Voigt profiles using VPFIT
;------------------------------------------------------------------------

Pro sp_vpfit

common sp_state
common sp_data
common sp_atom

; Currently fits "generic" H I Ly-alpha lines

; If a fit array does not yet exist, create one.  Assume
; flux array is normalized.
if (n_elements(fluxfit) eq 0) then fluxfit = 1 + 0.*flux

; Get min and max wavelengths over which to fit
print,'   Left-click twice to mark min and max wavelengths'
cursor,w1,y,/down
oplot,[w1,w1],[-1e6,1e6],color=250
cursor,w2,y,/down
oplot,[w2,w2],[-1e6,1e6],color=250
wlo = w1 < w2
whi = w1 > w2
wlo_str = strtrim(wlo,2)
whi_str = strtrim(whi,2)

; Include a buffer region that is large enough to hold a wavelength
; table
min_winc = (w1-100) > min(wave)
max_winc = (w2+100) < max(wave)
inc_ls   = where(wave ge min_winc and wave le max_winc,n_inc)

; In the future, allow user to select lines

; Write wavelengths, flux, and error to a temporary ASCII file.
; Include a line that specifies the FWHM.
openw,10,'sp_temp_data.dat'
printf,10,'# Temporary file created by SPECPLOT'
printf,10,'RESVEL 6.7'
for i=min(inc_ls),max(inc_ls) do printf,10,wave(i),flux(i),error(i),1.0
close,10

; Create fort.13 file
openw,10,'fort.13'
printf,10,' %% sp_temp_data.dat  1  '+wlo_str+'  '+whi_str
printf,10,' '
close,10

; Call VPFIT
spawn,'vpfit'

; Generate Voigt profiles from the resulting fort.26 file
temp_fit = abs_lines(wave,f26='fort.26',fwhm=6.7,/lya)
fluxfit  = fluxfit * temp_fit

; Replot, including fit
state.plot_fluxfit = 1
sp_draw_plot

end

;------------------------------------------------------------------------
; Plot metal lines
;------------------------------------------------------------------------

Pro sp_plot_metals,z

common sp_state
common sp_data
common sp_atom

; Get plot parameters
new_wmin       = ''
new_wmax       = ''
new_fmin       = ''
new_dv         = ''
new_normstatus = ''
new_ps         = ''
read,new_wmin,prompt='   Min wavelength ['+strtrim(atom.metals_wmin,2)+']: '
read,new_wmax,prompt='   Max wavelength ['+strtrim(atom.metals_wmax,2)+']: '
read,new_fmin,prompt='   Min f-value ['+strtrim(atom.metals_fmin,2)+']: '
read,new_dv,prompt='   Velocity range (km/s) ['+strtrim(atom.metals_dv,2)+']: '
read,new_normstatus,$
     prompt='   Normalized spec? (0/1) ['+strtrim(state.norm,2)+'] '
read,new_ps,prompt='   PS plot? (0/1) ['+strtrim(atom.metals_ps,2)+']: '
if (new_wmin ne '') then atom.metals_wmin = new_wmin
if (new_wmax ne '') then atom.metals_wmax = new_wmax
if (new_fmin ne '') then atom.metals_fmin = new_fmin
if (new_dv ne '')   then atom.metals_dv   = new_dv
if (new_normstatus ne '') then state.norm = new_normstatus
if (new_ps ne '')   then atom.metals_ps   = new_ps

; Plot metal lines in all orders
metals_wave = binwave
metals_flux = state.scale_factor*binflux
if (state.plot_error eq 1) then metals_error = state.scale_factor*binerror $
   else metals_error = 0
if (atom.metals_ps eq 1) then begin
   ps_openfile,'z'+strtrim(z,2),/color
   charsz    = 1.0
   !x.margin = [6,1]
   !y.margin = [3,2]
endif else begin
   charsz = 2.0
endelse
zlines,z,wave=metals_wave,flux=metals_flux,error=metals_error,$
       wmin=atom.metals_wmin,$
       wmax=atom.metals_wmax,fmin=atom.metals_fmin,dv=atom.metals_dv,$
       norm=state.norm,/plot,/shortlist,charsize=charsz
if (atom.metals_ps eq 1) then ps_closefile
cleanplot,/silent

end

;------------------------------------------------------------------------
; Attempt to identify marked lines
;------------------------------------------------------------------------

Pro sp_id_lines

common sp_state
common sp_data
common sp_atom

; Maake sure at least two lines have been marked
if (n_elements(marked_lines) lt 2) then begin
   print,'Must mark at least two lines first'
   return
endif

; Get id parameters
new_ztol       = ''
new_zmax       = ''
read,new_ztol,prompt='   z tolerance ['+strtrim(atom.ztol,2)+']: '
read,new_zmax,prompt='   z max       ['+strtrim(atom.zmax,2)+']: '
if (new_ztol ne '') then atom.ztol = new_ztol
if (new_zmax ne '') then atom.zmax = new_zmax

; Attempt to match up lines
zlines,marked_lines,ztol=atom.ztol,zmax=atom.zmax,/id,/shortlist

end

;------------------------------------------------------------------------
; Edit arrays
;------------------------------------------------------------------------

Pro sp_edit_arrays

common sp_state
common sp_data

; Select array to edit
if (state.plot_error or state.plot_flux2) then begin
   flux_y  = 0.9
   error_y = 0.8
   flux2_y = 0.7
   xyouts,0.1,flux_y,/norm, 'Array 1',charsize=2
   options_y = [flux_y]
   options   = ['1']
   if state.plot_error then begin
      xyouts,0.1,error_y,/norm,'Array 2',charsize=2,color=250
      options_y = [options_y,error_y]
      options   = [options,'2']
   endif
   if state.plot_flux2 then begin
      xyouts,0.1,flux2_y,/norm,   'Array 3',charsize=2,color=150
      options_y = [options_y,flux2_y]
      options   = [options,'3']
   endif
   print,'Select array to edit (click in window)'
   cursor,x_select,y_select,/down,/norm
   n_options   = n_elements(options)
   options_ind = findgen(n_options)
   choice      = 0 > round(interpol(options_ind,$
                                    options_y,y_select)) < (n_options-1)
   array_to_edit = options(choice)
endif else begin
   array_to_edit = '1'
endelse
print,'   Editting array '+strtrim(array_to_edit,2)

; Select wavelength region to edit
print,'   Click twice for x range'
cursor,x1,y1,/down,/data
oplot,replicate(x1,2),[-1e6,1e6],linestyle=2,color=80
cursor,x2,y2,/down,/data
oplot,replicate(x2,2),[-1e6,1e6],linestyle=2,color=80

; Enter new value
; Kludge for editing metal masks
if (array_to_edit eq 2) then begin
   new_value = 1.1
endif else begin
   new_value = ''
   read,new_value,prompt = '   Enter new value [cancel]: '
endelse
xlo = x1 < x2
xhi = x1 > x2
if (new_value ne '') then begin
   case array_to_edit of
      '1': begin
              edit_ls = where(wave ge xlo and wave le xhi,n_edit)   
              if (n_edit ne 0) then flux(edit_ls) = new_value
           end
      '2': begin
              edit_ls = where(wave ge xlo and wave le xhi,n_edit)   
              if (n_edit ne 0) then error(edit_ls) = new_value
           end
      '3': begin
              edit_ls = where(wave2 ge xlo and wave2 le xhi,n_edit)
              if (n_edit ne 0) then flux2(edit_ls) = new_value
           end
   endcase
endif

; Replot
sp_bin_arrays,state.binsz

; Flag that arrays have been modified
state.modified_arrays = 1

end

;------------------------------------------------------------------------
; Save arrays to FITS files
;------------------------------------------------------------------------

Pro sp_save_arrays

common sp_state
common sp_data

; Allow user to change filenames

new_edt_flux_file  = ''
new_edt_error_file = ''
new_edt_flux2_file = ''

read,new_edt_flux_file,prompt='Flux file ['+file.edt_flux+']:'
   if (new_edt_flux_file ne '') then file.edt_flux = new_edt_flux_file
   writefits,file.edt_flux,flux,fluxhdr

if state.plot_error then begin
   read,new_edt_error_file,prompt='Error file ['+file.edt_error+']:'
      if (new_edt_error_file ne '') then file.edt_error = new_edt_error_file
      writefits,file.edt_error,error,fluxhdr
endif

if state.plot_flux2 then begin
   read,new_edt_flux2_file,prompt='Flux2 file ['+file.edt_flux2+']:'
      if (new_edt_flux2_file ne '') then file.edt_flux2 = new_edt_flux2_file
      writefits,file.edt_flux2,flux2,flux2hdr
endif

print,'Output files written.'

state.modified_arrays = 0

end

;------------------------------------------------------------------------
; Main program - see beginning of file for header
;------------------------------------------------------------------------

Pro specplot,fluxfile,errorfile,wavefile=wavefile,norm=norm,z=z,$
             f2=f2,w2=w2,yrange=yrange,block=block,lines=lines

common sp_state
common sp_data
common sp_atom

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:'
   print,'   specplot,fluxfile,errorfile,wavefile=wavefile,/norm,z=z,'
   print,'            f2=f2,w2=w2,yrange=yrange,/block'
   return
endif

; Initialize common blocks and settings 
specplot_initialize

; Read in data arrays
if (size(fluxfile,/tname) eq 'STRING') then begin
   fluxexists = findfile(fluxfile,count=nfluxfiles)
   if (nfluxfiles eq 0) then begin
      print,'Flux file not found.'
      return
   endif else begin
      if keyword_set(wavefile) then begin
         ; Get wavelength information from separate file
         flux = readfits(fluxfile,fluxhdr)
         if (size(wavefile,/tname) eq 'STRING') then begin
            waveexists = findfile(wavefile,count=nwavefiles)
            if (nwavefiles eq 0) then begin
               print,'Wavelength file not found.'
               wave = findgen(n_elements(flux))
            endif else begin
               wave = readfits(wavefile,wavehdr)
            endelse
         endif else begin
            ; Array input
            wave = wavefile
         endelse
      endif else begin
         ; Get wavelength information from flux header
         flux = rflam(fluxfile,fluxhdr,wave)
      endelse
      file.flux = fluxfile
   endelse
endif else begin
   ; Array input
   flux = fluxfile
   mkhdr,fluxhdr,flux
   ; Wavelength information
   if keyword_set(wavefile) then begin
      if (size(wavefile,/tname) eq 'STRING') then begin
         waveexists = findfile(wavefile,count=nwavefiles)
         if (nwavefiles eq 0) then begin
            print,'Wavelength file not found.'
            wave = findgen(n_elements(flux))
         endif else begin
            wave = readfits(wavefile,wavehdr)
         endelse   
      endif else begin
         ; Array input
         wave = wavefile
      endelse
   endif else begin
      print,'No wavelength information.'
      wave = findgen(n_elements(flux))
   endelse
endelse

; If only a single wavelength order has been provided for a 
; multi-order flux array, use the same wavelengths for each
; order.
flux_sz = size(flux)
wave_sz = size(wave)
if (flux_sz(0) eq 2 and wave_sz(0) eq 1) then begin
   big_wave = make_array(dimension=flux_sz(1:2),type=wave_sz(2))
   for i=0,flux_sz(2)-1 do big_wave(*,i) = wave
   wave = big_wave
endif

; If an error file has been provided, read in the error array.
; Otherwise, use a fake error array.
if (n_params() ge 2) then begin
   if (size(errorfile,/tname) eq 'STRING') then begin      
      errorexists = findfile(errorfile,count=nerrorfiles)
      if (nerrorfiles eq 0) then begin
         print,'Error file not found.'
         return
      endif else begin
         error = readfits(errorfile,errorhdr)
         file.error = errorfile
      endelse
   endif else begin
      ; Array input
      error = errorfile
      mkhdr,errorhdr,error
   endelse
   state.plot_error = 1
endif else begin
   errorhdr  = fluxhdr
   error     = 1 + 0.*flux
   state.plot_error = 0
endelse

; If a second spectrum has been provided, read it in.
; Read in data arrays
if keyword_set(f2) then begin
   if (size(f2,/tname) eq 'STRING') then begin
      flux2exists = findfile(f2,count=nf2s)
      if (nf2s eq 0) then begin
         print,'Flux file 2 not found.'
         return
      endif else begin
         if keyword_set(w2) then begin
            ; Get wavelength information from separate file
            flux2 = readfits(f2,flux2hdr)
            wave2exists = findfile(w2,count=nw2s)
            if (nw2s eq 0) then begin
               print,'Wavelength file 2 not found.'
               wave2 = findgen(n_elements(flux2))
            endif else begin
               wave2 = readfits(w2,wave2hdr)
            endelse
         endif else begin
            ; Get wavelength information from flux2 header,
            ; if available
            flux2hdr = headfits(f2)
            lam0 = sxpar(flux2hdr,'CRVAL1',count=lam0_count)
            if (lam0_count ne 0) then begin
               flux2 = rflam(f2,flux2hdr,wave2)
            endif else begin
               ; Assume input arrays have the same wavelengths
               flux2 = readfits(f2,flux2hdr)
               wave2 = wave
            endelse
         endelse
         file.flux2 = f2
      endelse
   endif else begin
      ; Array input
      flux2 = f2
      mkhdr,flux2hdr,flux2
      ; Wavelength information
      if keyword_set(w2) then begin
         if (size(w2,/tname) eq 'STRING') then begin
            wave2exists = findfile(w2,count=nw2s)
            if (nw2s eq 0) then begin
               print,'Wavelength file 2 not found.'
               wave2 = findgen(n_elements(flux2))
            endif else begin
               wave2 = readfits(w2,wave2hdr)
            endelse   
         endif else begin
            ; Array input
            wave2 = w2
         endelse
      endif else begin
         ; Assume input arrays have the same wavelengths
         wave2 = wave
      endelse
   endelse
   state.plot_flux2 = 1
endif else begin
   flux2 = 0
   wave2 = 0
   state.plot_flux2 = 0
endelse

; Plotting normalized arrays?
if keyword_set(norm) then state.norm = 1 else state.norm = 0

; Specifcy plot y range?
if keyword_set(yrange) then begin
   state.fixedyrange    = 1
   state.fixed_ydispmin = yrange(0)
   state.fixed_ydispmax = yrange(1)
endif

; Use at least floating point for plotting
input_type = size(flux,/type)
flux       = 1. * flux
error      = 1. * error

; Determine number of orders (rows)
sz = size(flux)
if (sz(0) eq 1) then begin
   state.norders = 1
   state.npix    = sz(1)
endif else begin
   state.norders = sz(2)
   state.npix    = sz(1)
endelse

; Object redshift specified?
if keyword_set(z) then state.zqso = z

; Set defaults for plotting metal lines
if keyword_set(z) then state.zqso = z
atom.metals_wmin = lya*(1+state.zqso) > min(wave)
atom.metals_wmax = max(wave)
atom.metals_fmin = 0.047    ; Small enough to include O I 1302

; Set the initial binned arrays to be the input arrays
binflux  = flux
binerror = error
binwave  = wave

; Initialize output file names
flux_root      = filename(file.flux,/noextension,dir=dir)
file.edt_flux  = dir+flux_root+'_edt.fits'
error_root     = filename(file.error,/noextension,dir=dir)
file.edt_error = dir+error_root+'_edt.fits'
flux2_root     = filename(file.flux2,/noextension,dir=dir)
file.edt_flux2 = dir+flux2_root+'_edt.fits'

; Start with no lines marked
marked_lines = -1

; Use rainbow color indices
device,decomp=0
loadct,39,/silent

; One plot per window
!p.multi = [0,1,1]

; Create a new window if one does not already exist
if not(xregistered('specplot_window')) then create_specplot_window

; Start with first order
change_order,0,/no_save

; Print help reminder
print,"Type '?' for menu."

; Register widgets with optional blocking
if keyword_set(block) then no_blocking=0 else no_blocking=1
xmanager,'specplot_window',state.base_id,group_leader=group,$
         event_handler='resize_specplot_window',no_block=no_blocking

; Return marked lines (this only works with blocking)
lines = marked_lines

return
end
