Pro xshooter_multiextract,filelist,outfile,arm=arm,response=response,$
                    respwave=respwave,bin=bin,angstrom=angstrom,$
                    extra_var=extra_var,fid_row=fid_row,fid_wave=fid_wave,$
                    medscale=medscale,vacuum=vacuum,notellcorr=notellcorr

;+--------------------------------------------------------------------------
;
; XSHOOTER_MULTIEXTRACT       2/2010
;
; Perform simutaneous optimal extraction on multiple 2D frames.
;
; Note: Assumes each exposure has the same number of orders, and that
; the orders are listed in the orderlist files in order or increasing
; (or decreasing) wavelength/slit position.
;
; Must edit the filenames for the atmospheric extinction model
; (ext_file) in the program body below.
; 
; INPUTS
;        filelist   - String, name of multi-column ASCII file
;                     containing list of 2D files from which to
;                     extract spectrum, OR 4 x n string array with
;                     this information.  Format is [rootname,
;                     orderlist, slitgrid, telluric_file], where
;                     orderlist and slitgrid are the inputs into
;                     xshooter_reduce for each object in the list, and
;                     telluric_file is a 1D file with the telluric
;                     absorption spectrum to be applied to this frame
;                     (typically output from atm_correct; may be
;                     omitted).  The program will look for the files
;                     <root>_sub.fits, <root>_var.fits,
;                     <root>_wav.fits, <root>_spec.fits,
;                     <root>_wspec.fits, and <root>_profile.dat, which
;                     are output by xshooter_reduce.
;
;                     Note: The orderlists should contain only the
;                     orders that were actually reduced.  For example,
;                     if extracting spectra from NIR exposures taken
;                     with the K-blocking filter, orders >= 14 in the
;                     standard nir_orderlist.dat should be commented
;                     out.
;
; KEYWORDS
;        arm        - String specifying which X-Shooter arm: 'UVB',
;                     'VIS', or 'NIR'.  Must be set.
;        response   - String, FITS filename name of reponse function
;                     (2D array).  If set, input fluxes will be scaled
;                     before combining.  Must also set respwave keyword.
;        respwave   - String, FITS filename name of reponse function  
;                     wavelengths (2D array)
;        bin        - Wavelength bin, in km/s.  Default is 8.  If the
;                     /angstrom keyword is set, then this value is in
;                     angstroms, and the default is 0.5.
;        angstrom   - If set, output pixels will be a constant size
;                     in Angstroms.  Default is to use constant size
;                     in km/s.
;        extra_var  - String, FITS filename of 2D variance array to be
;                     added to the variance array for each exposure
;                     prior to extracting.  Useful for including variance
;                     not formally computed by longslit_reduce (e.g.,
;                     the variance in the dark frames, or variance for
;                     individual bad pixels).  Default is
;                     to use only the formal variance array for each
;                     exposure, <root>_var.fits.
;        fid_row    - Scalar, fiducial row to be used for relative
;                     flux scaling. Defaults defined in program body.
;                     Note: This will be different than the 'fiducial
;                     order' in xshooter_reduce.  The bluest extracted
;                     order will be row 0.
;        fid_wave   - 2-element array giving the minimum and maximum
;                     wavelengths within the fid_row to be used for relative
;                     flux calibration.  Default is to use entire wavelength
;                     interval.
;        medscale   - NOT ACTIVE If set, exposures will be scaled to
;                     roughly match one another using median flux
;                     values prior to performing optimal extraction.
;                     Recommended for very faint exposures.  Default
;                     is to attempt to remove differences in spectral
;                     shape as a function of x by fitting a polynomial
;                     to the median flux ratio across each order.
;        vacuum     - If set, assumes input wavelengths are in air,
;                     and converts to vacuum wavelengths
;        notellcorr - If set, will not perform telluric absorption 
;                     correction even if telluric spectra are specified
;                     in the input filelist.
;
; OUTPUTS
;        outfile    - String, <root> name of output files.  Program
;                     will write extracted 1D spectrum to
;                     <outfile>.fits and the 1D error to
;                     <outfile>e.fits.  Order-by-order extractions are
;                     written to <outfile>+'_multi_flux.fits',
;                     <outfile>+'_multi_err.fits', and
;                     <outfile>+'_multi_wave.fits'.
;
; HISTORY
;        Adapted from NE_MULTIEXTRACT  2/18/2010 GDB
;        FID_ROW, FID_WAVE keywords added 9/19/2011 GDB
;        Telluric absorption correction of 2D frames added 6/14/2012 GDB
;        ANGSTROM keyword added 8/23/2012
;        Atmospheric extinction correction added 10/16/2013 GDB
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
   print,'CALLING SEQUENCE:'
   print,'   xshooter_multiextract,filelist,outfile,response=response,bin=bin'
   print,'                   extra_var=extra_var,fid_row=fid_row,'
   print,'                   fid_wave=fid_wave,/medscale,/vacuum,/notellcorr'
   return
endif

;;; Which arm?
if not(keyword_set(arm)) then begin
   print,'Must specify which X-Shooter arm'
   return
endif
; Allow upper or lower case
case arm of
   'uvb': arm = 'UVB'
   'UVB': arm = 'UVB'
   'vis': arm = 'VIS'
   'VIS': arm = 'VIS'
   'nir': arm = 'NIR'
   'NIR': arm = 'NIR'
   else : begin
             print,'X-Shooter arm not recognized'
             return
          end
endcase

;;; Read in file information
if (n_elements(filelist) eq 1) then begin
   ; Read exposure information from ASCII file
   exposure_templ =  {   $
             version: 1.0, $
             datastart: 0, $
             delimiter: ' ', $
             missingvalue: 'Unknown', $
             commentsymbol: '#', $
             fieldcount: 4, $
             fieldtypes: [7,7,7,7], $
             fieldnames: ['root','orderlist','slit','tell_file'], $
             fieldlocations: [0,23,51,75], $
             fieldgroups: [0,1,2,3] $
           }
   exposure  = read_ascii(filelist,template=exposure_templ)
   rootname  = exposure.root
   orderlist = exposure.orderlist
   slitgrid  = exposure.slit     
   tell_file = exposure.tell_file
endif else begin
   ; Array input of exposure information
   rootname  = reform(filelist(0,*))
   orderlist = reform(filelist(1,*))
   slitgrid  = reform(filelist(2,*))
   tell_file = reform(filelist(3,*))  ; Should be '0' if none
endelse
n_exp = n_elements(rootname)

;;; Atmospheric extinction model
;;; X-Shooter model is taken from ESO reduction pipeline.  Models for
;;; vis and nir are identical to uvb, and only extend to 1.04 microns.
ext_file  = '~/eso/xshooter/flux_cal/calibrations/'+$
            'xsh_paranal_extinct_model_uvb.fits'
ext_model = mrdfits(ext_file,1,ext_hdr)
ext_w     = 10.*ext_model.lambda   ; Angstroms
ext_k     = ext_model.extinction   ; mag/airmass
; Assume a constant value at wavelength higher than those modeled
nbin_ext  = n_elements(ext_w)
ext_dw    = ext_w(nbin_ext-1) - ext_w(nbin_ext-2)
ext_w     = [ext_w,ext_w(nbin_ext-1)+ext_dw]
ext_k     = [ext_k,ext_k(nbin_ext-1)]

;;; Read in 2D files.  If a reference file for the telluric
;;; correction is given, apply telluric correction to the
;;; 2D frames.
testframe   = readfits(rootname(0)+'_sub.fits',hdr)
sz          = size(testframe)
xsz         = sz(1)
ysz         = sz(2)
big_sub     = dblarr(xsz,ysz,n_exp)
big_var     = dblarr(xsz,ysz,n_exp)
big_wav     = dblarr(xsz,ysz,n_exp)
big_slit    = dblarr(xsz,ysz,n_exp)
big_airmass = 1. + fltarr(n_exp)    ; Default is 1.0
big_texp    = fltarr(n_exp)
for i=0,n_exp-1 do begin
   ; 2D arrays
   exp_sub = readfits(rootname(i)+'_sub.fits',tmp_hdr)
   exp_var = readfits(rootname(i)+'_var.fits')
   exp_wav = readfits(rootname(i)+'_wav.fits')
   ; Airmass
   airmass = gxpar(tmp_hdr,'AIRMASS',count=n_airmass)
   if (n_airmass eq 0) then $
      airmass = gxpar(tmp_hdr,'HIERARCH ESO TEL AIRM END',count=n_airmass)
   if (n_airmass eq 0) then begin
      print,'NO AIRMASS INFORMATION.  Taking airmass=1.0'
      airmass = 1.0
   endif
   ; Exposure time
   texp = gxpar(hdr,'EXPTIME',count=n_texp)
   if (n_texp eq 0) then begin
      print,'NO EXPOSURE TIME INFORMATION'
      texp = 0.0
   endif
   ; Correction for atmospheric extinction
   corr_ls    = where(exp_wav gt 0)
   k          = fltarr(xsz,ysz)
   k(corr_ls) = interpol(ext_k,ext_w,exp_wav(corr_ls))
   frac_trans = 1. + fltarr(xsz,ysz)
   frac_trans(corr_ls) = 10.^(-0.4*airmass*k(corr_ls))
   exp_sub    = exp_sub / frac_trans
   exp_var    = exp_var / (frac_trans^2)
   ; Telluric correction
   if (tell_file(i) ne '0' and tell_file(i) ne 'Unknown' and $
       not(keyword_set(notellcorr))) then begin
      ; Read in telluric absorption spectrum, interpolate onto
      ; 2D wavelengths, and divide out from 2D frames.
      ftell   = rflam(tell_file(i),htell,wtell)
      tell_ls = where(exp_wav ge min(wtell) and exp_wav le max(wtell))
      ftell2d = 1.0 + fltarr(xsz,ysz)
      ftell2d(tell_ls) = interpol(ftell,wtell,exp_wav(tell_ls))
      exp_sub          = exp_sub / ftell2d
      exp_var          = exp_var / (ftell2d^2)
   endif
   ; Convert to vacuum wavelengths?  Perform on individual
   ; exposures to save memory.
   if keyword_set(vacuum) then airtovac,exp_wav
   ; Save arrays
   big_sub(*,*,i)  = exp_sub
   big_var(*,*,i)  = exp_var
   big_wav(*,*,i)  = exp_wav
   big_slit(*,*,i) = readfits(slitgrid(i))
   big_airmass(i)  = airmass
   big_texp(i)     = texp
endfor

;;; Arrays of x and y values
xarr = rebin(findgen(xsz),xsz,ysz)  ; These should be floating point
yarr = rebin(rotate(findgen(ysz),1),xsz,ysz)

;;; Add extra variance (optional)
if keyword_set(extra_var) then begin
   additional_var = readfits(extra_var)
   for i=0,n_exp-1 do $
      big_var(*,*,i) = big_var(*,*,i) + additional_var
endif

;;; Read in order information for each exposure
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
example_orderinfo = read_ascii(orderlist(0),template=order_templ)
n_orders = n_elements(example_orderinfo.center)
big_orderinfo = replicate(example_orderinfo,n_exp)
for i=0,n_exp-1 do $
   big_orderinfo(i) = read_ascii(orderlist(i),template=order_templ)

;;; Read object profile information for each exposure
profile_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 6, $
          fieldtypes: [4,5,4,4,4,4], $
          fieldnames: ['ordernum','meanwav','fwhm','pos','offset','shift'], $
          fieldlocations: [4,14,28,38,52,63], $
          fieldgroups: [0,1,2,3,4,5] $
        }
example_profile = read_ascii(rootname(0)+'_profile.dat',template=profile_templ)
big_profile     = replicate(example_profile,n_exp)
for i=0,n_exp-1 do big_profile(i) = $
            read_ascii(rootname(i)+'_profile.dat',template=profile_templ)

;; TESTING - For extracting "blank" spectra from other parts of the
;;           slit.  Must also comment out refitting profile with spline
;;           below.
;big_profile.pos = big_profile.pos - 2*big_profile.offset

;;; Masking for NIR arm
mask = lonarr(xsz,ysz)
; No masking for UVB and VIS arms.  Orders are well away from
; pre/overscan regions
if (arm eq 'NIR') then begin
   ; Red edge x(y) polynomial
   r_terms = [1459.7389, 0.65775340, -0.00090907266, 3.2113316e-06]
   r_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(r_terms)-1 do r_yfit = r_yfit + r_terms(i)*(yarr^i)
   r_mask_ls = where(xarr gt r_yfit)
   ; Blue edge 
   b_terms = [433.63313, -0.30479697, 0.00011715153, -1.1558709e-06]
   b_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(b_terms)-1 do b_yfit = b_yfit + b_terms(i)*(yarr^i)
   b_mask_ls = where(xarr lt b_yfit)
   mask(r_mask_ls)     = 1
   mask(b_mask_ls)     = 1
   mask(0:3,*)         = 1
   mask(xsz-4:xsz-1,*) = 1
   mask(*,0:3)         = 1
endif

;;; Scale flux in exposures to match
scale_list     = 1 + fltarr(n_exp)
temp_spec      = readfits(rootname(0)+'_spec.fits')
temp_err       = readfits(rootname(0)+'_err.fits')
temp_wspec     = readfits(rootname(0)+'_wspec.fits')
; Select order for relative flux scaling
case arm of
   'UVB': fiducial_order =  3
   'VIS': fiducial_order = 10  ; 11 for z > 5 QSOs
   'NIR': fiducial_order =  4
endcase
if keyword_set(fid_row) then fiducial_order = fid_row
positive_ls = where(temp_wspec(*,fiducial_order) gt 0)
fit_wmin    = min(temp_wspec(positive_ls,fiducial_order))
fit_wmax    = max(temp_wspec(positive_ls,fiducial_order))
if keyword_set(fid_wave) then begin
   fit_wmin = min(fid_wave) > fit_wmin
   fit_wmax = max(fid_wave) < fit_wmax
endif
fid_ls = where(temp_wspec(*,fiducial_order) ge fit_wmin and $
               temp_wspec(*,fiducial_order) le fit_wmax)
fiducial_spec  = temp_spec(fid_ls,fiducial_order)
fiducial_wspec = temp_wspec(fid_ls,fiducial_order)
; Correct for atmospheric extinction before fitting
fid_airmass   = big_airmass(0)
fid_k         = interpol(ext_k,ext_w,fiducial_wspec)
fid_trans     = 10.^(-0.4*fid_airmass*fid_k)
fiducial_spec = fiducial_spec / fid_trans
; Scale exposures by doing a rough polynomial fit to
; the overall spectral shape in the fiducial order.  Should
; be insensitive to outliers.  Fiducial order must have
; significant flux  and a fairly simple shape to work properly.
; WARNING: Fit may become negative.
fiducial_pfit  = robust_poly_fit(fiducial_wspec,fiducial_spec,2,$
                                 fiducial_specfit)
pos_ls         = where(fiducial_specfit gt 0)
for i=1,n_exp-1 do begin
   temp_spec  = readfits(rootname(i)+'_spec.fits')
   temp_wspec = readfits(rootname(i)+'_wspec.fits')
   temp_ls    = where(temp_wspec(*,fiducial_order) ge fit_wmin and $
                      temp_wspec(*,fiducial_order) le fit_wmax)
   spec       = temp_spec(temp_ls,fiducial_order)
   wspec      = temp_wspec(temp_ls,fiducial_order)
   ; Correct for atmospheric extinction before fitting
   temp_airmass = big_airmass(i)
   temp_k       = interpol(ext_k,ext_w,wspec)
   temp_trans   = 10.^(-0.4*temp_airmass*temp_k)
   spec         = spec / temp_trans
   pfit       = robust_poly_fit(wspec,spec,2,specfit)
   match_fit  = interpol(specfit,wspec,fiducial_wspec)
   scale_list(i) = median(fiducial_specfit(pos_ls)/match_fit(pos_ls))
   ;match_spec = interpol(spec,wspec,fiducial_wspec)
   ;scale_list(i)  = median(fiducial_spec/match_spec)  ; Results vary
   big_sub(*,*,i) = big_sub(*,*,i) * scale_list(i)
   big_var(*,*,i) = big_var(*,*,i) * (scale_list(i)^2)
endfor
print,'Relative scaling (smaller values mean higher signal):'
for i=0,n_exp-1 do print,rootname(i)+': '+strtrim(scale_list(i),2)

; Calculate a median effective exposure time for flux calibration,
; given that counts in each frame will be scaled to the first
; exposure.  If the first exposure is comparatively faint, then the
; effective exposure time should be less than the actual exposure time
; of the first exposure.
effective_texp = median(big_texp*scale_list)

;;;; TESTING - Scale flux in exposures to match fiducial exposure.
;;;;           Attempt to remove differences in spectral shape as a
;;;;           function of x by fitting the median flux ratio across
;;;;           each order.
;scale_list    = 1 + fltarr(n_exp)
;fid_spec      = readfits(rootname(0)+'_spec.fits')
;fid_wspec     = readfits(rootname(0)+'_wspec.fits')
;pfit_order    = 2
;scale_pfit    = fltarr(pfit_order+1,n_exp)
;scale_fn      = fltarr(1024,n_exp)
;scale_fn(*,0) = 1  ; Exposures will be scaled to first exposure
;for k=1,n_exp-1 do begin
;   ; Ratio of flux between the current and fiducial spectra
;   temp_spec  = readfits(rootname(k)+'_spec.fits')
;   temp_wspec = readfits(rootname(k)+'_wspec.fits')
;   ; Binned fluxes in each order
;   bin_x  = fltarr(20,n_orders)
;   bin_f1 = fltarr(20,n_orders)
;   bin_f2 = fltarr(20,n_orders)
;   for i=0,n_orders-1 do begin
;      ; Identify pixels with flux over a common wavelenth range
;      temp_wmin = min(temp_wspec(where(temp_wspec(*,i) gt 0),i))
;      temp_wmax = max(temp_wspec(where(temp_wspec(*,i) gt 0),i))
;      fid_wmin  = min(fid_wspec(where(fid_wspec(*,i) gt 0),i))
;      fid_wmax  = max(fid_wspec(where(fid_wspec(*,i) gt 0),i))
;      wmin      = temp_wmin > fid_wmin
;      wmax      = temp_wmax < fid_wmax
;      fid_ls    = where(fid_wspec(*,i) ge wmin and fid_wspec(*,i) le wmax)
;      temp_ls   = where(temp_wspec(*,i) ge wmin and temp_wspec(*,i) le wmax)
;      w1        = fid_wspec(fid_ls,i)
;      f1        = fid_spec(fid_ls,i)
;      w2        = temp_wspec(temp_ls,i)
;      f2        = temp_spec(temp_ls,i)
;      if (max(abs((w2-w1)/w1) gt 0.01)) then begin
;         print,'Wavelength intervals not equal.'
;         stop
;      endif
;      ; Convert wwavelengths to x-values, assuming spectra was
;      ; extracted across the entire array.
;      x = interpol([0.,1023.],[temp_wmin,temp_wmax],w1)
;      ; Discard ends of the spectra
;      fitls  = where(x ge 15 and x le 997)
;      fit_x  = x(fitls)
;      fit_f1 = f1(fitls)
;      fit_f2 = f2(fitls)
;      ; Compute median flux in bins
;      h         = histogram(fit_x,bin=50,min=0,rev=ri)
;      nbin      = n_elements(h)
;      for j=0,19 do begin
;         if (h(j) ne 0) then begin
;            bin_x(j,i)  = median(fit_x(ri(ri(j):ri(j+1)-1)))
;            bin_f1(j,i) = median(fit_f1(ri(ri(j):ri(j+1)-1)))
;            bin_f2(j,i) = median(fit_f2(ri(ri(j):ri(j+1)-1)))
;         endif
;      endfor
;   endfor
;   ; Ratio of binned fluxes
;   bin_r = bin_f2 / bin_f1
;   ; Fit the median ratio in each bin from orders that
;   ; should have good S/N
;   inner_xmed   = median(bin_x(*,4:8),dim=2)
;   inner_rmed   = median(bin_r(*,4:8),dim=2)
;   temp_2dscale = fltarr(xsz,ysz)
;      if keyword_set(medscale) then begin
;         temp_2dscale(*) = median(inner_rmed)
;         inner_pfit      = [median(inner_rmed),0,0]
;      endif else begin
;         ; Polynomial scaling
;         inner_pfit = robust_poly_fit(inner_xmed,inner_rmed,pfit_order,$
;                                      inner_rfit)
;         ; Evaluate scale value over entire array and divide through
;
;         for i=0,pfit_order do $
;            temp_2dscale = temp_2dscale + inner_pfit(i)*(xarr^i)
;      endelse
;   big_sub(*,*,k) = big_sub(*,*,k) / temp_2dscale
;   big_var(*,*,k) = big_var(*,*,k) / temp_2dscale
;   ; Save scale function for this exposure
;   scale_pfit(*,k) = inner_pfit
;   scale_fn(*,k)   = temp_2dscale(*,0)
;endfor

;;; Read in response function (optional)
if keyword_set(response) then begin
   raw_resp_fn  = readfits(response)
   raw_resp_wav = readfits(respwave)
   ; Temporarily divide out mean response function value
   ; to avoid floating point errors
   resp_ls         = where(raw_resp_wav gt 0)
   mean_resp_value = mean(raw_resp_fn(resp_ls))
   raw_resp_fn     = raw_resp_fn / mean_resp_value
   ; Suppress regions where adjacent orders have much higher
   ; sensitivity
   rfactor      = 5
   n_resp_order = n_elements(raw_resp_fn(0,*))
   resp_fn      = raw_resp_fn
   resp_wav     = raw_resp_wav
   for i=0,n_resp_order-1 do begin
      tmp_ls   = where(raw_resp_wav(*,i) gt 0,tmp_n)
      tmp_r    = raw_resp_fn(tmp_ls,i)
      tmp_w    = raw_resp_wav(tmp_ls,i)
      tmp_wmed = median(tmp_w)
      ind_ls   = [i-1,i+1]
      for j=0,1 do begin
         ind = ind_ls(j)
         if (ind lt 0 or ind gt n_resp_order-1) then continue
         tmp_ls2     = where(raw_resp_wav(*,ind) gt 0)
         tmp_r2      = raw_resp_fn(tmp_ls2,ind)
         tmp_w2      = raw_resp_wav(tmp_ls2,ind)
         tmp_wmed2   = median(tmp_w2)
         tmp_r2_match = interpol(tmp_r2,tmp_w2,tmp_w)
         outside_ls    = where(tmp_w lt min(tmp_w2) or $
                               tmp_w gt max(tmp_w2),n_outside)
         if (n_outside ne 0) then tmp_r2_match(outside_ls) = $
            (1e9 > (1000*median(tmp_r)))
         mask_ls = where(tmp_r ge rfactor*tmp_r2_match,n_mask)
         if (n_mask gt 0) then begin
            if (tmp_wmed2 gt tmp_wmed) then begin
               resp_fn(tmp_ls(min(mask_ls):tmp_n-1),i)  = 0.
               resp_wav(tmp_ls(min(mask_ls):tmp_n-1),i) = 0.
            endif else begin
               resp_fn (tmp_ls(0:max(mask_ls)),i) = 0.
               resp_wav(tmp_ls(0:max(mask_ls)),i) = 0.
            endelse
         endif
      endfor
      ; Mask extreme blue
      ls3100 = where(tmp_w lt 3100,n3100)
      if (n3100 ne 0) then begin
         resp_fn (tmp_ls(ls3100),i) = 0.
         resp_wav(tmp_ls(ls3100),i) = 0.
      endif
      ; Special masking for UVB arm
      if (arm eq 'UVB') then begin
         if (abs(tmp_wmed-4330) lt 100) then begin
            ; Mask end of order affected by bad columns, where
            ; the response fn may be bad.  This region is probably
            ; already masked above.
            mask_ls = where(tmp_w lt 4230)
            resp_fn (tmp_ls(mask_ls),i) = 0.
            resp_wav(tmp_ls(mask_ls),i) = 0.
         endif
         if (abs(tmp_wmed-5660) lt 100) then begin
            ; Mask region affected by dischroic, for which
            ; fluxing is extremely difficult.
            mask_ls = where(tmp_w gt 5600)
            resp_fn (tmp_ls(mask_ls),i) = 0.
            resp_wav(tmp_ls(mask_ls),i) = 0.
         endif
      endif
      ; Special masking for VIS arm
      if (arm eq 'VIS') then begin
         if (abs(tmp_wmed-5417) lt 100) then begin
            ; Mask bluest order, which is past dichroic
            resp_fn (*,i) = 0.
            resp_wav(*,i) = 0.
         endif
         ; Mask far red
         ls10270 = where(tmp_w gt 10270,n10270)
         if (n10270 gt 0) then begin
            resp_fn (tmp_ls(ls10270),i) = 0.
            resp_wav(tmp_ls(ls10270),i) = 0.
         endif
         ; Mask blueward of dichroic? (difficult to flux)
      endif
   endfor
endif else begin
   ; No response function
   mean_resp_value = 1.0
endelse

;;; Compute probability function for each order.  Also create
;;; order map for each exposure.
print,'Applying spatial profile and response function...'
big_prob     = fltarr(xsz,ysz,n_exp)
big_ordermap = -1 + bytarr(xsz,ysz,n_exp)
for i=0,n_exp-1 do begin
   exp_slit       = big_slit(*,*,i)
   exp_wav        = big_wav(*,*,i)
   exp_ordernum   = big_orderinfo(i).ordernum
   exp_center     = big_orderinfo(i).center
   exp_orderedge1 = big_orderinfo(i).left
   exp_orderedge2 = big_orderinfo(i).right
   exp_pos        = big_profile(i).pos
   exp_fwhm       = big_profile(i).fwhm  
   exp_prob       = fltarr(xsz,ysz)
   exp_sub        = big_sub(*,*,i)
   exp_var        = big_var(*,*,i)
   exp_ordermap   = -1 + bytarr(xsz,ysz)
   ; Read in extracted 1D spectrum for this exposure
   f1 = readfits(rootname(i)+'_spec.fits')
   e1 = readfits(rootname(i)+'_err.fits')
   w1 = readfits(rootname(i)+'_wspec.fits')
   if keyword_set(vacuum) then airtovac,w1
   ;; TESTING
   ;!p.multi = [0,4,4]
   for j=0,n_orders-1 do begin
      ; Illuminated region
      illum_ls   = where(exp_slit ge exp_orderedge1(j) and $
                         exp_slit le exp_orderedge2(j) and $
                         mask eq 0)
      illum_slit = exp_slit(illum_ls)
      exp_ordermap(illum_ls) = exp_ordernum(j)
      ; Gaussian parameters
      t1 = exp_pos(j)                         ; Position
      t2 = exp_fwhm(j) / (2.*sqrt(2*alog(2))) ; Width (sigma)
      const = 1./(t2*sqrt(2*!pi))             ; Normalized
      exp_prob(illum_ls) = const * exp(-1*((illum_slit-t1)^2)/(2*t2^2))
      ; For orders with high flux, re-compute probability function
      ; using spline fit.  First isolate pixels in this order.
      illum_sub  = exp_sub(illum_ls)
      illum_wav  = exp_wav(illum_ls)
      order_f1   = f1(*,j)
      order_e1   = e1(*,j)
      order_w1   = w1(*,j)
      pos_ls     = where(order_w1 gt 0 and order_e1 gt 0)
      order_f1   = order_f1(pos_ls)
      order_e1   = order_e1(pos_ls)
      order_w1   = order_w1(pos_ls)
      ; Does a substantial amount of this order contain high flux?
      nsig_thresh  = 5
      filt_f1      = median(order_f1,5)
      highflux_ls1 = where(filt_f1/order_e1 gt nsig_thresh,n_highflux1)
      if (n_highflux1 ge 0.1*n_elements(order_f1)) then begin
         print,'Re-evaluating profile for '+rootname(i)+$
               ' order '+strtrim(exp_ordernum(j),2)
         ; Select wavelengths with significant flux (filtered), and
         ; normalize counts along this order by the extracted flux
         ; (actual).
         match_f1      = interpol(order_f1,order_w1,illum_wav)
         match_filt_f1 = interpol(filt_f1,order_w1,illum_wav)
         match_e1      = interpol(order_e1,order_w1,illum_wav)
         highflux_ls2  = where(match_filt_f1/match_e1 ge nsig_thresh)
         norm_sub      = (illum_sub/match_f1)(highflux_ls2)
         norm_slit     = illum_slit(highflux_ls2)
         ; Reject outlying pixels
         reject       = 0.*norm_sub
         norm_h  = histogram(norm_slit,bin=0.5,rev=norm_ri)
         for t=0,n_elements(norm_h)-1 do begin
            bin_ls  = norm_ri(norm_ri(t):norm_ri(t+1)-1)
            bin_sub = norm_sub(bin_ls)
            bin_med = median(bin_sub)
            bin_sdev = robust_sigma(bin_sub)
            if bin_sdev eq -1 then begin
               ; Probably at the end of the slit, with too few points
               n_rej  = n_elements(bin_ls)
               rej_ls = lindgen(n_rej)
            endif else begin
               rej_ls   = where(abs(bin_sub-bin_med) ge 3*bin_sdev,n_rej)
            endelse
         if (n_rej ne 0) then reject(bin_ls(rej_ls)) = 1
         endfor
         keep_ls   = where(reject eq 0)
         keep_sub  = norm_sub(keep_ls)
         keep_slit = norm_slit(keep_ls)
         ; Fit spline
         sset = bspline_iterfit(keep_slit,keep_sub,bkspace=0.5,yfit=yfit,rej=3)
         ; Evaluate over a 1D array
         slit_binsz = 0.1
         n_slitbin  = (exp_orderedge2(j) - exp_orderedge1(j))/slit_binsz + 1
         slit_1d    = exp_orderedge1(j) + slit_binsz*findgen(n_slitbin)
         profile_1d = bspline_valu(slit_1d,sset)
         ; Get normalization.  Suppress edges where spline may
         ; diverge.  Note: this will not be appropriate for sources
         ; with multiple peaks that are well separated.
         ind_lo_1d = max(where(profile_1d le 0 and slit_1d le exp_pos(j)))
         ind_hi_1d = min(where(profile_1d le 0 and slit_1d ge exp_pos(j)))
         if (ind_lo_1d ne -1) then profile_1d(0:ind_lo_1d) = 0.
         if (ind_hi_1d ne -1) then profile_1d(ind_hi_1d:n_slitbin-1) = 0.
         norm_factor = 1./(total(profile_1d)*slit_binsz)
         ; Evaluate over entire order
         illum_prob = norm_factor * bspline_valu(illum_slit,sset)
         if (ind_lo_1d ne -1) then $
            illum_prob(where(illum_slit le slit_1d(ind_lo_1d))) = 0.
         if (ind_hi_1d ne -1) then $
            illum_prob(where(illum_slit ge slit_1d(ind_hi_1d))) = 0.
         exp_prob(illum_ls) = illum_prob
         ;; TESTING
         ;plot,keep_slit,keep_sub,psym=3,xr=t1+[-20,20],/xs,$
         ;     title=strtrim(exp_ordernum(j),2)
         ;oplot,slit_1d,profile_1d,color=250
         ;key=get_kbrd()
      endif
      ; Scale 2D counts by response function (optional)
      if keyword_set(response) then begin
         temp_response = resp_fn(*,j)
         temp_respwave = resp_wav(*,j)
         resp_ls       = where(temp_respwave gt 0,n_resp)
         if (n_resp eq 0) then begin
            ; No usable wavelength range
            exp_prob(illum_ls) = 0.
         endif else begin
            temp_response = temp_response(resp_ls)
            temp_respwave = temp_respwave(resp_ls)
            resp_match    = $
                interpol(temp_response,temp_respwave,exp_wav(illum_ls))
            exp_sub(illum_ls) = exp_sub(illum_ls) * resp_match
            exp_var(illum_ls) = exp_var(illum_ls) * (resp_match^2)
            ; Suppress regions beyond the wavelength range of
            ; the response function
            noresp_ls = where(exp_wav(illum_ls) lt min(temp_respwave) or $
                              exp_wav(illum_ls) gt max(temp_respwave),n_noresp)
            if (n_noresp ne 0) then exp_prob(illum_ls(noresp_ls)) = 0.
            ; Suppress points with very low throughput (This should
            ; already be mostly done above where the the resp fn is read
            ; in)
            nonzero_ls = where(temp_response gt 0)
            low_ls = where(resp_match gt 100.*min(temp_response(nonzero_ls)),$
                           n_low)
            if (n_low ne 0) then exp_prob(illum_ls(low_ls)) = 0.
         endelse
      endif
   endfor
   big_prob(*,*,i)     = exp_prob
   big_ordermap(*,*,i) = exp_ordermap
   if keyword_set(response) then begin
      big_sub(*,*,i) = exp_sub
      big_var(*,*,i) = exp_var
   endif
endfor

;;;; Inverse variance array
;big_invvar = 1./big_var
;neg_ls     = where(big_var le 0,n_neg)
;if (n_neg ne 0) then big_invvar(neg_ls) = 0
;
;;;; Delete large arrays to conserve memory
;big_ordermap = 0
;big_var      = 0
;
;;;; Calculate dispersion along each order
;print,'Calculating dispersion...'
;case arm of 
;   'UVB': pixarr = yarr
;   'VIS': pixarr = yarr
;   'NIR': pixarr = xarr
;endcase
;big_disp    = fltarr(xsz,ysz,n_exp)
;disp_porder = 3  ; Does not have to fit wave solution exactly
;for i=0,n_exp-1 do begin
;   exp_slit       = big_slit(*,*,i)
;   exp_wav        = big_wav(*,*,i)
;   exp_orderedge1 = big_orderinfo(i).left
;   exp_orderedge2 = big_orderinfo(i).right
;   exp_disp       = 0.*exp_wav
;   for j=0,n_orders-1 do begin
;      illum_ls = where(exp_slit ge exp_orderedge1(j) and $
;                       exp_slit le exp_orderedge2(j) and $
;                       exp_wav gt 0,n_illum)
;      pix = pixarr(illum_ls)
;      pfit = poly_fit(pix,exp_wav(illum_ls),disp_porder,yfit=wavfit)
;      illum_disp = dblarr(n_illum)
;      for k=1,disp_porder do illum_disp = illum_disp + k*pfit(k)*(pix^(k-1))
;      exp_disp(illum_ls) = illum_disp
;   endfor
;   big_disp(*,*,i) = exp_disp
;endfor
;big_disp = abs(big_disp)
;
;;;; Delete large arrays to conserve memory
;big_slit     = 0
;exp_disp     = 0
;exp_ordermap = 0
;exp_prob     = 0
;exp_slit     = 0
;exp_sub      = 0
;exp_var      = 0
;exp_wav      = 0
;mask         = 0
;neg_ls       = 0
;pixarr       = 0
;testframe    = 0
;xarr         = 0
;yarr         = 0

;;; Delete large arrays to conserve memory
orig_slit      = 0
reg_slit       = 0
big_ordermap   = 0
ILLUM_LS       = 0
ILLUM_PROB     = 0
ILLUM_SLIT     = 0
ILLUM_SUB      = 0
ILLUM_WAV      = 0
MATCH_E1       = 0
MATCH_F1       = 0
MATCH_FILT_F1  = 0
ORDER_CORR     = 0
ORDER_LS       = 0
ORDER_SLIT     = 0
ORDER_Y        = 0
RESP_MATCH     = 0
TEMP_SLIT      = 0
TESTFRAME      = 0
EXP_POS        = 0
EXP_ORDERMAP   = 0
EXP_PROB       = 0
EXP_SLIT       = 0
EXP_SUB        = 0
EXP_VAR        = 0
EXP_WAV        = 0
CORR_SLIT      = 0
MASK           = 0

;; Plot probability distribution, to check whether cutoffs are sensible
;plothist,big_prob(where(big_prob ge 0.01)),bin=0.01,xr=[0,1.5],$
;         title='Probability Distribution'
;oplot,replicate(0.05,2),[-1e6,1e6],color=250,linestyle=1
;oplot,replicate(1,2),[-1e6,1e6],color=250,linestyle=1

;;; Select pixels with signal from large arrays in order
;;; to conserve memory
min_prob = 0.05  ; This is an okay value, but the final spectrum is not
                 ; highly sensitive.
keep_ls  = where(big_prob ge min_prob,n_keep)
big_prob = (temporary(big_prob))(keep_ls)
big_sub  = (temporary(big_sub))(keep_ls)
big_var  = (temporary(big_var))(keep_ls)

;;; Calculate dispersion along each order
print,'Calculating dispersion...'
if (arm eq 'NIR') then pixarr = xarr else pixarr = yarr
big_disp    = fltarr(xsz,ysz,n_exp)
disp_porder = 3  ; Does not have to fit wave solution exactly
for i=0,n_exp-1 do begin
   exp_slit       = big_slit(*,*,i)
   exp_wav        = big_wav(*,*,i)
   exp_orderedge1 = big_orderinfo(i).left
   exp_orderedge2 = big_orderinfo(i).right
   exp_disp       = 0.*exp_wav
   for j=0,n_orders-1 do begin
      illum_ls = where(exp_slit ge exp_orderedge1(j) and $
                       exp_slit le exp_orderedge2(j) and $
                       exp_wav gt 0,n_illum)
      pix = pixarr(illum_ls)
      pfit = poly_fit(pix,exp_wav(illum_ls),disp_porder,yfit=wavfit)
      illum_disp = dblarr(n_illum)
      for k=1,disp_porder do illum_disp = illum_disp + k*pfit(k)*(pix^(k-1))
      exp_disp(illum_ls) = illum_disp
   endfor
   big_disp(*,*,i) = exp_disp
endfor

; Make sure dispersion is positive
if pfit(1) lt 0 then big_disp = -big_disp

;;; Select pixels with signal from large arrays in order
;;; to conserve memory
big_slit = (temporary(big_slit))(keep_ls)
big_wav  = (temporary(big_wav ))(keep_ls)
big_disp = (temporary(big_disp))(keep_ls)

;;; Inverse variance array
big_invvar = 1./big_var
neg_ls     = where(big_var le 0,n_neg)
if (n_neg ne 0) then big_invvar(neg_ls) = 0

;;; Delete large arrays to conserve memory
big_var      = 0
neg_ls       = 0

;;; Delete large arrays to conserve memory
big_slit     = 0
exp_disp     = 0
exp_ordermap = 0
exp_prob     = 0
exp_slit     = 0
exp_sub      = 0
exp_var      = 0
exp_wav      = 0
neg_ls       = 0
pixarr       = 0
testframe    = 0
xarr         = 0
yarr         = 0

;;; Run optimal extraction
if keyword_set(angstrom) then binning_mode = 'linear' else $
   binning_mode = 'loglin'
case binning_mode of
   'linear' : begin
                 wavebin = 0.5     ; Default in Angstroms
                 log_lin = 0
                 case arm of 
                    'UVB': lam0 = 1000.
                    'VIS': lam0 = 5336.8d ; To match pipeline reduction
                    'NIR': lam0 = 9940.0d
                 endcase
              end
   'loglin' : begin
                 wavebin = 8.0   ; Default in km/s
                 log_lin = 1
                 lam0    = 1000. ; OPT_EXTRACT default
              end
endcase
if keyword_set(bin) then wavebin=bin
case arm of
   'UVB': satlevel = 6.5e4
   'VIS': satlevel = 6.5e4
   'NIR': satlevel = 1e6
endcase
;min_prob = 0.05  ; This is an okay value, but the final spectrum is not
;                 ; highly sensitive.

; Full frame
print,'Performing optimal extraction...'
good_ls  = where(big_invvar gt 0 and big_wav gt 0 and $
                 big_disp gt 0 and big_disp lt 10 and $
                 big_prob gt min_prob and big_prob lt 1.0,n_good)
spec = opt_extract(big_sub(good_ls),big_wav(good_ls),big_invvar(good_ls),$
                   big_prob(good_ls),speclam=speclam,bin=wavebin,$
                   loglin=log_lin,lam0=lam0,maxcount=satlevel,$
                   specvar=specvar,disp=big_disp(good_ls),clip=5,$
                   modelcounts=modelcounts)

;; Order by order
;multi_spec = fltarr(10000,n_orders)
;multi_var  = fltarr(10000,n_orders)
;multi_wave = fltarr(10000,n_orders)
;for i=0,n_orders-1 do begin
;   order_num = example_orderinfo.ordernum(i)
;   good_ls   = where(big_invvar gt 0 and big_wav gt 0 and $
;                     abs(big_disp) gt 0 and abs(big_disp) lt 1 and $
;                     big_prob gt min_prob and $
;                     big_ordermap eq order_num)
;   temp_spec = opt_extract(big_sub(good_ls),big_wav(good_ls),$
;                           big_invvar(good_ls),big_prob(good_ls),$
;                           speclam=temp_speclam,bin=wavebin,$
;                           loglin=log_lin,maxcount=satlevel,$
;                           specvar=temp_specvar,disp=big_disp(good_ls),$
;                           clip=5)   
;   npix = n_elements(temp_spec)
;   multi_spec(0:npix-1,i) = temp_spec
;   multi_var(0:npix-1,i)  = temp_specvar
;   multi_wave(0:npix-1,i) = temp_speclam
;endfor
;; Remove columns without entries from any order
;keep_ls    = where(total(multi_wave,2) ne 0)
;multi_spec = multi_spec(keep_ls,*)
;multi_var  = multi_var(keep_ls,*)
;multi_wave = multi_wave(keep_ls,*)

; Rough conversion to flux / unit time.  Also apply mean response
; value
spec       = mean_resp_value * spec    / effective_texp
specvar    = (mean_resp_value^2) * specvar / (effective_texp^2)
;multi_spec = mean_resp_value * multi_spec / effective_texp
;multi_var  = mean_resp_value * multi_var  / (effective_texp^2)

;;; Write files

hspec = hdr
case binning_mode of
   'linear': begin
                sxaddpar,hspec,'CRVAL1',min(speclam)
                sxaddpar,hspec,'CDELT1',wavebin
                sxaddpar,hspec,'CD1_1',wavebin
                sxaddpar,hspec,'CRPIX1',1
                sxaddpar,hspec,'CTYPE1','LAMBDA'
                sxaddpar,hspec,'CTYPE2','PIXEL'
             end
   'loglin': begin
                c = 2.99792458d5
                sxaddpar,hspec,'CRVAL1',alog10(min(speclam))
                sxaddpar,hspec,'CDELT1',alog10(wavebin/c + 1)
                sxaddpar,hspec,'CD1_1',alog10(wavebin/c + 1)
                sxaddpar,hspec,'CRPIX1',1
                sxaddpar,hspec,'CTYPE1','LINEAR'
                sxaddpar,hspec,'CTYPE2','PIXEL'
             end
endcase
;sxaddpar,hspec,'CRVAL1',min(speclam)
;sxaddpar,hspec,'CDELT1',wavebin
;sxaddpar,hspec,'CD1_1',wavebin
;sxaddpar,hspec,'CRPIX1',1
;sxaddpar,hspec,'CTYPE1','LAMBDA'
;sxaddpar,hspec,'CTYPE2','PIXEL'

spec    = float(spec)
specerr = float(sqrt(specvar))
writefits,outfile+'.fits',spec,hspec
writefits,outfile+'e.fits',specerr,hspec

;multi_spec = float(multi_spec)
;multi_err  = float(sqrt(multi_var))
;writefits,outfile+'_multi_flux.fits',multi_spec
;writefits,outfile+'_multi_err.fits',multi_err
;writefits,outfile+'_multi_wave.fits',multi_wave


;;;; TESTING
;writefits,outfile+'_scalefn.fits',scale_fn

return
end
