Pro xshooter_reduce,object,wavefile,slitfile,orderlist,arm=arm,ybin=ybin,$
                    skyref=skyref,skyframe=skyframe,flatfield=flatfield,$
                    flatvar=flatvar,darkframe=darkframe,darkvar=darkvar,$
                    nod=nod,output=output,findrange=findrange,order=order,$
                    fit=fit,fiducial=fiducial,object_xy=object_xy,$
                    fid_xy=fid_xy,fid_fwhm=fid_fwhm,fid_mask=fid_mask,$
                    noobject=noobject,scattered=scattered,vshift=vshift,$
                    altvar=altvar,_extra=ex

;+-------------------------------------------------------------------
;
; XSHOOTER_REDUCE     12/2009
;
; Reduce X-Shooter data using mutliple calls to longslit_reduce.
;
; INPUTS
;         object      - Object FITS filename
;         wavefile    - FITS filename of calibrated wavelength map
;                       (see xshooter_wavegrid.pro and
;                       xshooter_wavecalib.pro).  If these are in
;                       vacuum (air), then output wavelengths will be
;                       in vacuum (air) heliocentric.  Corrections
;                       made using skylines if skyref keyword is set.
;         slitfile    - FITS filename of slit position map (see
;                       xshooter_slitgrid.pro)
;         orderlist   - Name of 5-column ASCII file containing, for
;                       each order, the the CCD name ('XSHOOT'),
;                       the order number (which is arbitrary, but
;                       should be consecutive integers), and should
;                       not be 0), the minimum slit position covered
;                       by the order, and the maximum slit position
;                       covered by the order.  Format is: [CCD, order
;                       #, center, left edge, right edge].  Output by
;                       xshooter_slitgrid.pro.
;
; KEYWORDS
;         arm         - String specifying which X-Shooter arm: 'UVB',
;                       'VIS', or 'NIR'.  Must be set.
;         ybin         - Factor by which to bin the arc exposure in the
;                        y-direction.  For example, if science exposures
;                        were taken in 1x2 binning, and arcs were taken
;                        in 1x1, then set ybin=2.
;         skyref      - String, FITS filename of reference sky exposure
;                       for which the slit positions and wavelengths
;                       have been calibrated.  Should be in a similar
;                       setup to the object being reduced (shifts ~< 10
;                       pixels).  Skylines in the reference frame
;                       will be matched to skylines in the frame
;                       being reduced in order to register the slit
;                       positions and wavelengths to the current setup.
;         skyframe    - String.  If set, the skylines in skyref will
;                       be matched to the skylines in this exposure,
;                       rather than to those in the frame being
;                       reduced.  This exposure should therefore be
;                       taken in exdactly the same setup as the
;                       exposure being reduced.  Useful for standard
;                       star exposures that may not have pronounced
;                       skylines.
;         flatfield   - FITS filename of flat field array
;         flatvar     - FITS filename of flat field variance array
;         darkframe   - FITS filename of dark exposure.  This is
;                       typically a comined dark from several darks.
;                       Recommended for NIR exposures.
;         darkvar     - FITS filename, variance in the dark frame.
;                       Recommended for NIR exposures.  *** Variance
;                       in the dark frame not current used.  Use to
;                       identify bad pixels? ***
;         nod         - String, FITS filename of nodded exposure to use
;                       for primary sky subtraction.  This passes to 
;                       the longslit_reduce keyword skyref, which is 
;                       different from the skyref used here.  Default
;                       is no nod subtraction.
;         output      - Root name for output files.  Default is
;                       the input object filename.
;         findrange   - Two-elements array giving the min and max x-values
;                       over which to locate the object in the fiducial
;                       order.  DISABLED.
;         order       - 1D array specifying the orders to be reduced.
;                       Numbers correspond to the entry number in the
;                       input orderlist (1 being the first order listed
;                       in the file).  Must include the fiducial order.
;         fit         - 1D array specifying the orders in which the object
;                       should be indepently located and traced, and its
;                       FWHM measured.  These orders should have significant
;                       flux over most of the length of the array.  Other
;                       orders will have their object position, trace, and
;                       FWHM extrapolated from the fitted orders, which
;                       is useful for orders with little or no flux.  Default
;                       is to independently fit all orders.  Must include
;                       the fiducial order.
;         fiducial    - Scalar, order number to be used as the
;                       fiducial order (the order number is determined
;                       by the order number entry in the input
;                       orderlist).  This order will be reduced first
;                       and willl be used to measure the initial
;                       profile FWHM (for Gaussian profiles).  The
;                       default is to use one of the middle orders.
;                       However, it is recommended to use the order
;                       that contains the most flux.
;         object_xy   - Two-element array giving the [x,y] pixel
;                       coordinates of a point at the center of the
;                       object trace in the the fiducial order.
;                       Useful when more than one object appears on
;                       the slit.  If set, the program uses this as an
;                       initial estimate but refines the position
;                       using a Gaussian fit.  Default is to automatically
;                       search for the brightest object in the slit.
;         fid_xy      - Two-element array giving the [x,y] pixel
;                       coordinates of a point at the center of the
;                       object trace in the the fiducial order.  Can
;                       be set along with fid_fwhm instead of
;                       attempting to automatically fit the object
;                       profile.  This keyword differs from object_xy
;                       in that the literal values are used for the
;                       object position, with no refinement.
;         fid_fwhm    - Scalar, spatial Gaussian FWHM, in pixels,
;                       of the object in the fiducial order.  Can be set
;                       along with fid_xy instead of attempting to
;                       automatically fit the object profile.
;         fid_mask    - Three-element array specifying a region on the
;                       fiducial order to mask.  Useful when multiple
;                       objects appear on the slit.  For the UVB and
;                       VIS arms, the array elements describe a
;                       horizontal region with [xmin,xmax,y], while
;                       for the NIR they describe a vertical region
;                       with [ymin,ymax,x].  The regions in other
;                       orders corresponding to these slit positions
;                       will also be masked.  Note that it is not
;                       necessary to set this keyword for
;                       single-object exposures, provided the object
;                       can be automatically located.
;         noobject    - Reduce without extracting a 1D spectrum, or
;                       masking object while fitting sky
;         scattered   - Subtract scattered light by fitting a 2D polynomial
;                       to the inter-order counts.
;         vshift      - Scalar, additional velocity shift in km/s to apply to 
;                       wavelengths
;         altvar      - If set, estimate 2D variance based on scatter
;                       about sky fit (same keyword in
;                       longslit_reduce).  Default is to compute
;                       variance based on amplitude of sky fit and
;                       extracted object spectrum, which can produce
;                       biases when combining multiple exposures.
;         _extra      - Additional keywords accepted by
;                       longslit_reduce, e.g., sprofile, maxiter,
;                       noshow, and noskysub.  Note that some keywords
;                       may not work due to their use in the program
;                       body below.
;
; OUTPUTS
;         Outputs written to output_sub.fits, output_spec.fits, etc.
;
; Side effects: If there is a journal file currently open, the program
;               will close it and open <output>.log.
;
; HISTORY
;         Adapted from MAGE_REDUCE 12/9/2009 GDB
;         FID_XY and FID_FWHM keywords added 2/26/2012 GDB
;         OBJECT_XY, FID_MASK keywords added 11/19/2012 GDB
;         Allowed for skyref and skyframe with different binning
;            from science frame 12/3/2012 GDB
;         Changed way in which trace offset and spatial profile
;            are estimated for orders not independently fit.
;            12/6/2012 GDB
;         VSHIFT keyword added 4/12/2013 GDB
;         Included secondary slit illumination fn correction 
;            2/10/2014 GDB
;         ALTVAR keyword added 5/12/2014 GDB
;--------------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:'
   print,' xshooter_reduce,object,wavefile,slitfile,orderlist,arm=arm,'
   print,'        ybin=ybin,skyref=skyref,skyframe=skyframe,'
   print,'        flatfield=flatfield,flatvar=flatvar,'
   print,'        darkframe=darkframe,darkvar=darkvar,'
   print,'        output=output,findrange=findrange,order=order,fit=fit,'
   print,'        fiducial=fiducial,object_xy=object_xy,fid_xy=fid_xy,'
   print,'        fid_fwhm=fid_fwhm,fid_mask=fid_mask,/noobject,'
   print,'        /scattered,vshift=vshift,/altvar,_extra=ex'
   return
endif

; Specify which arm
if not(keyword_set(arm)) then begin
   print,'Must specify which X-Shooter arm'
   return
endif
; Allow upper or lower case
case arm of
   'uvb': arm = 'UVB'
   'UVB': arm = 'UVB'
   'vis': arm = 'VIS'
   'VIS': arm = 'VIS'
   'nir': arm = 'NIR'
   'NIR': arm = 'NIR'
   else : begin
             print,'X-Shooter arm not recognized'
             return
          end
endcase

; Compile longslit reduce with the apporpriate include file
program_dir = '~/pro/'
spawn,'rm -f '+program_dir+'longslit_reduce.inc'  ; Should be a symbolic link
case arm of
   'UVB': tag = 'uvb'
   'VIS': tag = 'vis'
   'NIR': tag = 'nir'
endcase
spawn,'ln -s '+program_dir+'longslit_reduce_xshooter_'+tag+'.pro '+$
      program_dir+'longslit_reduce.inc'
resolve_routine,'longslit_reduce',/compile_full_file

; Root name for output
if keyword_set(output) then root = output else root = filename(object,/noexten)

; Start log.  If there is a journal already running, close it and start a
; new one.
if (!journal ne 0) then journal
journal,root+'.log'

print,''
print,'Reducing '+object
print,''

; Read in FITS files
obj  = readfits(object,hobj)
wave = readfits(wavefile,hwave)
slit = readfits(slitfile,hslit)
if keyword_set(darkframe) then dark = readfits(darkframe) $
   else dark = 0.
if keyword_set(darkvar) then dark_var = readfits(darkvar) $
   else dark_var = 0.

; Shift wavelengths?
if keyword_set(vshift) then begin
   c    = 2.99792458d5
   wave = wave * (1+vshift/c)
endif

; Method for computing variance?
if keyword_set(altvar) then begin
   ; Estiamte variance based on scatter around sky fit
   alt_var = 1
endif else begin
   ; Formal estimate based on amplitude of sky fit and
   ; object spectrum
   alt_var = 0
endelse

; Bias/Dark subtract
obj = xshooter_subbias(obj,arm=arm)  ; No effect for NIR
obj = obj - dark

; Exposure for nod subtraction (optional)
if keyword_set(nod) then begin
   nod_sky = readfits(nod)
   nod_sky = xshooter_subbias(nod_sky,arm=arm)
   nod_sky = nod_sky - dark
endif else begin
   ; Default is no nodded sky subtraction
   nod_sky = 0
endelse

; Bin exposure?
if keyword_set(ybin) then begin
   if (ybin ne 1) then begin
      orig_sz  = size(obj)
      orig_xsz = orig_sz(1)
      orig_ysz = orig_sz(2)
      obj = rebin(1.0*obj,orig_xsz,orig_ysz/2.)
   endif
endif

; Get size of individual CCD arrays
sz   = size(obj)
xsz  = sz(1)
ysz  = sz(2)
xarr = fltarr(xsz,ysz)
yarr = fltarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i

; Illuminated regions
mask = lonarr(xsz,ysz)
; No masking for UVB and VIS arms.  Orders are well away from
; pre/overscan regions
if (arm eq 'NIR') then begin
   ; Red edge x(y) polynomial
   r_terms = [1459.7389, 0.65775340, -0.00090907266, 3.2113316e-06]
   r_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(r_terms)-1 do r_yfit = r_yfit + r_terms(i)*(yarr^i)
   r_mask_ls = where(xarr gt r_yfit)
   ; Blue edge 
   b_terms = [433.63313, -0.30479697, 0.00011715153, -1.1558709e-06]
   b_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(b_terms)-1 do b_yfit = b_yfit + b_terms(i)*(yarr^i)
   b_mask_ls = where(xarr lt b_yfit)
   mask(r_mask_ls)     = 1
   mask(b_mask_ls)     = 1
   mask(0:3,*)         = 1
   mask(xsz-4:xsz-1,*) = 1
   mask(*,0:3)         = 1
endif

; Determine gain and read noise
case arm of
   'UVB': begin
             gn  = 0.67   ; 100k/1pt/hg clock
             rdn = 2.6
          end
   'VIS': begin
             gn  = 0.64   ; 100k/1pt/hg clock
             rdn = 3.2
          end
   'NIR': begin
             gn  = 2.12
             ; # readouts?
             texp = sxpar(hobj,'EXPTIME')
             if (texp lt 300) then rdn = 25. else rdn = 8.0
          end
endcase

;;; Register wavelength and slit position grids to current exposure
;;; using skylines.  Flag pixels that are not calibrated.
if not(keyword_set(skyref)) then begin
   print,'Skipping registering slit and wavelength arrays.'
endif else begin
   print,'Registering slit and wavelength arrays...'
   sky_ref = xshooter_subbias(readfits(skyref),arm=arm)
   ; If the science frame is binned, then make sure the
   ; sky reference frame is binned before registering.
   sky_ref_sz   = size(sky_ref)
   sky_ref_xsz  = sky_ref_sz(1)
   sky_ref_ysz  = sky_ref_sz(2)
   sky_ref_xbin = round(sky_ref_xsz / float(xsz))
   sky_ref_ybin = round(sky_ref_ysz / float(ysz))
   if (sky_ref_xbin ne 1 or sky_ref_ybin ne 1) then begin
      print,'Re-binning reference frame before registering'
      sky_ref = rebin(1.0*sky_ref,xsz,ysz)
   endif
   if keyword_set(skyframe) then begin
      ; Optional - Locate skylines in a separate exposure in the same
      ; setup
      print,'Registering to '+skyframe
      sky_frame = xshooter_subbias(readfits(skyframe),arm=arm)
      ; If the science frame is binned, then make sure the sky frame
      ; is binned before registering.
      sky_frame_sz   = size(sky_frame)
      sky_frame_xsz  = sky_frame_sz(1)
      sky_frame_ysz  = sky_frame_sz(2)
      sky_frame_xbin = round(sky_frame_xsz / float(xsz))
      sky_frame_ybin = round(sky_frame_ysz / float(ysz))
      if (sky_frame_xbin ne 1 or sky_frame_ybin ne 1) then begin
         print,'Re-binning sky frame before registering'
         sky_frame = rebin(1.0*sky_frame,xsz,ysz)
      endif
      xshooter_register,sky_ref,sky_frame,p,q,arm=arm
   endif else begin
      xshooter_register,sky_ref,obj,p,q,arm=arm
   endelse
   print,'POLY_2D terms: '
   print,p,q
   ; Mask uncalibrated pixels with NaN values.  This will tend to lose
   ; a few pixels at the edges of the array.  It may be possible to 
   ; interpolate the wavelength solution instead, and fill in the missing
   ; values, but this would only be worthwhile if the shift was large
   ; (> 10 pixels in wavelength).
   ; Keep track of uncalibrated pixels
   calibgrid = 1 + 0.*slit
   notcalib_ls = where(slit lt -999 or wave lt 0,n_notcalib)
   if (n_notcalib ne 0) then calibgrid(notcalib_ls) = 0.0
   newcalibgrid = poly_2d(calibgrid,p,q,1,missing=0.0)
   nan   = !values.f_nan
   wave = poly_2d(wave,p,q,1,missing=nan)
   slit = poly_2d(slit,p,q,1,missing=nan)
   new_notcalib_ls = where(newcalibgrid lt 1,n_new_notcalib)
   if (n_new_notcalib ne 0) then begin
      wave(new_notcalib_ls) = nan
      slit(new_notcalib_ls) = nan
   endif
   ;; Suppress pixels that were interpolated from uncalibrated pixels,
   ;; which should have been flagged.
   ;wave_ls = where(wave lt 0,n_wave)
   ;if (n_wave ne 0) then wave(wave_ls) = nan
   ;slit_ls = where(slit lt 0,n_slit)
   ;if (n_slit ne 0) then slit(slit_ls) = nan
endelse

; Read in flat files.  Assume bias has already been subtracted.
; If no flats are supplied then create fake flats for flagging
; masked pixels.
if keyword_set(flatfield) then flat = readfits(flatfield,hflat) $
   else flat = 1 + fltarr(xsz,ysz)
if keyword_set(flatvar) then flat_var = readfits(flatvar,hflatvar) $
   else flat_var = fltarr(xsz,ysz)

; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo = read_ascii(orderlist,template=order_templ)
; Make sure all order numbers are unique
if (n_elements(orderinfo.ordernum) ne $
    n_elements(uniq(orderinfo.ordernum))) then begin
   print,'Error - All order numbers must be unique.'
   return
endif
; Sort orders in order of increasing slit position
sort_ls     = sort(orderinfo.center)
ccdname     = orderinfo.ccd(sort_ls)
ordernum    = orderinfo.ordernum(sort_ls)
ordercenter = orderinfo.center(sort_ls)
orderedge1  = orderinfo.left(sort_ls)
orderedge2  = orderinfo.right(sort_ls)
n_orders    = n_elements(ordercenter)

; Flag whether order number increases with wavelength.  Note that the
; "order numbers" here are not generally the true physical order
; numbers.
case arm of
   'UVB': red_dir = -1   ; Wavelength increases towards -x
   'VIS': red_dir = +1   ; Wavelength increases towards +x
   'NIR': red_dir = +1   ; Wavelength increases towards +y
endcase

; Optional - Specify the orders to reduce.
if keyword_set(order) then begin
   n_reduce    = n_elements(order)
   reduce_ls   = intarr(n_reduce)
   for i=0,n_reduce-1 do reduce_ls(i) = where(ordernum eq order(i))
   ordercenter = ordercenter(reduce_ls)
   ccdname     = ccdname(reduce_ls)
   ordernum    = ordernum(reduce_ls)
   orderedge1  = orderedge1(reduce_ls)
   orderedge2  = orderedge2(reduce_ls)
   n_orders    = n_elements(ordercenter)
endif

; Optional - Specify orders the orders for which to independently locate
; the object and measure its FWHM.  Default is to independently fit
; all orders.
do_independent_fit = lonarr(n_orders)
if keyword_set(fit) then begin
   for i=0,n_elements(fit)-1 do begin
      j = where(ordernum eq fit(i))
      do_independent_fit(j) = 1
   endfor
endif else begin
   do_independent_fit(*) = 1
endelse

; Keep track of orders for which independent tracing and fitting
; has been done.
independent_fit_made = lonarr(n_orders)

; Optional - Specify the order to be reduced first.  Default is
; to start with an order in the middle.
if keyword_set(fiducial) then begin
   fid_index = where(ordernum eq fiducial,n_fid)
   fid_index = fid_index(0)
   if (n_fid eq 0) then begin
      print,'Error - Fiducial order must be included in the orders to reduce.'
      return
   endif
endif else begin
   fid_index = floor(n_orders/2.)
   fiducial  = ordernum(fid_index)
endelse

; Flag bad pixels to be masked by longslit_reduce.
; (None)

; Optional - Specify a region along the fiducial order to be masked
; when performing the sky subtraction.  The corresponding regions
; in other orders will also be masked.
if keyword_set(fid_mask) then begin
   if (arm eq 'NIR') then begin
      ymin     = fid_mask(0)
      ymax     = fid_mask(1)
      xmin     = fid_mask(2)
      xmax     = fid_mask(2)
   endif else begin
      xmin     = fid_mask(0)
      xmax     = fid_mask(1)
      ymin     = fid_mask(2)
      ymax     = fid_mask(2)
   endelse
   ; Min/max slit positins on fiducial order to mask
   slit_cut = slit(xmin:xmax,ymin:ymax)
   fid_mask_slit_min = min(slit_cut)
   fid_mask_slit_max = max(slit_cut)
   ; Min/max offset from order center
   mask_offset_lo = fid_mask_slit_min - ordercenter(fid_index)
   mask_offset_hi = fid_mask_slit_max - ordercenter(fid_index)
   apply_mask     = 1
endif else begin
   apply_mask     = 0
endelse

; Optional - Set pixel range in spectral direction over which to locate object.
; Note: nominal_find_range not currently being used.
if keyword_set(findrange) then begin
   nominal_find_range = findrange
   print,'findrange keyword disabled'
   return
endif else begin
   case arm of
      'UVB' : nominal_find_range = [900,1100]
      'VIS' : nominal_find_range = [900,1100]
      'NIR' : nominal_find_range = [900,1100]
   endcase
endelse

; Optional - Subtract scattered light
if keyword_set(scattered) then begin
   print,'Subtracting scattered light...'
   ; Select regions in between orders
   maskval    = -100
   interorder = maskval + 0.*obj
   for i=0,n_orders-2 do begin
      inter_ls = where(slit ge orderedge2(i)+5   and $
                       slit le orderedge1(i+1)-5,n_inter)
      if (n_inter ne 0) then interorder(inter_ls) = obj(inter_ls)
   endfor
   ; Mask non-illuminated regions
   nonillum_ls = where(mask eq 1)
   if (n_nonillum ne 0) then interorder(nonillum_ls) = maskval
   ; Fit 2D polynomial
   fit_ls = where(interorder ne maskval,n_fit)
   if (n_fit eq 0) then begin
      print,'XSHOOTER_REDUCE: No good pixels for fitting scattered light.'
      return
   endif
   interorder_data = fltarr(3,n_fit)
   interorder_data(0:1,*) = array_indices(interorder,fit_ls)
   interorder_data(2,*)   = interorder(fit_ls)
   fit_order = 4
   temp_fit  = sfit(interorder_data,fit_order,/irregular,kx=fitterms)
   ; Evaluate polynomial over entire array and subtract
   scattered_light = 0.*obj
   for i=0,fit_order do begin
      for j=0,fit_order do begin
         scattered_light = scattered_light + fitterms(j,i)*(xarr^i)*(yarr^j)
      endfor
   endfor
   if (n_nonillum ne 0) then scattered_light(nonillum_ls) = 0.
   obj = obj - scattered_light
endif

; Suppress masked pixels
masked_ls = where(mask ne 0,n_masked)
if (n_masked ne 0) then flat_var(masked_ls) = -1

; Initalize large spectrum, error, and wavelength
; arrays in which to store the individual reduced orders.
big_wave = dblarr(10000,n_orders)
big_spec = dblarr(10000,n_orders)
big_err  = dblarr(10000,n_orders)

; Keep track of the median wavelength, Gaussian profile FWHM, slit
; position, offset from the slit center for each order.
; X-Shooter has an ADC for the VIS and UVB arms, so assume object
; will be at constant slit position within each order.
medwave_list   = dblarr(n_orders)
fwhm_list      = dblarr(n_orders)
position_list  = dblarr(n_orders)
offset_list    = dblarr(n_orders)
;trace_coeffs   = dblarr(5,n_orders)

; Initialize fits to FWHM, position, offset from the slit center,
; and linear trace term
fwhm_fit     = dblarr(n_orders)
position_fit = dblarr(n_orders)
offset_fit   = dblarr(n_orders)
;slope_fit    = dblarr(n_orders)

; Initialize reduced arrays
nosky    = fltarr(xsz,ysz)
noskyvar = fltarr(xsz,ysz)
wavefit  = dblarr(xsz,ysz)
skymodel = fltarr(xsz,ysz)

; Get an approximate median wavelength for each order
approx_medwave = fltarr(n_orders)
for i=0,n_orders-1 do begin
   this_order   = where(slit ge orderedge1(i) and $
                        slit le orderedge2(i) and $
                        wave gt 100,n_this_order)
   approx_medwave(i) = median(wave(this_order))
endfor 

; Method for estimating trace parameters for orders not independently
; fit has been changed (see below).  The following parameterization
; for the higher-order fit terms to the spatial FWHM has been retained
; for reference (cf. other echelle reduction routines).
;
;; OLD -- High order terms for trace parameters as a function of
;; wavelength.  These were measured from fits to standard stars, and
;; should be at least be better than assuming the FWHM is constant with
;; wavelenth.  For now, ignore higher order terms for the offset from
;; slit center and the slope.
;case arm of 
;   'UVB': begin 
;             w0      = 4172.44
;             fwhm_a1 = -0.000817110
;             fwhm_a2 = 2.93374e-07
;          end
;   'VIS': begin 
;             w0      = 7067.98
;             fwhm_a1 = -0.000378055
;             fwhm_a2 = 1.33297e-07
;          end
;   'NIR': begin 
;             w0      = 14513.0
;             fwhm_a1 = -0.000130330
;             fwhm_a2 = 6.80993e-09
;          end
;endcase
;fwhm_hot     = fwhm_a1*(approx_medwave-w0) + fwhm_a2*((approx_medwave-w0)^2)
;offset_hot   = 0.0
;;slope_hot    = 0.0


; Notes on trace and spatial profile parameters.  The following were
; determined from standard star spectra taken in a variety of seeing
; conditions and airmasses.  The trace offset and slope were measured
; from telluric standards taken with the ADCs disabled.  These notes
; should therefore represent more stringent requirement than those for
; nominal X-Shooter observations.
;
;    FWHM -- For each arm, the FWHM as a fn of order number has a
;            self-similar function.  The FWHM for orders not
;            independently fit can therefore be estimated by scaling
;            these functions using the values from fitted orders.
;            This should work reasonably well in the UVB and NIR arms.
;            There is more variation in the shape of FWHM vs position
;            in the VIS; however, the shape can be reduced to a linear
;            fit.  This is somewhat more robust, although there may
;            still be ~10-20% errors in FWHM.  See
;            87/scripts/measure_std_fwhm.scr.
;
;    Offset -- For each arm, the offset from the nominal order center
;              as a function of absolute slit position (i.e., x or y)
;              is roughly linear.  The offset is slightly more linear
;              in absolute slit position than in order number.  Note
;              that there should be no drift (change in offset between
;              orders) when the ADC is engaged.  See
;              87/scripts/measure_trace_offsets.scr.
;
;   Slope -- For the VIS and NIR the slope of the object trace is
;            relatively small, even in the bluest orders.  It can be
;            significant in the UVB, however (up to an end-to-end
;            change of ~4 pixels in the bluest orders at airmass~2).
;            For now, will not attempt to trace objects and measure
;            slope.  However, it may be necessary to do so for very
;            blue spectra taken at high airmass without the ADC.  See
;            measure_trace_slope.scr.

; Tabulated shape of FWHM as a fn of order number
case arm of
   'UVB': begin
             tab_ordernum      = 1 + lindgen(12)
             tab_unscaled_fwhm = [0.903, 0.894, 0.909, 0.935, 0.954,$
                                  0.980, 0.999, 1.020, 1.047, 1.079,$
                                  1.107, 1.149]
          end
   'VIS': begin
             tab_ordernum      = 1 + lindgen(15)
             tab_unscaled_fwhm = [1.207, 1.170, 1.130, 1.102, 1.075,$ 
                                  1.043, 1.016, 0.988, 0.969, 0.953,$ 
                                  0.939, 0.938, 0.947, 0.968, 0.985]
          end
   'NIR': begin
             tab_ordernum      = 1 + lindgen(15)
             tab_unscaled_fwhm = [1.128, 1.105, 1.082, 1.060, 1.039,$
                                  1.018, 0.999, 0.980, 0.961, 0.943,$ 
                                  0.925, 0.906, 0.888, 0.868, 0.848]
          end
endcase
; Only include those orders being reduced
unscaled_fwhm_fit = interpol(tab_unscaled_fwhm,tab_ordernum,ordernum)

; Loop through orders.  Start with fiducial order
index_list = [fid_index]

; Next do indepent orders blueward of the fiducial order, going
; red to blue
if (red_dir eq 1) then begin
   blue_indep_ls = where(ordernum lt fiducial and do_independent_fit,$
                         n_blue_indep)
   if (n_blue_indep ne 0) then index_list = [index_list,reverse(blue_indep_ls)]
endif else begin
   blue_indep_ls = where(ordernum gt fiducial and do_independent_fit,$
                         n_blue_indep)
   if (n_blue_indep ne 0) then index_list = [index_list,blue_indep_ls]
endelse

; Next do indepent orders redward of the fiducial order
if (red_dir eq 1) then begin
   red_indep_ls = where(ordernum gt fiducial and do_independent_fit,n_red_indep)
   if (n_red_indep ne 0) then index_list = [index_list,red_indep_ls]
endif else begin
   red_indep_ls = where(ordernum lt fiducial and do_independent_fit,n_red_indep)
   if (n_red_indep ne 0) then index_list = [index_list,reverse(red_indep_ls)]
endelse

; Finally, do non-independent orders (order in which these are done
; does not matter)
non_indep_ls = where(do_independent_fit eq 0,n_non_indep)
if (n_non_indep ne 0) then index_list = [index_list,non_indep_ls]

; Record heliocentric velocity.  Default is no heliocentric velocity
; correction made.
vhelio_applied = 0
vhelio         = 0

; Temporary file name
temp_file = 'temp_'+filename(root,/noexten)

for i=0,n_orders-1 do begin

   index = index_list(i)

   print,' '
   print,'Reducing '+object+', order '+strtrim(ordernum(index),2)
   print,' '

   ; Extrapolate fit and trace parameters from orders that have been
   ; independently fit.  For the FWHM, this is done by scaling
   ; tabulated ratios of the FWHM in different orders.  For the
   ; offset, a linear fit with absolute position on the detector is
   ; made.  The offset should be easier and more reliable to fit than
   ; the raw position.  If there are very few orders for which the
   ; offset has been measured independently, then use a constant
   ; value.  This is equivalent to doing an offset from the standard
   ; star from which the slit grid was created.  In that case, the
   ; extrapolated values may only be reasonable out to a few orders.
   ; Intra-order slopes are currently ignored (see above).
   indep_ls     = where(independent_fit_made eq 1,n_indep)
   n_for_linear = 6
   if (n_indep ge n_for_linear) then begin
      offset_order   = 1
      ;slope_order    = 1
   endif else begin
      offset_order   = 0
      ;slope_order    = 0
   endelse
   if (n_indep ge 1) then begin
      ; Scale FWHM values
      fwhm_scale = mean(fwhm_list(indep_ls)/unscaled_fwhm_fit(indep_ls))
      fwhm_fit   = unscaled_fwhm_fit * fwhm_scale
      ; Fit offset, slope, and (possibly) FWHM values
      if (n_indep ge n_for_linear) then begin
         if (arm eq 'VIS') then begin
            ; Reduce FWHM vs position to a linear fit by parametrizing
            ; quadratic coeff in terms of the linear coeff
            fwhm_hot  = (9.0e-10)*((ordercenter-1200)^2)
            net_fwhm  = fwhm_list - fwhm_hot
            b0        = 1.28e-6
            b1        = 3.29e-4
            tmp_x     = ordercenter - 1200
            tmp_y     = net_fwhm
            fit_x     = tmp_x + b1*(tmp_x^2)
            fit_y     = tmp_y - b0*(tmp_x^2)
            pfit_fwhm = poly_fit(fit_x(indep_ls),fit_y(indep_ls),1)
            a0        = pfit_fwhm(0)
            a1        = pfit_fwhm(1)
            a2        = b0 + b1*a1
            net_fwhm_fit  = a0 + a1*tmp_x + a2*(tmp_x^2)
            fwhm_fit      = net_fwhm_fit + fwhm_hot
         endif
         pfit_offset = robust_poly_fit(ordercenter(indep_ls),$
                                       offset_list(indep_ls),$
                                       offset_order)
         ;pfit_slope  = robust_poly_fit(ordercenter(indep_ls),$
         ;                              trace_coeffs(1,indep_ls),$
         ;                              slope_order)
      endif else begin
         pfit_offset = poly_fit(ordercenter(indep_ls),offset_list(indep_ls),$
                                  offset_order)
         ;pfit_slope = poly_fit(ordercenter(indep_ls),trace_coeffs(1,indep_ls),$
         ;                      slope_order)  
      endelse
      ; Reconstruct total fitted values
      offset_fit = fltarr(n_orders)
      for k=0,offset_order do $
         offset_fit = offset_fit + pfit_offset(k)*(ordercenter^k)
      ;slope_fit = fltarr(n_orders)
      ;for k=0,slope_order do $
      ;   slope_fit = slope_fit + pfit_slope(k)*(ordercenter^k)
   endif else begin
      fwhm_fit     = fltarr(n_orders)
      offset_fit   = fltarr(n_orders)
      ;slope_fit    = fltarr(n_orders)
   endelse
   ; Compute actual slit positions
   position_fit = ordercenter + offset_fit

   if not(keyword_set(noobject)) then begin
      ; For the first order, start with no estimates for the object
      ; profile FWHM.  For subsequent orders that are being independently
      ; traced and fit, start with the fitted FWHM estimate.  For orders
      ; that are not to be independently fit, use extrapolated values for
      ; the object position, object FWHM, and trace slope from orders for
      ; which independent fits have been made.
      if (index eq fid_index) then begin
         if keyword_set(object_xy) then begin
            slitpos = slit(object_xy(0),object_xy(1))
         endif else begin
            slitpos = 0   ; Return fitted slit position
         endelse
         fwhm    = 0   ; FWHM estimate
         ;slope   = 0
         force   = 0
         independent_fit_made(index) = 1
         ; Optional - User-specified values
         if (keyword_set(fid_xy) or keyword_set(fid_fwhm)) then begin
            if (keyword_set(fid_xy) and keyword_set(fid_fwhm)) then begin
               slitpos   = slit(fid_xy(0),fid_xy(1))
               fwhm      = 1.*fid_fwhm
               force     = 1
            endif else begin
               print,'Must specify both fid_xy and fid_fwhm (or neither)'
               return
            endelse
         endif
      endif else begin
         if (do_independent_fit(index) eq 1) then begin
            ;slitpos = 0
            slitpos = position_fit(index)
            fwhm    = fwhm_fit(index)
            ;slope   = 0
            force   = 0
            independent_fit_made(index) = 1
         endif else begin
            ; Use fitted values
            slitpos = position_fit(index)
            fwhm    = fwhm_fit(index)
            ;slope   = slope_fit(index)
            force   = 1
            print,'Using fitted values: Position = '+strtrim(slitpos,2)
            print,'                     FWHM     = '+strtrim(fwhm,2)
            ;print,'                     Slope    = '+strtrim(slope,2)
            print,' '
         endelse
      endelse
      no_object = 0
   endif else begin
      ; Do not mask object while fitting sky or extract 1D spectrum
      slitpos = 0
      fwhm    = 0
      force   = 0
      independent_fit_made(index) = 0
      no_object = 1
   endelse

   ;; If using an estimate for the slope of the object trace, adjust
   ;; slit grid before passing to longslit_reduce.  If not, trace
   ;; the object.
   ;; ---> This is handled better in esi_reduce.
   ;temp_slit = slit
   ;if not(keyword_set(noobject)) then begin
   ;   if (slope ne 0) then begin
   ;      slit_correction = slope*findgen(xsz)
   ;      for j=0,xsz-1 do temp_slit(j,*) = slit(j,*) - slit_correction(j)
   ;      do_trace = 0
   ;   endif else begin
   ;      do_trace = 1
   ;   endelse
   ;endif else begin
   ;   do_trace = 0
   ;endelse

   ; VIS and UVB have an ADC, so assume object trace will follow a
   ; path path of constant slit position -- i.e., no tracing done.
   ; Even without the ADC, the object trace will be at a fairly
   ; constant (to within ~1 pixel) slit position within each order
   ; in the VIS and NIR arms.  However, if the ADC is disabled
   ; or not working, there may be significant intra-order drift
   ; in the UVB arm at high airmasses.  In that case, may need
   ; to perform tracing.
   do_trace = -1
   temp_slit = slit

   ; Note: If the trace_obj keyword is set in the call to xshooter_reduce,
   ; then this will be passed to longslit_reduce if do_trace=0 (can
   ; be used to prevent longslit_reduce from tracing object).

   ; Pass header to longslit_reduce
   temp_hdr = hobj

   ; Set approximate slit positions
   temp_ordercenter = ordercenter(index)
   temp_slitmin     = orderedge1(index)
   temp_slitmax     = orderedge2(index)

   ; Mask a specified region along the slit?
   if apply_mask then $
      objmask = ordercenter(index) + [mask_offset_lo,mask_offset_hi] else $
      objmask = 0

   ; Set break point spacing according to approximate dispersion
   this_order   = where(slit ge temp_slitmin and slit le temp_slitmax and $
                        wave gt 100,n_this_order)
   if (arm eq 'UVB' or arm eq 'VIS') then begin
      rough_disp   = (max(wave(this_order))-min(wave(this_order))) / $
                     (max(yarr(this_order))-min(yarr(this_order))) 
   endif else begin
      rough_disp   = (max(wave(this_order))-min(wave(this_order))) / $
                     (max(xarr(this_order))-min(xarr(this_order))) 
   endelse
   ; Use a wider break point spacing where possible works
   ; significantly better than using the min spacing everywhere with
   ; no adjustment, particularly for removing skylines and in regions
   ; of low object flux.  The minimum break point spacing should be
   ; set just small enough to avoid periodic residuals in the sky
   ; emission lines.
   bk_space     = 3.0*rough_disp
   min_bk_space = 0.5*rough_disp  ; 0.5 seems best for VIS & NIR
   print,'Setting break point spacing = '+strtrim(bk_space,2)
   print,'    min break point spacing = '+strtrim(min_bk_space,2)

   ; Do extraction in wavelength bins of constant velocity width
   log_lin = 1

   ; Perform secondary slit illumination function correction.  This
   ; helps with bright skylines, particularly in the NIR.
   fix_illum = 1

   ; Reduce order.
   longslit_reduce,obj,wave,temp_slit,temp_nosky,temp_noskyvar,$
      temp_wavefit,temp_skymodel,hdr=temp_hdr,/xshooter,/nobias,$
      flatfield=flat,flatvar=flat_var,skyref=nod_sky,gn=gn,rdnoise=rdn,$
      fwhm=fwhm,slitpos=slitpos,force=force,trace_obj=do_trace,$
      noobject=no_object,meanshift=mean_shift,outfile=temp_file,$
      objmask=objmask,subrange=[temp_slitmin,temp_slitmax],$
      waveref=wave,bkspace=bk_space,min_bkspace=min_bk_space,/set_bkpts,$
      loglin=log_lin,fixillum=fix_illum,altvar=alt_var,_extra=ex

   ; Check - Did longslit_reduce run properly?
   if (n_elements(temp_nosky) eq 1 or $
       n_elements(temp_noskyvar) eq 1 or $
       n_elements(temp_wavefit) eq 1 or $
       n_elements(temp_skymodel) eq 1) $
      then longslit_ok = 0 else longslit_ok = 1

   if longslit_ok then begin

      ; Store the section of the 2D reduced frames for this order.
      order_ls = where(slit ge temp_slitmin and slit le temp_slitmax)
      nosky(order_ls)    = temp_nosky(order_ls)
      noskyvar(order_ls) = temp_noskyvar(order_ls)
      wavefit(order_ls)  = temp_wavefit(order_ls)
      skymodel(order_ls) = temp_skymodel(order_ls)
      
      ; Read the new spectrum back in and store in the large array.
      if not(keyword_set(noobject)) then begin
         checkexists = findfile(temp_file+'_spec.fits',count=n_specfiles)
         checkexists = findfile(temp_file+'_err.fits',count=n_errfiles)
         if not(n_specfiles eq 0 or n_errfiles eq 0) then begin
            spec = rflam(temp_file+'_spec.fits',hspec,wspec)
            err  = rflam(temp_file+'_err.fits',herr,werr)
            n_pix = n_elements(spec)
            big_wave(0:n_pix-1,index) = wspec
            big_spec(0:n_pix-1,index) = spec
            big_err(0:n_pix-1,index)  = err
         endif
      endif

      ; Get heliocentric velocity.  This will be the same for
      ; all orders.
      temp_hdr    = headfits(temp_file+'_sub.fits')
      temp_vhelio = sxpar(temp_hdr,'HELIOVEL',count=n_vhelio)
      if (n_vhelio ne 0) then begin
         vhelio_applied = 1
         vhelio         = temp_vhelio
      endif
      
      ; Record the median wavelength, profile FWHM, slit position,
      ; offset from the standard star, and mean wavelength shift
      ; for this order.  Calculate offset from slit center.
      if not(keyword_set(noobject)) then begin
         medwave_list(index)   = median(wspec,/even)
         fwhm_list(index)      = fwhm
         position_list(index)  = slitpos
         offset_list(index)    = slitpos - ordercenter(index)
      endif

      ;; If a trace was made, read in and store trace information
      ;if not(keyword_set(noobject)) then begin
      ;   if do_trace then begin
      ;      trace_templ = {   $
      ;                version: 1.0, $    
      ;                datastart: 0L, $   
      ;                delimiter: ' ', $
      ;                missingvalue: -99.0, $
      ;                commentsymbol: '#', $
      ;                fieldcount: 1L, $
      ;                fieldtypes: 4L, $
      ;                fieldnames: 'value', $
      ;                fieldlocations: 0L, $
      ;                fieldgroups: [0L] $
      ;              }
      ;      trace_data = read_ascii(temp_file+'_trace.dat',$
      ;                              template=trace_templ)
      ;      ; The first numbers are the polynomial fit to the object
      ;      ; trace.  The last number is the object centroid, which
      ;      ; should be nearly identical to the zeroeth term in the
      ;      ; trace coefficients.
      ;      all_numbers  = trace_data.value
      ;      n_numbers    = n_elements(all_numbers)
      ;      temp_trace_coeffs = all_numbers(0:n_numbers-2)
      ;      n_coeffs     = n_elements(temp_trace_coeffs)
      ;      trace_coeffs(0:n_coeffs-1,index) = temp_trace_coeffs
      ;   endif else begin
      ;      ; Record fitted values
      ;      trace_coeffs(0:1,index) = [slitpos,slope]
      ;   endelse
      ;endif

   endif

endfor

; Save the 2D reduced arrays.
hout = hobj
sxaddpar,hout,'BZERO',0.0
get_date,date,/timetag
sxaddhist,'Reduced with XSHOOTER_REDUCE '+date+' (UTC)',hout
sxaddpar,hout,'ORIGEXP',filename(object),'Filename of original exposure.'
sxaddpar,hout,'WAVEGRID',filename(wavefile),'Wavelength-proxy grid'
sxaddpar,hout,'SLITGRID',filename(slitfile),'Slit-position-proxy grid'
if keyword_set(arc) then $
   sxaddpar,hout,'ARC',filename(arc),'Comparison lamp'
if keyword_set(flatfield) then $
   sxaddpar,hout,'FLATFLD',filename(flatfield),'Flat field'
if keyword_set(flatvar) then $
   sxaddpar,hout,'FLATVAR',filename(flatvar),'Flat field variance'
if vhelio_applied then begin
   sxaddpar,hout,'HELIOCNT','Heliocentric correction applied.'
   sxaddpar,hout,'HELIOVEL',vhelio,'lam_hcnt = lam_obs*(1+heliovel/c)'
endif
writefits,root+'_sub.fits',nosky,hout
writefits,root+'_var.fits',noskyvar,hout
writefits,root+'_sky.fits',skymodel,hout
writefits,root+'_wav.fits',wavefit,hout

; Record object information
if not(keyword_set(noobject)) then begin

   ; Remove extra zeros and save the big spec and error arrays.
   if (n_orders gt 1) then begin
      nonzero_ls  = where(total(big_wave,2) gt 0)
      max_nonzero = max(nonzero_ls)
      big_wave    = big_wave(0:max_nonzero-1,*)
      big_spec    = float(big_spec(0:max_nonzero-1,*))
      big_err     = float(big_err(0:max_nonzero-1,*))
   endif else begin
      big_spec    = float(big_spec)
      big_err     = float(big_err)
   endelse
   writefits,root+'_wspec.fits',big_wave,hout
   writefits,root+'_spec.fits',big_spec,hout
   writefits,root+'_err.fits',big_err,hout

   ; Output object profile and wavelength calibration information.
   openw,profile_lun,root+'_profile.dat',/get_lun
   for i=0,n_orders-1 do $
      printf,profile_lun,ordernum(i),medwave_list(i),fwhm_list(i),$
             position_list(i),offset_list(i),$
             format='(f8.2,f14.3,3(f12.4))'
   free_lun,profile_lun
   
   ;; Output trace information
   ;openw,trace_lun,root+'_trace.dat',/get_lun
   ;for i=0,n_orders-1 do printf,trace_lun,ordernum(i),trace_coeffs(*,i),$
   ;                             format='(6(f13.6))'
   ;free_lun,trace_lun

   ; Final fit to object profile and trace parameters (for evaluating
   ; the fits post-processing)
   indep_ls = where(independent_fit_made eq 1,n_indep,$
                    complement=fitted_ls,ncomp=n_fitted)
   if (n_indep ge 1) then begin
      ; Scale FWHM values
      fwhm_scale = mean(fwhm_list(indep_ls)/unscaled_fwhm_fit(indep_ls))
      fwhm_fit   = unscaled_fwhm_fit * fwhm_scale
      ; Fit offset and slope values
      if (n_indep ge n_for_linear) then begin
         if (arm eq 'VIS') then begin
            ; Reduce FWHM vs position to a linear fit by parametrizing
            ; quadratic coeff in terms of the linear coeff
            fwhm_hot  = (9.0e-10)*((ordercenter-1200)^2)
            net_fwhm  = fwhm_list - fwhm_hot
            b0        = 1.28e-6
            b1        = 3.29e-4
            tmp_x     = ordercenter - 1200
            tmp_y     = net_fwhm
            fit_x     = tmp_x + b1*(tmp_x^2)
            fit_y     = tmp_y - b0*(tmp_x^2)
            pfit_fwhm = poly_fit(fit_x(indep_ls),fit_y(indep_ls),1)
            a0        = pfit_fwhm(0)
            a1        = pfit_fwhm(1)
            a2        = b0 + b1*a1
            net_fwhm_fit  = a0 + a1*tmp_x + a2*(tmp_x^2)
            fwhm_fit      = net_fwhm_fit + fwhm_hot
         endif
         pfit_offset = robust_poly_fit(ordercenter(indep_ls),$
                                       offset_list(indep_ls),$
                                       offset_order)
         ;pfit_slope  = robust_poly_fit(ordercenter(indep_ls),$
         ;                              trace_coeffs(1,indep_ls),$
         ;                              slope_order)
      endif else begin
         pfit_offset = poly_fit(ordercenter(indep_ls),offset_list(indep_ls),$
                                  offset_order)
         ;pfit_slope = poly_fit(ordercenter(indep_ls),trace_coeffs(1,indep_ls),$
         ;                      slope_order)  
      endelse
      ; Reconstruct total fitted values
      offset_fit = fltarr(n_orders)
      for k=0,offset_order do $
         offset_fit = offset_fit + pfit_offset(k)*(ordercenter^k)
      ;slope_fit = fltarr(n_orders)
      ;for k=0,slope_order do $
      ;   slope_fit = slope_fit + pfit_slope(k)*(ordercenter^k)
   endif else begin
      fwhm_fit   = fltarr(n_orders)
      offset_fit = fltarr(n_orders)
      ;slope_fit  = fltarr(n_orders)
   endelse
   ; Reconstruct fitted position
   position_fit = ordercenter + offset_fit
   
   ; Plot profile and trace information
   fwhm_plotmin   = min(fwhm_list) < min(fwhm_fit)
   fwhm_plotmax   = max(fwhm_list) > max(fwhm_fit)
   offset_plotmin = min(offset_list) < min(offset_fit)
   offset_plotmax = max(offset_list) > max(offset_fit)
   ;slope_plotmin  = min(trace_coeffs(1,*)) < min(slope_fit)
   ;slope_plotmax  = max(trace_coeffs(1,*)) > max(slope_fit)
   ps_openfile,root+'_trace',/portrait
   !p.multi    = [0,2,3]
   !p.charsize = 1.2
   !x.charsize = 1.2
   !y.charsize = 1.2
   dx          = max(approx_medwave) - min(approx_medwave)
   !x.range    = [min(approx_medwave)-0.05*dx,max(approx_medwave)+0.05*dx]
   !x.style    = 1
   ; FWHM
   plot,approx_medwave,fwhm_list,psym=1,title='FWHM',$
        yr=[fwhm_plotmin,fwhm_plotmax]
   if (n_fitted ne 0) then $
      oplot,approx_medwave(fitted_ls),fwhm_list(fitted_ls),psym=6
   oplot,approx_medwave,fwhm_fit
   plot,approx_medwave,fwhm_list-fwhm_fit,psym=1,title='FWHM Residuals'
   if (n_fitted ne 0) then $
      oplot,approx_medwave(fitted_ls),(fwhm_list-fwhm_fit)(fitted_ls),psym=6
   ; Offset
   plot,approx_medwave,offset_list,psym=1,title='Offset',$
        yr=[offset_plotmin,offset_plotmax]
   if (n_fitted ne 0) then $
      oplot,approx_medwave(fitted_ls),offset_list(fitted_ls),psym=6
   oplot,approx_medwave,offset_fit
   plot,approx_medwave,offset_list-offset_fit,psym=1,title='Offset Residuals'
   if (n_fitted ne 0) then $
      oplot,approx_medwave(fitted_ls),(offset_list-offset_fit)(fitted_ls),psym=6
   ;; Slope
   plot,[0],[0]
   plot,[0],[0]
   ;plot,approx_medwave,trace_coeffs(1,*),psym=1,title='Slope',$
   ;     yr=[slope_plotmin,slope_plotmax]
   ;if (n_fitted ne 0) then $
   ;   oplot,approx_medwave(fitted_ls),trace_coeffs(1,fitted_ls),psym=6
   ;oplot,approx_medwave,slope_fit
   ;plot,approx_medwave,trace_coeffs(1,*)-slope_fit,psym=1,title='Slope Residuals'
   ;if (n_fitted ne 0) then $
   ;   oplot,approx_medwave(fitted_ls),(trace_coeffs(1,*)-slope_fit)(fitted_ls),$
   ;         psym=6
   ps_closefile

endif

;; Remove temporary files.
print,'Removing temporary files...'
spawn,'"rm" '+temp_file+'_err.fits'
spawn,'"rm" '+temp_file+'_profile.ps'
spawn,'"rm" '+temp_file+'_sky.fits'
spawn,'"rm" '+temp_file+'_spec.fits'
spawn,'"rm" '+temp_file+'_spec.ps'
spawn,'"rm" '+temp_file+'_sub.fits'
if not(keyword_set(noobject)) then begin
   ;spawn,'"rm" '+temp_file+'_trace.ps'
   spawn,'"rm" '+temp_file+'_trace.dat'
   spawn,'"rm" '+temp_file+'_var.fits'
   spawn,'"rm" '+temp_file+'_wav.fits'
endif

print,'Done.'

; Close log
journal

; Restore plot settings
cleanplot,/silent

return
end
