Pro xshooter_response,exposure,air=air,check=check
  
;+--------------------------------------------------------------------------
;
; XSHOOTER_RESPONSE
;
; Generate an X-Shooter response function from a standard star
; spectrum.  
;
; The new response function is generated by multiplying a nominal
; repsonse function by the ratio of fluxes in the input raw spectrum
; and the spectrum used to generate the nominal response function.
; Corrections are made for atmospheric extinction and exposure time,
; as well as for the intrinsic spectra of the standard stars, which
; may be different.
;
; The input spectrum must be one of the seven standard stars used for
; X-Shooter flux calibration: BD+17, EG274, Feige110, GD153, GD71,
; LTT3218, and LTT7987.
;
; Currently only recommended for UVB and VIS arms. 
;
; The directories containing the calibrated standard star spectra
; (cal_dir) and nominal response functions (nom_resp_dir) must be
; edited in the program body below, along with the filenames for the
; atmospheric extinction model (ext_file) and telluric absorption
; model (tell_file).  The calibrated standard star spectra and
; atmospheric extinction model are taken from the ESO X-Shooter
; pipeline.  The telluric absorption model is taken from the NSO.
;
; Wavelengths for the input spectrum are assumed to be in vacuum,
; unless the /air keyword is set.
;
; INPUTS
;        exposure   - String, root file name of standard star
;                     exposure, with no ".fits" or other extension
;                     (e.g., "XSHOO.2013-05-06T09:26:33.931").  The
;                     program will look for the files
;                     <exposure>_spec.fits and <epxosure>_wspec.fits,
;                     output by xshooter_reduce.pro.  May include
;                     directory name.  Wavelengths must be in vacuum,
;                     unless the /air keyword is set.
; 
; KEYWORDS
;       air         - If set, assumes input wavelengths
;                     (<exposure>_wspec.fits) are in air.
;       check       - If set, displays fits to the flux ratios
;                     in each order.
;
; OUTPUTS
;        Writes response function to Resp-<exposure>.fits and
;        corresponding wavlengths to Resp-<esposure)>_wave.fits (files
;        will be written to the directory containing the input
;        spectrum).
;
; HISTORY
;        Written 10/16/2013 GDB
;        Updated nominal response functions 2/27/2014 GDB
;-------------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:  xshooter_response,exposure,/air,/check'
   return
endif

; In the XSHOOTER_REDUCE IDL pipeline, the following four directories
; may be set to the calib/ directory included in the distribution.

; Directory in which calibrated standard spectra are kept
cal_dir      = '~/stds/xshooter/'
; Directory in which nominal response functions are kept
nom_resp_dir = '~/eso/xshooter/flux_cal/calibrations/'
; File name of atmospheric extinction model
ext_file     = '~/eso/xshooter/flux_cal/calibrations/'+$
               'xsh_paranal_extinct_model_uvb.fits'
; File name of telluric absorption model
tell_file    = '~/calib/telluric/nso/transdata_0.5_2.5.fits'

; X-Shooter arm
exposure_root = filename(exposure,dirname=exposure_dir)
hdr           = headfits(exposure_dir+exposure_root+'_spec.fits')
arm           = strsetlen(gxpar(hdr,'HIERARCH ESO SEQ ARM'),3)
if not(arm eq 'UVB' or arm eq 'VIS' or arm eq 'NIR') then begin
   print,'X-Shooter arm not recognized'
   return
endif

; Exit if NIR arm
if (arm eq 'NIR') then begin
   print,'Not currently recommended for NIR arm'
   return
endif

; Atmopsheric extinction model.  Assume a constant value beyond the
; highest modelled wavelength.
ext_model = mrdfits(ext_file,1,ext_hdr)
ext_w     = 10.*ext_model.lambda
ext_k     = ext_model.extinction   ; mag/airmass
nbin_ext  = n_elements(ext_w)
ext_dw    = ext_w(nbin_ext-1) - ext_w(nbin_ext-2)
ext_w     = [ext_w,ext_w(nbin_ext-1)+ext_dw]
ext_k     = [ext_k,ext_k(nbin_ext-1)]

; Nominal response function, along with the raw extracted
; spectrum from which it was generated
nom_resp     = readfits(nom_resp_dir+'Resp-'+arm+'.fits')
nom_respwave = readfits(nom_resp_dir+'Resp-'+arm+'_wave.fits')
case arm of
   'UVB': nom_exposure = 'XSHOO.2013-05-07T06:59:24.937'
   'VIS': nom_exposure = 'XSHOO.2013-05-07T06:59:30.107'
   'NIR': nom_exposure = 'XSHOO.2013-05-07T06:59:33.209'
endcase
nom_root    = nom_resp_dir+nom_exposure
nom_fraw    = readfits(nom_root+'_spec.fits',nom_hdr)
nom_eraw    = readfits(nom_root+'_err.fits')
nom_w       = readfits(nom_root+'_wspec.fits')
nom_airmass = gxpar(nom_hdr,'HIERARCH ESO TEL AIRM END')
nom_texp    = gxpar(nom_hdr,'EXPTIME')
nom_vhelio  = gxpar(nom_hdr,'HELIOVEL')
nom_sz      = size(nom_fraw)
nom_n_pix   = nom_sz(1)
nom_n_order = nom_sz(2)
nom_ls      = where(nom_w gt 0)
; Correct for extinction
nom_k                  = 0.*nom_fraw
nom_k(nom_ls)          = interpol(ext_k,ext_w,nom_w(nom_ls))
nom_frac_trans         = 0.*nom_fraw
nom_frac_trans(nom_ls) = 10.^(-0.4*nom_airmass*nom_k(nom_ls))
nom_fcorr              = 0.*nom_fraw
nom_ecorr              = 0.*nom_eraw
nom_fcorr(nom_ls)      = nom_fraw(nom_ls) / nom_frac_trans(nom_ls)
nom_ecorr(nom_ls)      = nom_eraw(nom_ls) / nom_frac_trans(nom_ls)
; Divide out calibrated spectrum and exposure time.  These standard
; star spectra have been created from calibraiton spectra included in
; the X-Shooter pipeline.  Units are erg/s/cm2/Ang.  Wavelengths are
; in AIR.
nom_star       = 'EG274'
nom_fcal       = readfits(cal_dir+nom_star+'_flux.fits')
nom_wcal       = readfits(cal_dir+nom_star+'_wave.fits')
nom_fcal_match = 0.*nom_fraw
nom_fcal_match(nom_ls) = interpol(nom_fcal,nom_wcal,nom_w(nom_ls))
nom_fnorm      = 0.*nom_fraw
nom_enorm      = 0.*nom_eraw
nom_fnorm(nom_ls) = nom_fcorr(nom_ls) / nom_fcal_match(nom_ls) / nom_texp
nom_enorm(nom_ls) = nom_ecorr(nom_ls) / nom_fcal_match(nom_ls) / nom_texp

; Read in raw extracted spectrum
fraw    = readfits(exposure_dir+exposure_root+'_spec.fits')
eraw    = readfits(exposure_dir+exposure_root+'_err.fits')
w       = readfits(exposure_dir+exposure_root+'_wspec.fits')
; Convert to air wavelengths
if not(keyword_set(air)) then vactoair,w
airmass = gxpar(hdr,'HIERARCH ESO TEL AIRM END')
texp    = gxpar(hdr,'EXPTIME')
vhelio  = gxpar(hdr,'HELIOVEL')
sz      = size(fraw)
n_pix   = sz(1)
n_order = sz(2)
ls      = where(w gt 0)
; Correct for extinction
k              = 0.*fraw
k(ls)          = interpol(ext_k,ext_w,w(ls))
frac_trans     = 0.*fraw
frac_trans(ls) = 10.^(-0.4*airmass*k(ls))
fcorr          = 0.*fraw
ecorr          = 0.*eraw
fcorr(ls)      = fraw(ls) / frac_trans(ls)
ecorr(ls)      = eraw(ls) / frac_trans(ls)
; Determine which star
targname   = strupcase(gxpar(hdr,'HIERARCH ESO OBS TARG NAME'))
star       = 'unknown'
if (strpos(targname,'BD') ne -1 and strpos(targname,'17') ne -1) $
   then star = 'BD+17'
if (strpos(targname,'EG') ne -1 and strpos(targname,'274') ne -1) $
   then star = 'EG274'
if (strpos(targname,'FEIGE') ne -1 and strpos(targname,'110') ne -1) $
   then star = 'Feige110'
if (strpos(targname,'GD') ne -1 and strpos(targname,'153') ne -1) $
   then star = 'GD153'
if (strpos(targname,'GD') ne -1 and strpos(targname,'71') ne -1) $
   then star = 'GD71'
if (strpos(targname,'LTT') ne -1 and strpos(targname,'3218') ne -1) $
   then star = 'LTT3218'
if (strpos(targname,'LTT') ne -1 and strpos(targname,'7987') ne -1) $
   then star = 'LTT7987'
if (star eq 'unknown') then begin
   print,'Target name not recognized'
   return
endif else print,'Star = '+star
; Divide out calibrated spectrum and exposure time
fcal       = readfits(cal_dir+star+'_flux.fits')
wcal       = readfits(cal_dir+star+'_wave.fits')
fcal_match = 0.*fraw
fcal_match(ls) = interpol(fcal,wcal,w(ls))
fnorm      = 0.*fraw
enorm      = 0.*eraw
fnorm(ls)  = fcorr(ls) / fcal_match(ls) / texp
enorm(ls)  = ecorr(ls) / fcal_match(ls) / texp

; Check that they have the same number of orders
if (n_order ne nom_n_order) then begin
   print,'Spectra have different numbers of orders'
   return
endif

; Translate wavelengths to observed (geocentric).
c        = 2.99792458d5
nom_wobs = nom_w / (1+nom_vhelio/c)
wobs     = w / (1+vhelio/c)

; Telluric absorption model
ftell_hires = rflam(tell_file,htell,wtell)
fwhm_kms    = 50.
ftell_lores = convol(ftell_hires,gaussker(fwhm=fwhm_kms),/edge_trunc)
ftell       = 0.*fraw
ftell(ls)   = interpol(ftell_lores,wtell,wobs(ls))

; Record scaling values (if any)
wmid_list  = dblarr(n_order)
scale_list = fltarr(n_order)

; Adjusted response function
resp     = 0.*fraw
respwave = 0.*wobs

for i=0,n_order-1 do begin

   tmp_nom_f    = nom_fnorm(*,i)
   tmp_nom_w    = nom_wobs(*,i)
   tmp_nom_resp = nom_resp(*,i)
   tmp_nom_ls   = where(tmp_nom_w gt 0)
   tmp_f        = fnorm(*,i)
   tmp_w        = wobs(*,i)
   tmp_ls       = where(tmp_w gt 0)   
   tmp_ftell    = ftell(*,i)   ; Avoid regions of strong tell abs

   ; Arm-specific masking
   case arm of
      'UVB': begin
                wmin = 3050
                wmax = 1e6
             end
      'VIS': begin
                wmin = 5500
                wmax = 10200
             end
      'NIR': begin
                wmin = 1
                wmax = 1e6
             end
   endcase

   ; Interpolate nominal spectrum and response function onto new
   ; wavelengths.  Avoid regions of strong telluric absorption
   ; when fitting.
   covered_ls = where(tmp_w gt 0 and $
                      tmp_w ge min(tmp_nom_w) and $
                      tmp_w le max(tmp_nom_w) and $
                      tmp_w ge wmin           and $
                      tmp_w le wmax,n_covered)
   tmp_nom_fmatch    = 0.*tmp_f
   tmp_nom_respmatch = 0.*tmp_f
   tmp_nom_fmatch(covered_ls)    = interpol(tmp_nom_f(tmp_nom_ls),$
                                   tmp_nom_w(tmp_nom_ls),tmp_w(covered_ls))
   tmp_nom_respmatch(covered_ls) = interpol(tmp_nom_resp(tmp_nom_ls),$
                                   tmp_nom_w(tmp_nom_ls),tmp_w(covered_ls))
   x        = tmp_w(covered_ls)
   ratio    = tmp_f(covered_ls) / tmp_nom_fmatch(covered_ls)
   tell     = tmp_ftell(covered_ls)

   ; Correct for exposure time!

   ; Initial masking
   mask  = lonarr(n_covered)
   nbin  = 20
   xmed  = fltarr(nbin)
   rmed  = fltarr(nbin)
   rsdev = fltarr(nbin)
   nsig  = 2
   pixperbin = histogram(x,nbin=nbin,loc=x_lo,rev=ri)
   for t=0,nbin-1 do begin
      if (pixperbin(t) eq 0) then continue           
      ind_ls   = ri(ri(t):ri(t+1)-1)
      xbin     = x(ind_ls)
      rbin     = ratio(ind_ls)
      xmed(t)  = median(xbin)
      rmed(t)  = median(rbin)
      rsdev(t) = robust_sigma(rbin)
      mask_ls  = where(abs(rbin-rmed(t)) ge nsig*rsdev(t),n_mask)
      if (n_mask ne 0) then mask(ind_ls(mask_ls)) = 1
   endfor
   keep_ls   = where(pixperbin gt 1)
   xmed      = xmed(keep_ls)
   rmed      = rmed(keep_ls)
   rsdev     = rsdev(keep_ls)
   ratio_err = interpol(rsdev,xmed,x)

   ; Mask regions of strong telluric absorption
   tell_mask_ls = where(tell lt 0.9,n_tell_mask)
   if (n_tell_mask ne 0) then mask(tell_mask_ls) = 1

   ; Explicitly mask strong O2 bands
   o2_ls = where((x ge 6866 and x le 6960) or $
                 (x ge 7590 and x le 7707),n_o2)
   if (n_o2 ne 0) then mask(o2_ls) = 1

   ; Mask H-alpha
   halpha_ls = where(x ge 6536 and x le 6586,n_halpha)
   if (n_halpha ne 0) then mask(halpha_ls) = 1

   ; How to fit flux ratio?  Default is b-spline fit
   do_bspline = 1
   do_linear  = 0
   do_scalar  = 0
   do_median  = 0

   ; Order-specific fit method and masking
   wmid         = median(tmp_w(where(tmp_w gt 0)))
   wmid_list(i) = wmid
   if (abs(wmid-9500) le 200) then do_linear = 1  ; High tell abs
   if (arm eq 'UVB' and abs(wmid-5660) le 200) then do_median = 1 ; Dichroic
   if (abs(wmid-4330) le 200) then begin
      do_linear       = 1  ; Bad columns
      badcol_ls       = where(x le 4240)
      mask(badcol_ls) = 1
   endif
   if (arm eq 'UVB' and wmid ge 4400) then do_median = 1
   if (arm eq 'VIS' and abs(wmid-8975) le 200) then do_linear = 1
   if (arm eq 'VIS' and abs(wmid-10100) le 200) then do_linear = 1
   if (arm eq 'NIR') then do_scalar = 1
   if (arm eq 'NIR' and abs(wmid-10070) le 200) then begin
      bluemask_ls = where(x le 10090,n_bluemask)
      if (n_bluemask ne 0) then mask(bluemask_ls) = 1
   endif

   if (do_linear or do_scalar or do_median) then do_bspline = 0

   if do_scalar then begin
      fit_ls        = where(mask eq 0)
      med_ratio     = median(ratio(fit_ls))
      rfit          = med_ratio + 0.*ratio
      scale_list(i) = med_ratio
      new_mask      = mask
   endif

   if do_linear then begin
      fit_ls   = where(mask eq 0)
      pfit     = robust_poly_fit(x(fit_ls),ratio(fit_ls),1)
      rfit     = pfit(0) + x*pfit(1)
      new_mask = mask
   endif

   if do_median then begin
      ; Not masking pixels.  Intended to capture sharp
      ; discontinuities.
      nmed_pix = 31
      rfit     = median(ratio,nmed_pix)
      rfit(0:nmed_pix-1) = median(ratio(0:nmed_pix-1))
      rfit(n_covered-nmed_pix-1:n_covered-1) = $
                           median(ratio(n_covered-nmed_pix-1:n_covered-1))
      ; Interpolate over abs line near 4850 (kludge for Feige110,
      ; where the calibrated spectrum may be somewhat off)
      if (arm eq 'UVB' and abs(wmid-4905) lt 100) then begin
         good_ls = where(x le 4847 or x ge 4877)
         rfit    = interpol(rfit(good_ls),x(good_ls),x)
      endif
      if (arm eq 'UVB' and abs(wmid-4600) lt 100) then begin
         good_ls = where(x le 4675 or x ge 4700)
         rfit    = interpol(rfit(good_ls),x(good_ls),x)
      endif
      ; Smooth
      if (arm eq 'UVB' and abs(wmid-5660) lt 100) then smooth_fwhm = 10 $
         else smooth_fwhm = 50
      rfit = convol(rfit,gaussker(fwhm=smooth_fwhm),/edge_truncate)
      new_mask = replicate(0.,n_covered)
   endif

   if do_bspline then begin
      ; B-spline fit
      bfit_iter = 0
      max_iter  = 10
      bfit_done = 0
      new_mask  = mask
      if (arm eq 'UVB') then n_bkpts = 6
      if (arm eq 'VIS') then n_bkpts = 6
      if (arm eq 'VIS' and abs(wmid-5575) lt 100) then n_bkpts = 10
      if (arm eq 'VIS' and abs(wmid-5775) lt 100) then n_bkpts = 10
      if (arm eq 'VIS' and wmid ge 6000) then n_bkpts = 4
      if (arm eq 'VIS' and abs(wmid-7350) lt 200) then n_bkpts = 3
      if (arm eq 'VIS' and abs(wmid-8975) lt 200) then n_bkpts = 3
      if (arm eq 'NIR') then n_bkpts = 4
      while not(bfit_done) do begin
         bfit_iter++
         fit_ls       = where(new_mask eq 0)
         sset         = bspline_iterfit(x(fit_ls),ratio(fit_ls),nbkpts=n_bkpts)
         rfit         = bspline_valu(x,sset)
         temp_mask    = mask
         temp_mask_ls = where(mask eq 1 or $
                              abs(ratio-rfit) ge nsig*ratio_err,n_temp_mask)
         if (n_temp_mask ne 0) then temp_mask(temp_mask_ls) = 1
         if array_equal(temp_mask,new_mask) then bfit_done = 1
         if (bfit_iter ge max_iter) then bfit_done = 1
         new_mask = temp_mask
      endwhile
   endif

   ; Adjusted response function
   resp(covered_ls,i)     = tmp_nom_respmatch(covered_ls) / rfit
   respwave(covered_ls,i) = tmp_w(covered_ls)

   ; Check results (optional)
   if keyword_set(check) then begin
      ymin = 0.5 < (min(rfit)-0.05)
      ymax = 1.5 > (max(rfit)+0.05)
      plot,x,ratio,psym=6,yr=[ymin,ymax],/ys,xs=3,title=strtrim(wmid,2),$
           charsize=2
      tmp_ls = where(new_mask eq 1,n_tmp)
      if (n_tmp ne 0) then oplot,x(tmp_ls),ratio(tmp_ls),psym=6,color=220
      oplot,x,rfit,color=250                   
      oplot,xmed,rmed,psym=6,color=150,symsize=2
      print,'Hit any key for next order'
      key = get_kbrd()
   endif

endfor

if arm eq 'NIR' then stop

; Write response function
writefits,exposure_dir+'Resp-'+exposure_root+'.fits',resp,hdr
writefits,exposure_dir+'Resp-'+exposure_root+'_wave.fits',respwave,hdr

return
end
