Pro xshooter_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile,arm=arm

;+--------------------------------------------------------------------------
;
; XSHOOTER_SLITGRID      12/2009
;
; Trace orders in a bright standard star exposure to create a map of slit
; positions.  The slit position at a pixel is the x-intercept of the
; curve of constant position along the slit that would pass through
; the center of that pixel.
;
; Standard star should be at low airmass (less than 1.1)
;
; INPUTS
;         stdfile   - String, FITS filename of standard star exposure
;
; KEYWORDS
;         flatfile  - String, FITS filename of flatfield exposure
;                     (non-diffused).  Used to identify the edges
;                     of each order.
;         arm       - String specifying which X-Shooter arm: 'UVB',
;                     'VIS', or 'NIR'.  Must be set.
;
; OUTPUTS
;         slitgrid  - Name of FITS file to which to write slit grid
;         orderlist - Name of ASCII file to which to write list of
;                     order centers and edges.  Output only if the
;                     flatfile keyword is set.
;
; HISTORY
;         Adapted from MAGE_SLITGRID 12/6/2009
;         Modified way in which order centers are calculated 9/16/2011
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
 print,'CALLING SEQUENCE: '
 print,'   xshooter_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile,'
 print,'                     arm=arm'
 return
endif

device,decomp=0
loadct,39

;;; Which arm?
if not(keyword_set(arm)) then begin
   print,'Must specify which X-Shooter arm'
   return
endif
; Allow upper or lower case
case arm of
   'uvb': arm = 'UVB'
   'UVB': arm = 'UVB'
   'vis': arm = 'VIS'
   'VIS': arm = 'VIS'
   'nir': arm = 'NIR'
   'NIR': arm = 'NIR'
   else : begin
             print,'X-Shooter arm not recognized'
             return
          end
endcase

;;; Rough size of slit in pixels + a wide buffer
case arm of
   'UVB': slitsz = 80
   'VIS': slitsz = 80
   'NIR': slitsz = 60
endcase

;;; Read in standard star exposure
std = readfits(stdfile,hstd)
std = xshooter_subbias(std,arm=arm)  ; No effect for NIR

;;; For UVB and VUS, swap x and y such that orders run in the
;;; x-direction.
if (arm eq 'UVB' or arm eq 'VIS') then std = rotate(std,4)
sz  = size(std)
xsz = sz(1)
ysz = sz(2)

;;; Arrays of x and y values
xarr = rebin(findgen(xsz),xsz,ysz)  ; These should be floating point
yarr = rebin(rotate(findgen(ysz),1),xsz,ysz)

;;; Illuminated region
mask = lonarr(xsz,ysz)
; No masking for UVB and VIS arms.  Orders are well away from
; pre/overscan regions
if (arm eq 'NIR') then begin
   ; Red edge x(y) polynomial
   r_terms = [1459.7389, 0.65775340, -0.00090907266, 3.2113316e-06]
   r_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(r_terms)-1 do r_yfit = r_yfit + r_terms(i)*(yarr^i)
   r_mask_ls = where(xarr gt r_yfit)
   ; Blue edge 
   b_terms = [433.63313, -0.30479697, 0.00011715153, -1.1558709e-06]
   b_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(b_terms)-1 do b_yfit = b_yfit + b_terms(i)*(yarr^i)
   b_mask_ls = where(xarr lt b_yfit)
   mask(r_mask_ls)     = 1
   mask(b_mask_ls)     = 1
   mask(0:3,*)         = 1
   mask(xsz-4:xsz-1,*) = 1
   mask(*,0:3)         = 1
endif

;;; Median cut through the array
if (arm eq 'UVB') then begin
   cut_xmin = 450   ; 2x binning in spectral dir
   cut_xmax = 500
   cut_ymin = 330
   cut_ymax = 1610
endif else if (arm eq 'VIS') then begin
   cut_xmin = 1085  ; 2x binning in spectral dir
   cut_xmax = 1135
   cut_ymin = 240
   cut_ymax = 1920
endif else begin
   cut_xmin = 890
   cut_xmax = 960
   cut_ymin = 50
   cut_ymax = 1060
endelse
cut_xmid = 0.5*(cut_xmin+cut_xmax)
cut      = median(std(cut_xmin:cut_xmax,*),dim=1)

;;; Locate peaks.  Avoid edges of illuminated region.
first_peak = 1
for j=cut_ymin,cut_ymax do begin
   case arm of
      'UVB': nside = 60
      'VIS': nside = 50
      'NIR': nside = 40
   endcase
   med_ymin = (0 > (j-nside)) < (ysz-1-2*nside)
   med_ymax = med_ymin + 2*nside
   if (cut(j) eq max(cut(j-10:j+10)) and $
       cut(j) gt median(cut(med_ymin:med_ymax)) + $
                    5*robust_sigma(cut(med_ymin:med_ymax)) and $
       cut(j) lt 60000. and $          ; Avoid saturated lines
       cut(j) ne cut(j-1)) then begin  ; Avoid double-counting
      if (first_peak) then begin
         peak_list  = j
         first_peak = 0
      endif else begin
         peak_list  = [peak_list,j]
      endelse
   endif
endfor
n_peak = n_elements(peak_list)

;;; Plot peak locations
window,9,xs=1000,ys=500,title='Peak Locations'
!p.multi = [0,1,1]
plot,cut,psym=10
for i=0,n_peak-1 do oplot,replicate(peak_list(i),2),[-1e6,1e6],color=250

;;; Illuminated regions
xrange = [0,xsz-1]
yrange = [0,ysz-1]

;;; Master slit grid.  Flag non-illuminated regions.
slit = fltarr(xsz,ysz) - 1000

;;; Y-intercept of each trace
y_intercept = fltarr(n_peak)

;;; Trace, fit, and compute slit grid in the vicinity of each peak.
print,'Tracing...'
window,10,xs=1000,ys=800,title='Traces'
!p.multi = [0,ceil(sqrt(n_peak)),ceil(sqrt(n_peak))]
for j=0,n_peak-1 do begin
   ; Avoid non-illuminated regions
   xstart = cut_xmid
   ystart = peak_list(j)
   medlen = 3.
   traceline,std,xstart,ystart,xtr,ytr,/xdir,merit=merit,$
             median=medlen,range=xrange,bound=yrange,nside=10,/noplot,/bygauss,$
             coord=mask,cbound=[0,0]
   ; Reject bad points
   keep_ls = where(merit gt 0,n_keep)
   xtr   = xtr(keep_ls)
   ytr   = ytr(keep_ls)
   merit = merit(keep_ls)
   ; Use a lower-order polynomial for the preliminary fit.
   npoly_prelim  = 3
   temp_trcoeffs = robust_poly_fit(xtr,ytr,npoly_prelim,yfit,yerr)
   ; Reject outliers and refit to get uncertainties in
   ; coefficients.  User a higher-order polynomial for longer orders.
   ls = where(abs(ytr-yfit) le ((3*yerr) > 1.0))
   xlength = max(xtr(ls))-min(xtr(ls))
   npoly_trace = 3
   if (xlength gt  800) then npoly_trace = 4
   if (xlength gt 1700) then npoly_trace = 5
   trcoeffs = poly_fit(xtr(ls),ytr(ls),npoly_trace,$
                       sigma=sig,yerror=yerr,yfit=yfit)
   y_intercept(j) = trcoeffs(0)

   ; Plot residuals
   plot,xtr(ls),ytr(ls)-yfit,psym=3,xs=3,/ystyle,yr=[-1,1],$
        title=strtrim(j,2)+' ('+strtrim(xlength,2)+')',charsize=2
   oplot,[-1e6,1e6],[0,0],linestyle=2
   ; Create a temporary 2D slit array assuming all loci of constant
   ; slit position run parallel to this trace
   temp_slit = xarr
   for k=1,npoly_trace do temp_slit = temp_slit - trcoeffs(k)*(xarr^k)

   ; Extract the region around this trace and save in the master
   ; slitgrid.  Assume the std is at the center of the slit, and
   ; include a margin at the edge of the slit.  Use wider margins for
   ; bluer orders.  Only save the blue end of the reddest order.
   if (j eq n_peak-1) then separation = (peak_list(j)-peak_list(j-1)) $
      else separation = (peak_list(j+1)-peak_list(j))
   order_ls = where(temp_slit ge trcoeffs(0) - separation/2. and $
                    temp_slit le trcoeffs(0) + separation/2.)
   slit(order_ls) = temp_slit(order_ls)
endfor

;;; Write slitgrid to file.  Un-swap x and y first for UVB and VIS.
sxaddpar,hstd,'BZERO',0.0
if (arm eq 'UVB' or arm eq 'VIS') then begin
   writefits,slitgrid,rotate(slit,4),hstd
endif else begin
   writefits,slitgrid,slit,hstd
endelse

;;; Identify order edges (optional)
if keyword_set(flatfile) then begin
   ;flat = xshooter_subbias(readfits(flatfile,hflat))
   flat = readfits(flatfile,hflat)
   if (arm eq 'UVB' or arm eq 'VIS') then flat = rotate(flat,4)
   window,11,xs=1000,ys=500
   !p.multi   = [0,1,1]
   print,'Left-click once on each edge of each order, left then right.'
   print,'Right-click when done with each order.'
   slit_slice = slit(xsz/2.-200:xsz/2.+200,0:ysz-1)
   flat_slice = flat(xsz/2.-200:xsz/2.+200,0:ysz-1)
   orderedge_list = fltarr(2*n_peak)
   for i=0,n_peak-1 do begin
      x_min = y_intercept(i) - (2./3)*slitsz
      x_max = y_intercept(i) + (2./3)*slitsz
      y_min = 0
      plot_ls = where(slit_slice ge x_min and slit_slice le x_max)
      y_max = cutvalue(flat_slice(plot_ls),99.9)
      plot,slit_slice(plot_ls),flat_slice(plot_ls),psym=3,xr=[x_min,x_max],/xs,$
           yr=[y_min,y_max]
      getpoints,temp_orderedge,/plotx,color=150
      orderedge_list(2*i:2*i+1) = temp_orderedge
   endfor
   
   ;;; Compute center of each order
   n_orderedge = n_elements(orderedge_list)
   if (n_orderedge mod 2 ne 0) then begin
      print,'XSHOOTER_SLITGRID: Error - There should be an even number of order' 
      print,'               edges marked (two for each order).  Exiting without'
      print,'               writing order list.'
      return
   endif
   n_order        = n_orderedge / 2
   leftedge_list  = orderedge_list(2*lindgen(n_order))
   rightedge_list = orderedge_list(1+2*lindgen(n_order))
   ;center_list    = (leftedge_list + rightedge_list) / 2.
   ; Fit a smooth function to the order midpoints
   midpoint_list  = (leftedge_list + rightedge_list) / 2.
   center_polyorder = floor(n_order/2.) < 5
   ordernum         = findgen(n_order)
   center_pfit      = robust_poly_fit(ordernum,midpoint_list,center_polyorder)
   center_list      = 0.*midpoint_list
   for i=0,center_polyorder do center_list = center_list + $
                                             center_pfit(i)*(ordernum^i)
   
   ;;; Write order info to file
   print,'Writing order info to file...'
   openw,unit,orderlist,/get_lun
   printf,unit,'#  CCD   Order    Center   Left edge  Right edge'
   for i=0,n_order-1 do $
      printf,unit,'XSHOOTER',i+1,center_list(i),leftedge_list(i),$
                  rightedge_list(i),format='(a6,i6,3(f12.3))'
   free_lun,unit
endif

; Restore plot settings
cleanplot,/silent

return
end
