Pro xshooter_wavecalib,arcfile,orderlist,slitgrid,wave_uncalib,wave_calib,$
                 linelist=linelist,dispersion=dispersion,refwave=refwave,$
                 refarc=refarc,nstep=nstep,slitrange=slitrange,$
                 tweak=tweak,auto=auto,order=order,dark=dark,arm=arm,$
                 ybin=ybin

;+----------------------------------------------------------------------------
;
; XSHOOTER_WAVECALIB
;
; Wavelength calibrate X-Shooter orders by hand.
;
; The default is to use a third order polynomial for the wavelength
; solution.  The user can change this interactively.
;
; INPUTS
;         arcfile      - FITS filename of arc lamp exposure.  Program
;                        will check the header to determine whether this
;                        is the blue or the red side.
;         orderlist    - Name of ASCII file containing the order
;                        information [CCD, order #, center, left edge,
;                        right edge], as output by xshooter_slitgrid.pro.
;         slitgrid     - FITS filename of slit position map (see
;                        xshooter_slitgrid.pro)
;         wave_uncalib - FITS filename of uncalibrated wavelength map,
;                        as produced by xshooter_wavecalib.  Also used
;                        to name output files containing wavelength fit
;                        coefficients and identified skylines for each
;                        order.
;
; KEYWORDS
;         linelist     - String, name of ASCII file containing line
;                        wavelengths.  Default is OH line
;                        list suitable for moderate resolution.  Wavelengths
;                        should be listed in the first column and have
;                        units of angstroms.
;         dispersion   - Approximate dispersion, in units of
;                        units(linelist)/units(wave_uncalib).  This is
;                        generally angstroms/pixel.  Default is 0.6,
;                        which is a very approximate value near 1.5 microns
;                        (though it changes with wavelength).  If
;                        using skylines to tweak an already-calibrated
;                        array, set the /tweak keyword, in which case
;                        the units are angstroms/ansgtrom, and the
;                        dispersion is set to 1.  This keyword has no
;                        effect if the /tweak keyword is set.
;         refwave      - FITS filename of an already-calibrated wave
;                        grid to use as a reference.  If set, the
;                        program will automatically identify arc lines
;                        based on the wavelengths in the refwave
;                        file.  The refwave array should have
;                        wavelengths very similar to the array being
;                        calibrated (within a resolution element.
;                        Default is to identify arc lines by hand.
;         refarc       - FITS filename of the arc lamp exposure from
;                        which the refwave wavelength solution was
;                        generated.  If set, program will attempt to
;                        offset the refwave wavelength solution by
;                        an integer number of pixels in x and y to
;                        match the input arcfile.  The setups used to
;                        take the two arcs should not be different by
;                        more than 40 pixels in x or y.  Default is to
;                        assume that the input arcfile and the arc
;                        used to generate the refwave wavelength
;                        solution were taken at very similar setups.
;         nstep        - Scalar, number of +/- steps in x and y away from
;                        zero offset to search for the offset between the
;                        input arc and the reference arc.  Default is 30.
;         slitrange    - 2-element array [n_below,n_above] giving the
;                        number of pixels above and below the center
;                        of the order to use for identifying lines.
;                        Useful for avoiding objects when identifying
;                        skylines.  Default is to use all illuminated
;                        pixels in the order.  Note: n_below should be
;                        a negative number if using pixels with smaller
;                        slit positions than the order center.
;         tweak        - If set, program assumes that a second arc or
;                        (sky exposure) is being used to tweak a
;                        previously made wavelength solution.  The
;                        dispersion is set to 1 (since the dispersion
;                        is in angstroms per angstrom), and a linear
;                        fit to the identified wavelengths is used
;                        (although the user can change this
;                        interactively).
;         auto         - Automatically identify lines without user
;                        input.  Recommended only if the /tweak
;                        keyword is set.
;         order        - 1D array specifying the orders to be fit.
;                        Numbers correspond to the entry number in the
;                        input orderlist, rather than the physical
;                        order number.
;         dark         - FITS filename of dark exposure to be subtracted
;                        from arc prior to identifying lines.  Recommended
;                        for NIR exposures.
;         arm          - String specifying which X-Shooter arm: 'UVB',
;                        'VIS', or 'NIR'.  Must be set.
;         ybin         - Factor by which to bin the arc exposure in the
;                        y-direction.  For example, if science exposures
;                        were taken in 1x2 binning, and arcs were taken
;                        in 1x1, then set ybin=2.
;
; OUTPUTS
;         wave_calib   - Name of file to which to write calibrated
;                        wave grid as a multi-extension FITS file.
;
; HISTORY
;         Adapted from MAGE_WAVECALIB 12/7/2009 GDB
;-----------------------------------------------------------------------------

if (n_params() lt 5) then begin
   print,'CALLING SEQUENCE: '
   print,'   xshooter_wavecalib,arcfile,orderlist,slitgrid,wave_uncalib,'
   print,'                wave_calib,linelist=linelist,dispersion=dispersion,'
   print,'                refwave=refwave,refarc=refarc,nstep=nstep,'
   print,'                slitrange=slitrange,/tweak,/auto,order=order,'
   print,'                dark=dark'
   return
endif

; Which arm?
if not(keyword_set(arm)) then begin
   print,'Must specify which X-Shooter arm'
   return
endif
; Allow upper or lower case
case arm of
   'uvb': arm = 'UVB'
   'UVB': arm = 'UVB'
   'vis': arm = 'VIS'
   'VIS': arm = 'VIS'
   'nir': arm = 'NIR'
   'NIR': arm = 'NIR'
   else : begin
             print,'X-Shooter arm not recognized'
             return
          end
endcase

; Get spectral format for this arm
format_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,3,4,4,4], $
          fieldnames: ['arm','phys_ordernum','wmin','wblaze','wmax'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
format_data   = read_ascii('~/pro/xshooter/xshooter_orders.dat',$
                           template=format_templ)
arm_ls        = where(format_data.arm eq arm)
format_order  = format_data.phys_ordernum(arm_ls)
format_wmin   = 10. * format_data.wmin(arm_ls)      ; Angstroms
format_wblaze = 10. * format_data.wblaze(arm_ls)
format_wmax   = 10. * format_data.wmax(arm_ls)

; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,3,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo    = read_ascii(orderlist,template=order_templ)
tmp_ordernum    = orderinfo.ordernum
tmp_ordercenter = orderinfo.center
tmp_orderedge1  = orderinfo.left
tmp_orderedge2  = orderinfo.right
; Sort orders in order of increasing slit position
order_order = sort(tmp_ordercenter)
ordernum    = tmp_ordernum(order_order)
ordercenter = tmp_ordercenter(order_order)
orderedge1  = tmp_orderedge1(order_order)
orderedge2  = tmp_orderedge2(order_order)
n_orders    = n_elements(ordercenter)

; Get physical order numbers
case arm of
   'UVB': begin
             ref_ordernum = 18
             ref_slitpos  = 864.
             sign         = 1    ; Phys order num increases with slitpos
          end
   'VIS': begin
             ref_ordernum = 20
             ref_slitpos  = 1253.
             sign         = -1   ; Phys order num decreases with slitpos
          end
   'NIR': begin
             ref_ordernum = 15
             ref_slitpos  = 619.
             sign         = -1   ; Phys order num decreases with slitpos
          end
   else : begin
             print,'Arm name not recognized'
             return
          end
endcase
ref_index = where(abs(ordercenter-ref_slitpos) eq $
                  min(abs(ordercenter-ref_slitpos)))
ref_index = ref_index(0)
phys_ordernum = ref_ordernum + sign*(ordernum - ordernum(ref_index))

; Optional - Specify the orders to fit.
if keyword_set(order) then begin
   n_fit         = n_elements(order)
   fit_ls        = intarr(n_fit)
   for i=0,n_fit-1 do fit_ls(i) = where(ordernum eq order(i))
   ordercenter   = ordercenter(fit_ls)
   ordernum      = ordernum(fit_ls)
   orderedge1    = orderedge1(fit_ls)
   orderedge2    = orderedge2(fit_ls)
   phys_ordernum = phys_ordernum(fit_ls)
   n_orders      = n_elements(ordercenter)
endif

; Read in arc, slit grid, and uncalibrated wavelength grid
arc     = xshooter_subbias(readfits(arcfile,harc),arm=arm)
slit    = readfits(slitgrid,hslit)
uncalib = readfits(wave_uncalib,huncalib)

; Bin arc exposure?
if keyword_set(ybin) then begin
   ybinsz   = float(ybin)
   orig_sz  = size(arc)
   orig_xsz = orig_sz(1)
   orig_ysz = orig_sz(2)
   arc = rebin(1.0*arc,orig_xsz,orig_ysz/ybinsz)
endif else begin
   ybinsz = 1.0
endelse

; Array size
sz  = size(arc)
xsz = sz(1)
ysz = sz(2)

; Arrays of x and y-values
sz  = size(arc)
xsz = sz(1)
ysz = sz(2)
xarr = fltarr(xsz,ysz)
yarr = fltarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i

; Optional - subtract dark exposure
if keyword_set(dark) then begin
   darkarr = readfits(dark)
   arc     = arc - darkarr
endif

; Masked region
mask = lonarr(xsz,ysz)
; No masking for UVB and VIS arms.  Orders are well away from
; pre/overscan regions
if (arm eq 'NIR') then begin
   ; Red edge x(y) polynomial
   r_terms = [1459.7389, 0.65775340, -0.00090907266, 3.2113316e-06]
   r_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(r_terms)-1 do r_yfit = r_yfit + r_terms(i)*(yarr^i)
   r_mask_ls = where(xarr gt r_yfit)
   ; Blue edge 
   b_terms = [433.63313, -0.30479697, 0.00011715153, -1.1558709e-06]
   b_yfit  = fltarr(xsz,ysz)
   for i=0,n_elements(b_terms)-1 do b_yfit = b_yfit + b_terms(i)*(yarr^i)
   b_mask_ls = where(xarr lt b_yfit)
   mask(r_mask_ls)     = 1
   mask(b_mask_ls)     = 1
   mask(0:3,*)         = 1
   mask(xsz-4:xsz-1,*) = 1
   mask(*,0:3)         = 1
endif

; Read in reference calibrated wave grid (optional)
; **** Set values for X-Shooter ****
if keyword_set(refwave) then begin
   print,'refwave keyword not yet implemented for x-shooter'
   return
   ;ref_wave = readfits(refwave,href)
   ;have_ref = 1
   ;; Offset the reference wave grid to match the input arc (optional)
   ;if (keyword_set(refarc) and not(keyword_set(tweak))) then begin
   ;   ;ref_arc = xshooter_subbias(readfits(refarc))
   ;   ref_arc = readfits(refarc)
   ;   ; Match the two arc exposures over a region likely to contain
   ;   ; several arc lines.  
   ;   print,'Calculating offset between arcs.  This may take a minute...'
   ;   nside = 100
   ;   ; Lower left
   ;   x_ll = 200.
   ;   y_ll = 200.
   ;   if keyword_set(nstep) then n_step = nstep else n_step = 30
   ;   offset_ll  = im_offset(ref_arc,arc,nx=n_step,ny=n_step,$
   ;                          subx=[x_ll-nside,x_ll+nside],$
   ;                          suby=[y_ll-nside,y_ll+nside],filter=0)
   ;   print,'Offset between arcs in lower left (x,y) =',strtrim(offset_ll,2)
   ;   ; Upper right
   ;   x_ur = 530.    ; This is more upper center, to avoid sat lines
   ;   y_ur = 1850.   ; in red and bad columns in blue
   ;   offset_ur  = im_offset(ref_arc,arc,nx=n_step,ny=n_step,$
   ;                          subx=[x_ur-nside,x_ur+nside],$
   ;                          suby=[y_ur-nside,y_ur+nside],filter=0)
   ;   print,'Offset between arcs in upper right (x,y) =',strtrim(offset_ur,2)
   ;   ; Shift, stretch, and rotate reference wavelength solution.
   ;   ref_wave = im_rot(ref_wave,[x_ll,x_ur],[y_ll,y_ur],$
   ;                     [x_ll+offset_ll(0),x_ur+offset_ur(0)],$
   ;                     [y_ll+offset_ll(1),y_ur+offset_ur(1)])
   ;endif
endif else begin
   ref_wave = -1
   have_ref = 0
endelse

; Line list
if keyword_set(linelist) then line_list = linelist else $
   line_list = '~/calib/nirspec/linelists/lowd_ir_ohlines.lst'  ; NIR
line_templ = {   $
          version: 1.0, $    
          datastart: 0L, $   
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 1L, $
          fieldtypes: 5L, $
          fieldnames: 'wavelength', $
          fieldlocations: 0L, $
          fieldgroups: [0L] $
        }
linedata  = read_ascii(line_list,template=line_templ)
linewaves = linedata.wavelength
linenames = strtrim(string(linewaves),2)

; Calibrated wavelength arrays
calib = 0d0*uncalib

; Loop through orders 
for i=0,n_orders-1 do begin

   print,' '
   print,'Order number '+strtrim(ordernum(i),2)+$
         ' ('+strtrim(phys_ordernum(i),2)+')'

   ; Determine the slit positions covered by this order by
   ; finding the midpoints between the current order and the
   ; adjacent orders.
   if (n_orders gt 1) then begin
      case i of
         0:    begin
                  slit_hi = 0.5*(orderedge2(i)+orderedge1(i+1))
                  slit_lo = slit_hi - (ordercenter(i+1)-ordercenter(i))
               end
         n_orders-1: begin
                  slit_lo = 0.5*(orderedge2(i-1)+orderedge1(i))
                  slit_hi = slit_lo + (ordercenter(i)-ordercenter(i-1))
               end
         else: begin
                  slit_lo = 0.5*(orderedge2(i-1)+orderedge1(i))
                  slit_hi = 0.5*(orderedge2(i)+orderedge1(i+1))
               endelse
      endcase   
      this_region = where(slit ge slit_lo and slit lt slit_hi,n_region)
   endif else begin
      this_region = where(slit ge orderedge1(0)-2 and $
                          slit le orderedge2(0)+2,n_region)
   endelse

   ; Select pixels just for wavelength calibration
   if keyword_set(slitrange) then begin
      n_below = slitrange(0)
      n_above = slitrange(1)
   endif else begin
      n_below = -1000
      n_above = +1000
   endelse

   ; Default is that there either is no reference wavelength
   ; solution, or the reference solution does not cover this
   ; order
   ref_covered = 0
   if have_ref then begin
      calib_ls = where(slit ge (ordercenter(i)+n_below > orderedge1(i)) and $
                       slit le (ordercenter(i)+n_above < orderedge2(i)) and $
                       mask eq 0,n_calib)
      if (n_calib gt 100) then ref_covered=1 else begin
         print,'Reference wavelength solution does not cover this order.'
         ref_covered = 0
      endelse
   endif
   if not(ref_covered) then begin
      calib_ls = where(slit ge ordercenter(i)+n_below > orderedge1(i) and $
                       slit le ordercenter(i)+n_above < orderedge2(i) and $
                       mask eq 0,n_calib)
   endif

   ; Get wavelength calibration coefficients
   coeffs = 0
   xid    = 0
   waveid = 0
   ; Calculate resolution in angstroms.  Assume 1.0" slit for UVB
   ; and 0.9" slit for VIS and NIR
   this_order = where(format_order eq phys_ordernum(i))
   this_order = this_order(0)
   wmid = format_wblaze(this_order)
   case arm of
      'UVB': resoln = wmid / 5100.
      'VIS': resoln = wmid / 8800.
      'NIR': resoln = wmid / 5600.
   endcase
   ; Calculate approximate dispersion in angstroms per binned pixel.
   if keyword_set(dispersion) then disp=dispersion else begin
      case arm of
         'UVB': pix_per_fwhm = 6.3  ; Unbinned values
         'VIS': pix_per_fwhm = 6.0
         'NIR': pix_per_fwhm = 4.0
      endcase
   endelse
   disp = ybinsz * resoln / pix_per_fwhm
   ; Optional - User-input dispersion, in units of calibrated
   ;            wavelength/uncalibrated wavelength
   if keyword_set(dispersion) then disp=dispersion
   ; Adjust a previously created wavelength solution?
   if keyword_set(tweak) then begin
      disp       = 1.0
      poly_order = 1
   endif else begin
      poly_order = 3
   endelse
   ; Perform fit automatically?
   if keyword_set(auto) then begin
      do_auto = 1
      no_show = 1
   endif else begin
      do_auto = 0
      no_show = 0
   endelse
   ; Limit for saturated pixels
   if (arm eq 'NIR') then satval=1e6 else satval=5e4
   ; Send only lines with wavelengths in the current order to
   ; the wave_calib widget
   wlo        = format_wmin(this_order) - 100
   whi        = format_wmax(this_order) + 100
   order_ls   = where(linewaves ge wlo and linewaves le whi)
   if ref_covered then begin
      ; Guesses at line positions based on reference wavelngth
      ; fit.
      this_order  = where(linewaves ge min(ref_wave(calib_ls)) and $
                          linewaves le max(ref_wave(calib_ls)),n_lines)
      guess_lines = linewaves(this_order)
      sort_ls     = sort(ref_wave(calib_ls))
      guess_x     = interpol(uncalib(calib_ls(sort_ls)),$
                             ref_wave(calib_ls(sort_ls)),guess_lines)
      wave_calib,uncalib(calib_ls),arc(calib_ls),linelist=linewaves(order_ls),$
                 coeffs=coeffs,xid=xid,waveid=waveid,dispersion=disp,$
                 resolution=resoln,order=poly_order,$
                 guess=[[guess_x],[guess_lines]],/rejectnoise,$
                 limit=limit,auto=do_auto,noshow=no_show
   endif else begin
      ; ID lines by hand
      wave_calib,uncalib(calib_ls),arc(calib_ls),linelist=linewaves(order_ls),$
                 coeffs=coeffs,xid=xid,waveid=waveid,dispersion=disp,$
                 resolution=resoln,order=poly_order,/rejectnoise,$
                 limit=limit,auto=do_auto,noshow=no_show
   endelse
   n_id = n_elements(xid)

   ; Write identified lines and fit coefficients after each order.
   ; Separate files are used in case the program crashes midway through.
   rootname = filename(wave_calib,/noextension,dirname=outdir)
   waveid_filename = $
      outdir+rootname+'_waveid_'+strtrim(ordernum(i),2)+'.dat'
   wavecoeffs_filename = $
      outdir+rootname+'_wavecoeffs_'+strtrim(ordernum(i),2)+'.dat'
   openw,id_unit,waveid_filename,/get_lun
   openw,coeff_unit,wavecoeffs_filename,/get_lun
   n_coeffs = n_elements(coeffs)
      case n_coeffs of
         1: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,1(g16))'
         2: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,2(g16))'
         3: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,3(g16))'
         4: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,4(g16))'
         5: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,5(g16))'
         6: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,6(g16))'
         7: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,7(g16))'
         8: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,8(g16))'
         9: printf,coeff_unit,ordernum(i),float(coeffs),format='(i6,9(g16))'
      endcase
   for h=0,n_id-1 do printf,id_unit,ordernum(i),xid(h),waveid(h)
   free_lun,id_unit
   free_lun,coeff_unit

   ; Calculate calibrated wavelength values.  If only one line was
   ; identified, perform a bulk shift.
   poly_order   = n_elements(coeffs) - 1
   temp_uncalib = uncalib(this_region)
   temp_calib   = 0d0*temp_uncalib
   if (n_id eq 1) then begin
      print,'Only one line found.'
      ; Wavelength shift of the identified line
      line_shift = waveid(0) - xid(0)
      ; If tweaking a previous wavelength solution, scale the 
      ; shift by the dispersion at each pixel.  
      if keyword_set(tweak) then begin
         ; Reconstruct wavelength solution and dispersion in the input
         ; wavelength solution (does not need to be exact).  Note:
         ; Calculated dispersion is in d(wave)/dy, which is not the same
         ; as d(wave)/d(uncalibrated wavelength pixels).
         old_terms    = poly_fit(yarr(calib_ls),uncalib(calib_ls),4)
         old_disparr  = fltarr(n_region)
         for j=1,3 do old_disparr = $
                old_disparr + j*old_terms(j)*((yarr(this_region))^(j-1))
         ; Dispersion at the wavelength of the identified line.
         sort_ls      = sort(temp_uncalib)
         line_disp    = interpol(old_disparr(sort_ls),temp_uncalib(sort_ls),$
                                 xid(0))
         ; Scale wavelength shift by the dispersion at each pixel
         scaled_shift = line_shift * (old_disparr/line_disp)
         temp_calib   = temp_uncalib + scaled_shift
         print,'Performing bulk shift scaled by dispersion:'
         print,'    Max shift = '+strtrim(max(scaled_shift),2)
         print,'    Min shift = '+strtrim(min(scaled_shift),2)
      endif else begin
         temp_calib   = temp_uncalib + line_shift
         print,'Performing bulk shift of '+strtrim(line_shift,2)
      endelse
   endif else begin
      ; Compute wavelengths based on fit to multiple lines
      for h=0,poly_order do temp_calib = temp_calib + $
                                         coeffs(h)*(temp_uncalib^h)
   endelse
   ; Save calibrated wavelength values
   calib(this_region) = temp_calib
endfor

; Write array(s) to FITS file.
writefits,wave_calib,calib,huncalib

return
end
